#include "pch.h"
#include <pthread.h>
#include "Optimise.h"

#define  Eps  0.0
//#define  Al  1		//Original
#define  Al  1
#define  Ga  2
#define  Be  0.5		//Original
//#define  Be  0.8

dataVec::dataVec(size_t size):vector<DOUBLE_TYPE>(size),m_error(0.0)
{
};

dataVec::dataVec(const dataVec & data) : vector<DOUBLE_TYPE>(data),m_error(data.m_error)
{
}

dataVec::~dataVec()
{
};

/*
dataVec & dataVec::operator = (const vector<DOUBLE_TYPE> & data)
{
	vector<DOUBLE_TYPE>::operator =(data);
	m_error = 0;
	return (*this);
}*/
dataVec & dataVec::operator = (DOUBLE_TYPE data)
{
	for (size_t i = 0; i < size();i++)
		at(i) = data;
	m_error = 0;
	return (*this);
}
dataVec & dataVec::operator /= (DOUBLE_TYPE data)
{
	for (size_t i = 0; i < size();i++)
		at(i) /= data;
	m_error = 0;
	return (*this);
}

dataVec & dataVec::operator = (const dataVec & data)
{
	vector<DOUBLE_TYPE>::operator =(data);
	m_error = data.m_error;
	return (*this);
}

dataVec & dataVec::operator += (const dataVec & data)
{
	for (size_t i = 0; i < size();i++)
		at(i) += data[i];
	m_error = 0;
	return (*this);
}

dataVec & dataVec::operator -= (const dataVec & data)
{
	for (size_t i = 0; i < size();i++)
		at(i) -= data[i];
	m_error = 0;
	return (*this);
}

dataVec operator * (const dataVec & data,DOUBLE_TYPE multiplier)
{
	dataVec dv(data.size());

	for (size_t i = 0; i < data.size();i++)
		dv[i] = data[i] * multiplier;
	return dv;
}

dataVec operator / (const dataVec & data,DOUBLE_TYPE divider)
{
	dataVec dv(data.size());

	for (size_t i = 0; i < data.size();i++)
		dv[i] = data[i] / divider;
	return dv;
}

dataVec operator + (const dataVec & data1,const dataVec & data2)
{
	dataVec dv(data1.size());

	for (size_t i = 0; i < data1.size();i++)
		dv[i] = data1[i] +  data2[i];
	return dv;
}

dataVec operator - (const dataVec & data1,const dataVec & data2)
{
	dataVec dv(data1.size());

	for (size_t i = 0; i < data1.size();i++)
		dv[i] = data1[i] -  data2[i];
	return dv;
}

/*
dataVec & dataVec::combine(const dataVec & dataA,double valA,const dataVec & dataB,double valB)
{
	for (size_t i = 0; i < m_size;i++)
		at(i) = dataA[i] * valA + dataB[i] * valB;
	m_error = 0;
	return (*this);
}
*/
dataVec & dataVec::noisyCombine(const dataVec & A,const dataVec & B,const dataVec & C)
{
	//	Make the choice as to which one to use based on the least significant bit of
	//	the randomise function
	int j = 1;
	int R = rand();
	for (size_t i = 0; i < size(); i++)
	{
		if (R & j)
			at(i) = C[i];
		else 
			at(i) = A[i];
		j *= 2;
		if (j > 65536)
		{
			j = 1;
			R = rand();
		}
	}
	return (*this);
}
void dataVec::log()
{
	size_t count = size();
	for (size_t i = 0;i < count;i++)
		at(i) = ::log(at(i));
}

void dataVec::exp()
{
	size_t count = size();
	for (size_t i = 0;i < count;i++)
		at(i) = ::exp(at(i));
}


DOUBLE_TYPE dataVec::average()
{
	size_t count = size();
	DOUBLE_TYPE ave = total();
	ave /= count;
	return ave;
}

DOUBLE_TYPE dataVec::total()
{
	DOUBLE_TYPE tot = 0;
	size_t count = size();
	for (size_t i = 0;i < count;i++)
		tot += at(i);
	return tot;
}

DOUBLE_TYPE dataVec::info_content()
{
	DOUBLE_TYPE info = 0;
	size_t count = size();
	for (size_t i = 0;i < count;i++)
	{
		if (at(i) > 0)
			info += at(i) * ::log(at(i))/Log2;
	}
	return info;
}


DOUBLE_TYPE dataVec::std_dev()
{
	size_t count = size();
	DOUBLE_TYPE ave = average();

	DOUBLE_TYPE sigf = 0;
	for (size_t i = 0;i < count; i++)
	{
		sigf += sqr(at(i) - ave);
	}

	sigf = sqrt(sigf/(count - 1));
	return sigf;

}

void optiVector::setOptimising(bool mode)
{
	for (size_t i = 0; i < size(); i++)
		at(i).optimise = mode;
}

DOUBLE_TYPE cumulateDifference(const dataVec & data1,const dataVec & data2)
{
	DOUBLE_TYPE difference = 0;
	for (size_t i = 0;i < min<size_t>(data1.size(),data2.size());i++)
	{
		difference += fabs(data1[i]-data2[i]);
	}
	return difference;
}

/*
void dataArray::resize(size_t size)
{
	vector<dataVec>::resize(size+1);
}
*/
dataArray::dataArray(size_t size): vector<dataVec>(size+1)
{
}

dataArray::~dataArray()
{
};


size_t dataArray::initialise(const optiDataType & data)
{
	//	A square array, plus one extra row for the reference 'M' row
	resize(1);
	at(0).clear();

	//	Setup the first vector in teh array with the entries in the
	//	input data that are to be optimised
	for (size_t i = 0;i < data.size();i++)
	{
		for (size_t j = 0;j < (*data[i]).size(); j++)
			if ((*data[i])[j].optimise)
				at(0).push_back((*data[i])[j].value);
	}

	//And then resize the master array so that there are N+1 entries
	resize(at(0).size()+1);

	//	And return the number of entries
	return at(0).size();
}

optiDataType dummyData;

COptimise::COptimise(const char * rootDir) : m_rootDir(rootDir),
	m_externalData(dummyData),
	m_recalcCount(0),m_totalIterations(0),m_outputCount(0)
{
}

COptimise::~COptimise(void)
{
}

void COptimise::loadExternalData(dataVec & data)
{
	//	Transfers from the internal arrays holding the current values of the parameters being
	//	optimised to the external data values that are used within the model
	size_t p = 0;
	for (size_t i = 0;i < m_externalData.size();i++)
	{
		for (size_t j = 0;j < m_externalData[i]->size(); j++)
			if ((*m_externalData[i])[j].optimise)
				(*m_externalData[i])[j].value = data[p++];
	}
}


void COptimise::CreateAndScoreOrthogonalVectors(DOUBLE_TYPE delta)
{
	size_t i;
	dataVec dx(m_size);

	//	We have N+1 copies of the orginal data
	for (i = 1;i <= m_size; i++)
		m_data[i] = m_data[0];

	//	Currently we move each of the vectors by the same absolute amount 
	//	in each dimension
	for (i = 0;i < m_size; i++)
		dx[i] = delta;

	//	Add the increments to create the N orthogonal vectors, with the O'th vector
	//	being the original vector
	for (i = 0;i < m_size; i++)
	{
		if (m_data[i+1][i] > dx[i])
			m_data[i+1][i] -= dx[i];
//		else if (m_data[i+1][i] == 0)
//			m_data[i+1][i] = dx[i];
		else
			m_data[i+1][i] += dx[i];
	}

	//	And now find the errors for each vector, 
	for (i = 0;i <= m_size; i++)
	{
		GetError(m_data[i]);
		SaveResults(m_data[i].Error() == m_ErrLowest);

	}

}


void COptimise::InitData(DOUBLE_TYPE delta)
{
	size_t i;
	m_size = m_data.initialise(m_externalData);

	DOUBLE_TYPE rmsTot = 0;
	for (i = 0; i < m_size;i++)
		rmsTot += sqr(m_data[0][i]);

	rmsTot = sqrt(rmsTot/m_size);

	m_recalcCount = m_totalIterations = m_outputCount = -(int)m_size;
	m_ErrLowest = 1E+20;

	CreateAndScoreOrthogonalVectors(delta);

}

typedef struct str_thdata
{
    int thread_no;
    char message[100];
} thdata;

void * test_func(void * ptr)
{
	thdata *data = (thdata *) ptr;  /* type cast to a pointer to thdata */
    
    /* do the work */
    printf("Thread %d says %s \n", data->thread_no, data->message);

	return NULL;
}


void COptimise::GetError(dataVec & data)
{
	m_totalIterations++;
	m_recalcCount++;
	m_outputCount++;

	loadExternalData(data);
	m_ErrLast = ErrorFunc();
	data.SetError(m_ErrLast);

	if (m_ErrLast < m_ErrLowest)
	{
		m_ErrLowest = m_ErrLast;
		m_DataLowest = data;
	}

}


DOUBLE_TYPE COptimise::optimise(optiDataType & data,int maxIter,int outputEvery,DOUBLE_TYPE delta)
{
	int average_count = 0;
	DOUBLE_TYPE last_err = 1000000;

	Err_pointer_lowest = 0;
	Err_pointer_highest = 0;
	Error_second_highest = 0;


	//	Pointers to the locations of the external data being optimised
	m_externalData = data;
/*
	For testing threading
	pthread_t pt;
	thdata data1;           

	data1.thread_no = 1;
    strcpy(data1.message, "Hello!");

	pthread_create(&pt,NULL,&test_func,(void*)&data1);
*/
	InitData(delta);

	bool Optiexit = false;

	//	As we don't change the contents of Xl, this can just be a pointer type
//	dataVec * Xl = &m_data[m_size];

	dataVec Xo(m_size);
	dataVec Xr(m_size);
	dataVec Xe(m_size);

	m_DataLowest.resize(m_size);

	do
	{
		
		Error_highest= 0;    
		Error_second_highest= 0;
		m_ErrLowest = 1.00E+20;                                     
		
		size_t I;
		for (I = 0 ;I <= m_size;I++)
		{
			if ( m_data[I].Error() > Error_highest )
			{
				Error_second_highest= Error_highest;
				Err_pointer_sec_highest= Err_pointer_highest;
				Error_highest= m_data[I].Error();
				Err_pointer_highest= I;
			}
			else if ( m_data[I].Error() > Error_second_highest )
			{
				Error_second_highest= m_data[I].Error();
				Err_pointer_sec_highest=I;
			};
			if ( m_data[I].Error() < m_ErrLowest )          
			{
				m_ErrLowest= m_data[I].Error();
				Err_pointer_lowest= I;
			}
		};

		

		// Check for exit.}
		if (m_totalIterations>=(int)maxIter)
			Optiexit=true;


		//	Pointers to the original data.  
		m_DataLowest=m_data[Err_pointer_lowest];
		const dataVec & Xh=m_data[Err_pointer_highest];

		Err= m_ErrLowest;

		// Check for exit.}
		if ( m_ErrLowest < Eps )
			Optiexit=true;
		else
		{
			//	Xo is the average of all the vectors except the worst
			Xo = m_data[0];
			for ( I= 1; I <= m_size;I++)
				Xo += m_data[I];
			Xo = (Xo - Xh)/m_size;

			//	Xr is the average, plus a bit pointing away from the worst result
			//	ie downhill is in the opposite direction from uphill
            //                      {CALCULATE VEC, yyy < =R Xr}
			//	Default Al = 2
			Xr = (Xo * (1+Al)) - (Xh * Al);
//			Xr = Xo - (Xh-Xo)*Al;
			GetError(Xr);

			if ( Xr.Error() == m_ErrLowest )           //                       {ERROR Fr<FL?}
			{
				SaveResults(true);
				//	New vector looks better, try a second that is the same more in the same direction
	
				Xe = (Xr * Ga) + (Xo * (1-Ga));	//Ga = 2
				GetError(Xe);

				//	Pick the one that gave the best results, and replace the worst with it
				if ( Xe.Error() == m_ErrLowest )                 //    {ERROR Fe<Error_lowest?}
				{
					SaveResults(true);

					cout << ".";
					dataVec Xe2(m_size);
					Xe2.noisyCombine(Xo,Xr,Xe);
					GetError(Xe2);
					if ( Xe2.Error() == m_ErrLowest )
					{

//						cout << "Noisy Double better than the lowest" << "\n";
						m_data[Err_pointer_highest]= Xe2;
						SaveResults(true);
					}
					else
					{
//						cout << "Double better than the lowest" << "\n";
						m_data[Err_pointer_highest]= Xe;
					}
				}
				else
				{
//					cout << "Better than the lowest" << "\n";
					m_data[Err_pointer_highest]= Xr;
				}
			}
			else
			{
				//	The new one we have found is better than the second worst, so use it to replace
				//	the worst, (which it must also be better than, by definition
				if ( Xr.Error() <= Error_second_highest )    //                 {S, yyy < =RE VEC, yyy < =R}
				{
//					cout << "Better than the second highest" << "\n";
					m_data[Err_pointer_highest]= Xr;
				}
				else
				{
					//	The new one is only better than the worst, so replace it
					if ( Xr.Error() < Error_highest )     //     {S, yyy < =RE VEC, yyy < =R Xr}
					{ 
//						cout << "Better than the highest" << "\n";
						m_data[Err_pointer_highest]= Xr;
					};

					//	Trial vector which is half the worst plus hald the average
					Xe = (Xh * Be)  + (Xo * (1-Be));
					GetError(Xe);

					//	If the average is worse than the highest then take more drastic action
					if ( Xe.Error() > Error_highest )
					{
						if ((last_err/m_ErrLowest) < 1.0001)
							average_count++;
						else
							average_count = 0;
						cout << "average count = " << average_count << "\n";
						last_err = m_ErrLowest;
						if (average_count > 10)
							Optiexit= true;




						Err = m_ErrLowest ;

						//{ The second term does not make much sense, but was there in the original}
						//	Anyway, Eps is zero, so it had no effect anyway
						if ( (Err>Eps) ) //|| (Err<Mx) )
						{
static 	bool averaging = true;
							//	Swap between two strategies for restarting if we get stuck.
							if (averaging)
							{
							//	Change all the vectors to be midway between the current value 
							//	and the lowest.  For the row that is the lowest, the values will
							//	end up being (essentially) unchanged
								cout << "AVERAGING.......\n";
								SaveProgress("Averaging");
								for ( I= 0 ; I <= m_size ;I++)
								{	//                                  { 36 }
									//								m_data[I].combine(m_data[I],0.5,Xl,0.5);
									m_data[I] = (m_data[I]* 0.5) + (m_DataLowest * 0.5);
									GetError(m_data[I]);
									if ( m_data[I].Error() == m_ErrLowest )
									{
										SaveResults(true);
									}
								};
								cout << "finished AVERAGING\n";
							}
							else
							{
//								double delta = cumulateDifference(Xl,Xh)/10;
								char msg[100];
								sprintf(msg,"Reorthogonalising -  delta = %f",delta);
								SaveProgress(msg);
								cout << msg << "\n";
								m_data[0] = m_DataLowest;
								CreateAndScoreOrthogonalVectors(delta);
								cout << "Finished reorthogonalising\n";

							}
							averaging = !averaging;
						}
						else
							Optiexit= true;
					}
					else
					{
						m_data[Err_pointer_highest]= Xe;
					};
				};
			};
		};
		if (m_outputCount > outputEvery)
		{
			m_outputCount = 0;
			loadExternalData(m_DataLowest);
			SaveResults(false);
		}
	} while (!Optiexit);

	loadExternalData(m_DataLowest);
	SaveResults(false);

	return m_ErrLowest.WithWeightings;
}

#define PROGRESS_FILENAME "progress.csv"


void COptimise::SaveProgress(const char * status)
{
	FILE * f = fopen((m_rootDir + PROGRESS_FILENAME).c_str(),"a");
	if (f == 0) 
	{
		cout << "Failed to open " << PROGRESS_FILENAME << "\n";
	}
	else
	{
		fprintf(f,"%i,%f,%f",m_totalIterations,(status)?m_ErrLowest.WithWeightings:m_ErrLast.WithWeightings,
			(status)?m_ErrLowest.NoWeightings:m_ErrLast.NoWeightings);
		if (status)
			fprintf(f,",%s",status);
		fprintf(f,"\n");

		fclose(f);
	}
}
