#pragma once
#include <vector>
using namespace std;

const float Log2 = log((float)2.0);

#ifdef _DEBUG
#define DOUBLE_TYPE float
#define _USE_FLOAT
#else
#define DOUBLE_TYPE double
#endif


//	Holds the information relating to an item to be optimised as a pair
//	The 'first' item of the pair is the value being iptimised, and the second 
//	indicates whether optimisation is currently enabled
class optiItem
{
public:
	DOUBLE_TYPE value;
	bool optimise;
	//	Constructor.  Pass in the initial value and whether it is beinh optimised
	optiItem(DOUBLE_TYPE v,bool o):value(v),optimise(o){};
	optiItem():value(0),optimise(false){};
	
	void readFloat(const char * buffer) {
		value = (DOUBLE_TYPE)atof(buffer);
		const char * p = strchr(buffer,',');
		if (p)
			optimise = (*(p+1) == 'Y');
	};

};

//	A vector of values being optimised
struct optiVector: public vector<optiItem>
{
	optiVector(size_t i= 0,optiItem oi = optiItem()):vector<optiItem>(i,oi){};
	bool optimising()
	{
		for (size_t i = 0;i < size();i++)
			if (at(i).optimise)
				return true;
		return false;
	};
	void setOptimising(bool mode = true);
};

//	A set of such vectors
typedef vector<optiVector *> optiDataType;

/*	A class to contain both the error value representing the basic difference
	between the target and the model and the value that includes additional weightings
	to prevent negative coefficients etc.  Cast the duplet to the latter so that it
	is used with comparison operators during optimisation
	*/
struct ErrorPair {
	DOUBLE_TYPE WithWeightings;
	DOUBLE_TYPE NoWeightings;

	ErrorPair(DOUBLE_TYPE err = 0.0):WithWeightings(err),NoWeightings(err){};
	void operator = (const ErrorPair & val){
		WithWeightings = val.WithWeightings;
		NoWeightings = val.NoWeightings;
	};
	operator DOUBLE_TYPE() const {return WithWeightings;};
};


//	Internal class.  A consolidated set of the items being optimised
class dataVec	: public vector<DOUBLE_TYPE>
{
	ErrorPair m_error;
public:

	//	The current error value associated with this set
	ErrorPair Error(){return m_error;};
	void SetError(ErrorPair error){ m_error = error;};

	dataVec(size_t size=0);
	~dataVec();

	//Don't instantiate this to see if the compiler needs it.  If it does then there will be a 
	//	link error
	dataVec(const dataVec & data);	

	//	For accessing members
	//	For setting
//	dataVec & operator = (const vector<DOUBLE_TYPE> & data);
	dataVec & operator = (const dataVec & data);
	dataVec & operator += (const dataVec & data);
	dataVec & operator -= (const dataVec & data);
	dataVec & operator = (DOUBLE_TYPE val);
	dataVec & operator /= (DOUBLE_TYPE val);

	//For combining
	dataVec & noisyCombine(const dataVec & A,const dataVec & B,const dataVec & C);
	DOUBLE_TYPE average();
	DOUBLE_TYPE total();
	DOUBLE_TYPE info_content();
	DOUBLE_TYPE std_dev();

	void log();
	void exp();

};

dataVec operator * (const dataVec & data,DOUBLE_TYPE multiplier);
dataVec operator / (const dataVec & data,DOUBLE_TYPE divider);
dataVec operator + (const dataVec & data1,const dataVec & data2);
dataVec operator - (const dataVec & data1,const dataVec & data2);
DOUBLE_TYPE cumulateDifference(const dataVec & data1,const dataVec & data2);

class dataArray	: public vector<dataVec>
{
public:
	dataArray(size_t size=0);
	~dataArray();

//	void resize(size_t size);
	size_t initialise(const optiDataType & data);
};

class COptimise
{
	size_t m_size;
	dataArray m_data;

	optiDataType & m_externalData;
	int	m_maxIter;
	int m_outputEvery;

	DOUBLE_TYPE 	Error_highest, Error_second_highest,Eps,Err;
	size_t   Err_pointer_highest, Err_pointer_sec_highest, Err_pointer_lowest;

protected:
	string m_rootDir;

	int m_totalIterations;
	int m_recalcCount;
	int m_outputCount;
		
	ErrorPair m_ErrLowest;
	ErrorPair m_ErrLast;

	dataVec m_DataLowest;

public:
	COptimise(const char * rootDir = "");
	~COptimise(void);

	void InitData(DOUBLE_TYPE delta);
	void CreateAndScoreOrthogonalVectors(DOUBLE_TYPE delta);

	DOUBLE_TYPE optimise(optiDataType & data,int maxIter,int outputEvery,DOUBLE_TYPE delta);

	void loadExternalData(dataVec & data);

	void GetError(dataVec & data);

	//	To be implemented by the class representing the data to be optimised
	virtual ErrorPair ErrorFunc() = 0;
	virtual void SaveResults(bool toFile) = 0;

	void SaveProgress(const char * status = 0);
};

