
#include "pch.h"
#include "alignFile.h"
#include "mRNAinfo.h"


alignFile::alignFile(){};

bool alignFile::read(alignment & data)
{
	if(!readLine())
		return false;
	if (strncmp(m_line,"ERR",3)== 0)
	{
		data.id = lineData(' ');
	}
	else if (strncmp(m_line,"SRR",3)== 0)
	{
		data.id = lineData(' ');
	}
	else
	{
		lineData(':');
		data.id = lineData('#');
	}
	lineData('\t');

	data.forward = (*lineData('\t') == '+');
	data.gene = lineData('\t');
	data.chromosome = geneToChromosome(data.gene);
	data.location = atoi(lineData('\t'));
	data.sequence = lineData('\t');
	return true;
}

geneAlignMents * genomeAlignMents::find(const string & gene)
{
	iterator i = map<string,geneAlignMents>::find(gene);
	if (i == end()) return 0;
	return &(*i).second;
};

void genomeAlignMents::save(const string & filename)
{
	FILE * f = fopen(filename.c_str(),"w");
	for (genomeAlignMents::const_iterator i = begin(); i != end(); i++)
	{
		string geneName = (*i).first;
		const geneAlignMents & ga = (*i).second;
		for (geneAlignMents::const_iterator j = ga.begin();j != ga.end();j++)
		{
			alignPair ap = (*j).second;
			fprintf(f,"%s,%s,%i,%i\n",geneName.c_str(),ap.id.c_str(),ap.start,ap.end);
		}
	}
	fclose(f);
}

#define MAX_LENGTH 4000

void genomeAlignMents::saveLengths(const string & filename)
{
	vector<int> lengths;
	lengths.resize(MAX_LENGTH+1);
	for (genomeAlignMents::const_iterator i = begin(); i != end(); i++)
	{
		const geneAlignMents & ga = (*i).second;
		for (geneAlignMents::const_iterator j = ga.begin();j != ga.end();j++)
		{
			alignPair ap = (*j).second;
			lengths[min<size_t>(MAX_LENGTH,ap.length())]++;
		}
	}
	FILE * f = fopen(filename.c_str(),"w");
	if (f == 0)
	{
		cout << "Unable to open " << filename << "\n";
	}
	else
	{
		for (size_t i = 0;i < lengths.size(); i++)
		{
			fprintf(f,"%i,%i\n",i,lengths[i]);
		}
		fclose(f);
	}
}

bool genomeAlignMents::open(const string & filename)
{
	if(!dataFile::open(filename))
		return false;
	alignPair ap;

	while (readLine())
	{
		string gene = lineData(',');
		ap.id = lineData(',');
		ap.start = atoi(lineData(','));
		ap.end = atoi(lineData());
		geneAlignMents * ga = find(gene);
		if (ga == 0)
		{
			genomeAlignMents::iterator i = insert(
				genomeAlignMents::value_type(gene,geneAlignMents())).first;
			ga = &((*i).second);
		}
		ga ->insert(geneAlignMents::value_type(ap.start,ap));
	}
	return true;
}

