#include "pch.h"
#include "breakData.h"

#define RECREATE_TOTAL		//Recreates total from the data for the individual chromosome
							//Allows mitochondrial DNA to be excluded

bool breakData::m_forward = false;

int fromRoman(const char * s)
{
	int N = 0;
	int I = 0;
	for (size_t i = 0; i < strlen(s); i++)
	{
		switch (s[i])
		{
		case 'I': N++; I++; break;
		case 'V':
				if (I == 1) N += 3; else N += 5; break;
		}
	}
	return N;
}


breakDataInfo::breakDataInfo() :
	m_offset(0),m_length(0)
	{}

void breakDataInfo::setup(int offset,int length) 
{
	m_offset = offset;
	m_length = length;

	m_addVal = 1;
	m_mask = 0x03;
	for (int i = 0; i < m_length-1; i++)
	{
		m_mask <<= 2;
		m_mask |= 0x03;
		m_addVal *= 4;
	}
};



int breakDataInfo::asIndex(const char * seq)
{
	int index = 0;
	for (size_t i = 0; i < min<size_t>(m_length,strlen(seq));i++)
	{
		index *=4;
		switch (seq[i])
		{
		case 'A': case 'a': break;
		case 'C': case 'c': index += 1; break;
		case 'G': case 'g': index += 2; break;
		case 'T': case 't': index += 3; break;
		default : case 'N': return -1;
		}
	}
	return index;
}

string breakDataInfo::asString(size_t val)
{
	char buff[100];
	buff[m_length] = 0;

	for (int i = m_length-1;i >= 0; i--)
	{
		switch (val & 0x03)
		{
		case 0:
			buff[i] = 'A'; break;
		case 1:
			buff[i] = 'C'; break;
		case 2:
			buff[i] = 'G'; break;
		case 3:
			buff[i] = 'T'; break;
		}
		val >>= 2;
	}
	return buff;
}



breakData::breakData(breakDataInfo & bdi):	m_bdi(bdi),m_totalBreaks(0),
	m_totalDupNucleotides(0),m_totalUniqueNucleotides(0),m_totalWeightedUniqueNucleotides(0),m_Ncounter(0)
/*
#ifdef COMBINED_NUCS
			m_data(1 << (m_posOffset-m_negOffset+1))
#else
			m_data(1 << ((m_posOffset-m_negOffset+1) * 2))
#endif
*/
{
	m_dataLength = 1 << (m_bdi.length() * 2);

#ifdef _DEBUG
	m_data.resize(m_dataLength);
#else
	m_data = new countInfo[m_dataLength];
#endif
}
breakData::breakData(const breakData & data ) :m_bdi(data.m_bdi)
{
	operator = (data);
}

breakData & breakData::operator = (const breakData & data )
{
	m_bdi = data.m_bdi;
	chrName = data.chrName;
	m_totalBreaks = data.m_totalBreaks;
	m_totalUniqueNucleotides = data.m_totalUniqueNucleotides;
	m_totalWeightedUniqueNucleotides = data.m_totalWeightedUniqueNucleotides;
	m_totalDupNucleotides = data.m_totalDupNucleotides;
	m_currPtr = 0;
	m_Ncounter = 0;
	m_dataLength = data.m_dataLength;
#ifdef _DEBUG
	m_data = data.m_data;
#else
	m_data = new countInfo[m_dataLength];
	for (int i = 0;i < m_dataLength;i++)
		m_data[i] = data.m_data[i];
#endif
	return *this;
}


breakData::~breakData()
{
#ifndef _DEBUG
	delete [] m_data;
#endif
}

breakData & breakData::operator += (const breakData & bs)
{
	for (int i = 0; i < m_dataLength;i++)
	{
		m_data[i].count += bs.m_data[i].count;
		m_data[i].weightedCount += bs.m_data[i].weightedCount;
		m_data[i].dups += bs.m_data[i].dups;
		m_data[i].breaks += bs.m_data[i].breaks;
		m_data[i].scaledBreaks += bs.m_data[i].scaledBreaks;
	}
	m_totalBreaks += bs.m_totalBreaks;
	m_totalDupNucleotides += bs.m_totalDupNucleotides;
	m_totalUniqueNucleotides += bs.m_totalUniqueNucleotides;
	m_totalWeightedUniqueNucleotides += bs.m_totalWeightedUniqueNucleotides;

	return *this;
}

void breakData::setForward(bool forward)
{
	m_forward = forward;
}

bool breakDataSet::isSetup()
{
	return !empty();
}

void breakDataSet::resize(size_t s)
{
	vector<breakData>::resize(s+1,breakData(m_bdi));
	for (size_t i = 0;i < size();i++)
	{
#ifdef _DEBUG
		for (int j = 0;j < at(i).m_dataLength; j++)
			at(i).m_data[j].m_sequence = m_bdi.asString(j);
#endif
		char buffer[100];
		sprintf(buffer,"Chr %i",i);
		at(i).chrName = buffer;
	}
	at(0).chrName = "ref";
}

void breakDataSet::resetData()
{
	for (size_t i = 0;i < size();i++)
	{
		at(i).resetData();
	}
}

/*
breakDataSet::breakDataSet(size_t data_size)
{
	resize(data_size,breakData(m_bdi));
#ifdef _DEBUG
	for (size_t i = 0;i < size();i++)
	{
		for (int j = 0;j < at(i).m_dataLength; j++)
			at(i).m_data[j].m_sequence = m_bdi.asString(j);
	}
#endif

}
*/
breakDataSet::breakDataSet(int offset,int length)
{
	m_bdi.setup(offset,length);
}

PTR_TYPE breakData::addData(const char * nucleotides,float breaks)
{
	if (setCurrPtr(nucleotides) == -1) {
		//	Counter for skipping over Ns
		m_Ncounter = m_bdi.length();
	}
	else {
		m_totalBreaks += breaks;
		m_totalUniqueNucleotides++;
		m_totalWeightedUniqueNucleotides += 1;

		setCurrPtr(nucleotides);
		m_data[m_currPtr].count++;
		m_data[m_currPtr].breaks += breaks;
	}
	return m_currPtr;
}

#define SUBCOUNTER 100

PTR_TYPE breakData::addData(const char nucleotide,float breaks,bool duplicate,double weight)
{
static int subsetCounter = SUBCOUNTER;

	if (addChar(nucleotide) == -1) {
		//	Counter for skipping over Ns
		m_Ncounter = m_bdi.length();
	}
	else {
		if (m_Ncounter == 0)
		{
//			subsetCounter--;
//			if (subsetCounter <= 0)
//			{
			m_totalBreaks += breaks;
			if (duplicate)
			{
				m_totalDupNucleotides += weight;
				m_data[m_currPtr].dups++;
			}
			else
			{
				m_totalUniqueNucleotides ++;
				m_totalWeightedUniqueNucleotides += weight;
				m_data[m_currPtr].weightedCount += (float)weight;
				m_data[m_currPtr].count ++;
			}

			m_data[m_currPtr].breaks += breaks;
//				subsetCounter = SUBCOUNTER;
//			}
		}
		else
			m_Ncounter--;
	}
	return m_currPtr;

}

double breakData::normaliser(int nucleotide_type,int threshold)
{
	if(nucleotide_type == -1)
		nucleotide_type = (int)m_currPtr;
	if (m_data[nucleotide_type].count >= threshold)
	{
		if(m_data[nucleotide_type].scaledBreaks)
			return 1.0/m_data[nucleotide_type].scaledBreaks;
		else
			return 1.0;
	}
	else
		return 1.0;
}

void breakData::resetData()
{
	for (int i = 0; i < m_dataLength; i++)
	{
		m_data[i].breaks = 0;
		m_data[i].count = 0;
		m_data[i].dups = 0;
		m_data[i].scaledBreaks = 0;
		m_data[i].weightedCount = 0;
	}
	m_totalDupNucleotides = 0;
	m_totalWeightedUniqueNucleotides = 0;
	m_totalBreaks = 0;
	m_totalUniqueNucleotides = 0;
}
void breakData::normalise(bool logNorm)
{
	int i;
	for (i = 0; i < m_dataLength; i++)
	{
		m_data[i].scaledBreaks = (1.0 * m_dataLength * m_data[i].breaks) / m_totalBreaks;
	}

	if (m_totalWeightedUniqueNucleotides == 0)
	{
		for (i = 0; i < m_dataLength; i++)
		{
			//	We have unweighted data
			m_data[i].scaledBreaks *=  (double)m_totalUniqueNucleotides /((double)m_data[i].count * m_dataLength);
			if (logNorm)
#ifdef LOG2NORM
				if (m_data[i].scaledBreaks)
					m_data[i].scaledBreaks = log(m_data[i].scaledBreaks)/Log2;
				else
					m_data[i].scaledBreaks = DEFAULT_IF_NOBREAKS;
#else
				m_data[i].scaledBreaks = log(m_data[i].scaledBreaks + 1);
#endif
		}
	}
	else
	{
		for (i = 0; i < m_dataLength; i++)
		{
			m_data[i].scaledBreaks *=  (double)m_totalWeightedUniqueNucleotides /((double)m_data[i].weightedCount * m_dataLength);
			if (logNorm)
#ifdef LOG2NORM
				if (m_data[i].scaledBreaks)
					m_data[i].scaledBreaks = log(m_data[i].scaledBreaks)/Log2;
				else
					m_data[i].scaledBreaks = DEFAULT_IF_NOBREAKS;
#else
				m_data[i].scaledBreaks = log(m_data[i].scaledBreaks + 1);
#endif
		}
	}
}
void breakData::equilise()
{
	for (int i = 0; i < m_dataLength; i++)
	{
		m_data[i].scaledBreaks = (1.0 * m_dataLength * m_data[i].breaks) / m_totalBreaks;
	}
}



PTR_TYPE breakData::setCurrPtr(const char * str)
{
	m_currPtr = 0;
	for (int i = 0; i < m_bdi.length(); i++)
		addChar(str[i]);
	return m_currPtr;
}

PTR_TYPE breakData::addChar(const char c)
{
	if (m_forward)
	{
		m_currPtr <<= 2;

		switch (c)
		{
		case 'A':
		case 'a':
			break;
		case 'C':
		case 'c':
			m_currPtr += 1;
			break;
		case 'G':
		case 'g':
			m_currPtr += 2;
			break;
		case 'T':
		case 't':
			m_currPtr += 3;
			break;
		default:
			m_currPtr &= m_bdi.mask();
			return -1;
		}
	}
	else
	{
		m_currPtr >>= 2;

		switch (c)
		{
		case 'A':
		case 'a':
			m_currPtr += (3 * m_bdi.addVal());
			break;
		case 'C':
		case 'c':
			m_currPtr += (2 * m_bdi.addVal());
			break;
		case 'G':
		case 'g':
			m_currPtr += m_bdi.addVal();
			break;
		case 'T':
		case 't':
			break;
		default:
			m_currPtr &= m_bdi.mask();
			return -1;
		}
	}
	m_currPtr &= m_bdi.mask();
	return m_currPtr;
}
void breakData::addData(int position,float count,unsigned int dup, double breaks,double norm)
{
	m_data[position].count += count;
	m_data[position].dups += dup;
	m_data[position].breaks += breaks;
	m_data[position].scaledBreaks += norm;

	m_totalUniqueNucleotides += count;
	m_totalDupNucleotides += dup;
	m_totalBreaks += breaks;
}





void breakDataSet::output(const string & filename,int Nchrs,bool full)
{
	FILE * f = fopen(filename.c_str(),"w");

	if (f == 0)
	{
		cout << "Failed to open file\n";
		return;
	}
	size_t i;
	fprintf(f,"index,string");
	for (i = 0; i < size();i++)
	{
		if (full)
			fprintf(f,",%s,%s brks,%s norm,%s dups",at(i).chrName.c_str(),at(i).chrName.c_str(),at(i).chrName.c_str(),at(i).chrName.c_str());
		else
			fprintf(f,",%s",at(i).chrName.c_str());
	}
	fprintf(f,"\n");

	if (Nchrs == -1)
		Nchrs = (int)size();

	size_t scale = at(0).m_dataLength;
	for (size_t i = 0;i < scale; i++)
	{
		fprintf(f,"%i,%s",i,m_bdi.asString(i).c_str());
		for (int j = 0; j < Nchrs;j++)
		{
			if (full)
			fprintf(f,",%f,%f,%f,%f",at(j).m_data[i].weightedCount,
				(double)at(j).m_data[i].breaks,
				at(j).m_data[i].scaledBreaks,
				at(j).m_data[i].dups);
			else
				fprintf(f,",%f",at(j).m_data[i].weightedCount);
		}
		fprintf(f,"\n");
	}

	fclose(f);

}

bool breakDataSet::open(const string & filename,bool logNorm,int & Nchrs)
{
	bool combineChromosomes = (Nchrs < 0);

	/*	Opens a previously stored file of break data.  The file is normally stored
		with an entry for all 65536 combinations of  8 nucleotides.   If we are using paired
		nucleotides then the A's and T's are combined as are the C's and G's by manipulation
		of the index.

		At the end we renormalise because of the fact that we may have combined data, and we
		dont rely on the log stats of the orignal results, just the absolute counts.
	*/

	if (!dataFile::open(filename))
		return false;

	const char * pos;
	int offset = 0;
	int length = 0;

	if (pos = strstr(filename.c_str(),".bar"))
	{
		if (pos = strstr(pos,"_o"))
		{
			offset = atoi(pos+2);
			if (pos = strstr(pos,"_l"))
			{
				length = atoi(pos+2);
			}
		}
	}

	readLine();
	bool initLineIsIndex = false;
	if (strncmp(m_line,"Summary,",8)==0)
	{
		lineData(',');
		int length = atoi(lineData(','));
		int offset = atoi(lineData(','));
		m_bdi.setup(offset,length);
		resize(0);	//An entry just for the summary

		for (int i = 0;i < 65536;i++)
		{
			readLine();
			lineData(',');
			lineData(',');
			at(0).m_data[i].scaledBreaks = atof(lineData(','));

			if (!logNorm)
				at(0).m_data[i].scaledBreaks = 
					exp(at(0).m_data[i].scaledBreaks * Log2);

		}
	}
	else
	{
		if (strcmp(lineData(','),"index") == 0)
			initLineIsIndex = true;

		readLine();
		if (initLineIsIndex) lineData(',');
		const char * seq = lineData(',');
		if (length == 0)
			length = (int)strlen(seq);
		m_bdi.setup(offset,length);

		rewind();
		readLine();

		//	Read in the first line and get the chromosome names
		if (initLineIsIndex) lineData(',');
		vector<string> chrNames;

		lineData(',');
		const char * s = lineData(',');
		while ((s) && strlen(s))
		{

			if ((strcmp(s,"ref") == 0) || (strcmp(s,"all") == 0))
				chrNames.push_back("all");
			else
			{
				//	There is a bug in the generattion of the files that result in empty columns
				//	at the end labelled, for example  'chr 22' and 'chr 23'.  Check for these by
				//	checking for spaces.  Real chromosomes don't have spaces
				if (strchr(s,' ') == NULL)
					chrNames.push_back(s);
			}
			lineData(',');
			lineData(',');
			lineData(',');
			s = lineData(',');
		}
		if (Nchrs <= 0)
			Nchrs = (int)chrNames.size() -1;
		resize(Nchrs);	//An entry for each chromosome, plus the initial s						//  The resize function adds a column for the ref data

		multimap<int,int> chrMap;
		bool isRoman = false;
		for (int i = 0;i <= Nchrs; i++)
		{
			if (i ==0)
				at(i).m_chrNumber = 0;
			else
			{
				if (i == 1)
				{
					if (chrNames[i][3] == 'I')
						isRoman = true;
				}
				if (chrNames[i][3] == 'X')
					at(i).m_chrNumber = 100;
				else if (chrNames[i][3] == 'Y')
					at(i).m_chrNumber = 101;
				else if (chrNames[i][3] == 'M')
					at(i).m_chrNumber = 102;
				else if (isRoman)
					at(i).m_chrNumber = fromRoman(chrNames[i].c_str()+3);
				else
					at(i).m_chrNumber = atoi(chrNames[i].c_str()+3);
			}
			chrMap.insert(multimap<int,int>::value_type(at(i).m_chrNumber,i));

			at(i).chrName = chrNames[i];
		}

		for(multimap<int,int>::iterator i =  chrMap.begin(); i != chrMap.end();i++)
		{
			m_chrMap.push_back((*i).second);
		}

		m_seqs.resize(at(0).m_dataLength);

		int lineCounter = 0;
		int index = -1;
		while (	readLine())
		{
			if (initLineIsIndex) 
				index = atoi(lineData(','));
			else
				index++;

			m_seqs[index] = lineData(',');

			for (int i = 0;(i <= Nchrs) && dataRemains();i++)
			{
				float count = (float)atof(lineData(','));
				double breaks = atof(lineData(','));
				double norm = atof(lineData(','));
				int dups = atoi(lineData(','));
				at(i).addData(index,count,dups,breaks,norm);
			}
		}

#ifdef RECREATE_TOTAL
		chr(0).resetData();
		for(int i = 1;i <= Nchrs;i++)
		{
			switch(chr(i).chrName[3])
			{
			case 'M': break;
			default:
				chr(0) += chr(i);
			}
		}
#endif

		if (combineChromosomes)
		{
			string lastChr[2];
			m_combinedChrList[0] = (chr(1).chrName);
			if (Nchrs > 1) 
				m_combinedChrList[1] = (chr(2).chrName);
			for(int i = 3;i <= Nchrs;i++)
			{
				switch(chr(i).chrName[3])
				{
				case 'M': break;
				default:
					if (chr(1).TotalUniqueNucleotides() < chr(2).TotalUniqueNucleotides())
					{
						chr(1) += chr(i);
						if (i < 8)
							(m_combinedChrList[0] += ";") += chr(i).chrName;
						else 
							lastChr[0] = chr(i).chrName;
					}
					else
					{
						chr(2) += chr(i);
						if (i < 8)
							(m_combinedChrList[1] += ";") += chr(i).chrName;
						else 
							lastChr[1] = chr(i).chrName;
					}
				}
			}
			if (!lastChr[0].empty())
				(m_combinedChrList[0] += "..") += lastChr[0];
			if (!lastChr[1].empty())
				(m_combinedChrList[1] += "..") += lastChr[1];

			if (Nchrs > 2)
				Nchrs = 2;
		}

		//	Recreate the 0th data, excluding the mitochondrial contribution

		//	Need to recalculate the ratios as the originals may not be correct
		for (int i = 0;i <= Nchrs; i++)
		{
			chr(i).normalise(logNorm);
		}
	}

	close();
	return true;
}


