#ifndef BREAK_DATA_H
#define BREAK_DATA_H

#include <vector>
#include <map>
#include <string>
using namespace std;
//#define PTR_TYPE unsigned long long
#define PTR_TYPE size_t

#include "pssmData.h"
#include "dataFile.h"
//	Use this #define to combine together A&T and also C&G
//#define COMBINED_NUCS 1


#define LOG2NORM
#define DEFAULT_IF_NOBREAKS -1

int fromRoman(const char * s);

class countInfo
{
public:
	float count;	//float rather than int to cope with scaling for RNA data
	float weightedCount;	//float rather than int to cope with scaling for RNA data
	double scaledBreaks;
	float dups;
	double breaks;
#ifdef _DEBUG
	string m_sequence;
#endif
	countInfo():count(0),breaks(0),scaledBreaks(0),dups(0),weightedCount(0){};
};

//	Note that offset no longer seems to be being used by anything within readData
//	which is odd
class breakDataInfo
{
	friend class breakDataSet;
	int m_offset,m_length,m_mask,m_addVal;
public:
	void setup(int offset,int length);
	breakDataInfo();
	string asString(size_t val);
	int asIndex(const char * seq);
	int offset(){return m_offset;};
	int length(){return m_length;};
	int mask(){return m_mask;};
	int addVal(){return m_addVal;};

};

class breakData
{
	friend class breakDataSet;

public:
	breakDataInfo & m_bdi;
	string chrName;
	int m_chrNumber;	//The character number, derived from the string
private:
	double m_totalBreaks;
	double m_totalUniqueNucleotides;
	double m_totalWeightedUniqueNucleotides;
	double m_totalDupNucleotides;
	PTR_TYPE m_currPtr;
	int m_Ncounter;

	//	Use a C++ array rather than a vector for efficiency
protected:
	int m_dataLength;
public:
#ifdef _DEBUG
	vector<countInfo> m_data;
#else
	countInfo * m_data;
#endif


public:

	//	how far to the left and the right of the end of the segment we acquire data
	static void setForward(bool forward = true);
	static bool m_forward;
	size_t size(){ return m_dataLength;};
	double TotalUniqueNucleotides() {return m_totalUniqueNucleotides;};



	countInfo & operator [](size_t i){return m_data[i];};

	//	We need a copy constructor because it is used by vector operations, and teh default
	//	version means that all copies share the same array of countInfos!
	breakData(breakDataInfo & bdi);
	breakData(const breakData & data);
	~breakData();

	breakData & operator =(const breakData & bdi);
	breakData & operator += (const breakData & bs);

	//	Add info about the number of reads at the position 'posOffset' to the left of 
	//	the nucleotide passed to the method, or negOffset to the right of the start of the
	//	string that is passed in
	PTR_TYPE addChar(char c);
	PTR_TYPE addData(const char * nucleotides,float reads);
	PTR_TYPE addData(const char nucleotide,float reads,bool duplicate = false,double weight = 1);
	void addData(int position,float count,unsigned int dup, double reads, double norm);

	//	
	PTR_TYPE setCurrPtr(const char *);
	void normalise(bool logNorm);
	void equilise();
	double normaliser(int nucleotide_type = -1,int threshold= 0);
	void resetData();
	PTR_TYPE currPtr(){return m_currPtr;};
};

class breakDataSet : public	vector<breakData> , public dataFile
{
	vector<string> m_seqs;
	vector<int> m_chrMap;

protected:
	breakDataInfo m_bdi;
	string m_combinedChrList[2];

public:
	int chrInd(int i) {return m_chrMap[i];};
	breakData & chr(int i){return at(m_chrMap[i]);};
	breakDataSet(int offset = 0,int length = 0);
	bool isSetup();
	void resize(size_t size);
	void output(const string & filename,int Nchrs = -1,bool full = true);
	bool open (const string & filename,bool logNorm,int & Nchrs);
	const char * sequence(int i){return m_seqs[i].c_str();};
	int asIndex(const char * s){return m_bdi.asIndex(s);};
	void resetData();

	int seqLength() {return m_bdi.length();};
	int seqOffset() {return m_bdi.offset();};
};


#endif