#define _CRT_RAND_S
#include "pch.h"
#include <sys/stat.h>
#ifdef _WIN32
#include <io.h>
#define chmod _chmod
#endif

#include "breakListOptimiser.h"

#define OUTPUT_EVERY 100

//static float logSmall = log((float)0.5);

//#define COMP_LOG

ErrorPair breakListOptimiser::ErrorFunc()
{
#ifdef _DEBUG
	vector<DOUBLE_TYPE> errors(m_break_data.size());
#endif
	ErrorPair error;
	DOUBLE_TYPE outstanding = m_outstanding;

	for (size_t i = 0;i < m_break_data.size();i++)
	{
		{
			m_z[i] = (m_break_data[i].value + outstanding) * m_gene->m_model[m_offset+i][0];
			outstanding = outstanding + m_break_data[i].value - m_z[i];
		}

		DOUBLE_TYPE err;
#ifdef COMP_LOG
		if (m_z[i] > 0)
			err = (log(m_z[i]) - m_target[i]);
		else
			err = (logSmall - m_target[i]);
#else
		err = m_z[i] - m_target[i];
#endif
#ifdef _DEBUG
		errors[i] = err;
#endif

		err *= err;

		if (m_gene->m_readData[0].reads[m_offset+i] > m_threshold)
		{
			DOUBLE_TYPE mult = m_threshold/(m_threshold + m_gene->m_readData[0].reads[m_offset+i]);
			err *= (mult * mult);
		}

		error.NoWeightings += err;

		if (m_break_data[i].value < 0)
			error.WithWeightings += (m_break_data[i].value * m_break_data[i].value * 10);
			
	}

	error.WithWeightings += error.NoWeightings;
	return error;
}

void breakListOptimiser::SaveResults(bool toFile)
{
	static int counter = 0;
	if (counter++ % 1000 == 0)
		printf("\n%i,%4.1f,%4.1f\n",m_totalIterations,m_ErrLast.WithWeightings,m_ErrLowest.WithWeightings);
}

void breakListOptimiser::SetGene(mRNAinfo & gene)
{
	m_gene = &gene;
}

void breakListOptimiser::optimise(size_t offset,size_t size,int iterations,DOUBLE_TYPE threshold,
		DOUBLE_TYPE outstanding)
{
	m_outstanding = outstanding;
	m_threshold = threshold;
	m_offset = offset;

	m_break_data.resize(size);

	m_z.resize(size);
	m_target.resize(size);
	//	So a simplistic allocation

	DOUBLE_TYPE os= m_outstanding;

	float total = 0;
	for (size_t i = 0;i < size;i++)
	{
		total +=  m_gene->m_readData[0].reads[m_offset+i];
	}
	total /= size;

	for (size_t i = 0;i < size;i++)
	{
//		m_break_data[i].value = total;
//		m_break_data[i].optimise = true;

#ifdef COMP_LOG

		if(m_gene->m_readData[0].reads[m_offset+i] > 0)
			m_target[i] = log(m_gene->m_readData[0].reads[m_offset+i]);
		else
			m_target[i] = logSmall;
#else
		m_target[i] = m_gene->m_readData[0].reads[m_offset+i];
#endif

		float reads = m_gene->m_readData[0].reads[m_offset+i];
		if (reads <= 0)
		{
			m_break_data[i].value = 0;
			m_break_data[i].optimise = false;
		}
		else if (m_gene->m_model[m_offset+i][0] == 0)
		{
			m_break_data[i].value = 0;
			m_break_data[i].optimise = false;
		}
		else if (reads <= 1)
		{
			if ((i != (size-1)) && (m_gene->m_readData[0].reads[m_offset+i+1] > 4))
			{
				m_break_data[i].value =reads/m_gene->m_model[m_offset+i][0] - os;
				m_break_data[i].optimise = true;
			}
			else
			{
				m_break_data[i].value = 0;
				m_break_data[i].optimise = false;
			}
		}
		else
		{
			m_break_data[i].value = reads/m_gene->m_model[m_offset+i][0] - os;
			m_break_data[i].optimise = true;
		}
		if (m_break_data[i].value < 0)
			m_break_data[i].value = 0;
		os = os + m_break_data[i].value - reads;

	}
	optiDataType optiData;
	optiData.push_back(&m_break_data);

	COptimise::optimise(optiData,iterations,OUTPUT_EVERY,1);

	for (size_t i = 0;i < size;i++)
	{
		if (m_break_data[i].value < 0)
			m_gene->m_modelInfo[m_offset+i].m_breaks = 0;
		else
			m_gene->m_modelInfo[m_offset+i].m_breaks =  (float)m_break_data[i].value;
	}
}


void breakListOptimiser::readWeightParameters(std::ifstream & f,mystring & buffer)
{
	while (getDataLine(f,buffer))
	{
		const char * p;
		if (buffer.startsWith("carry over,",p))
			m_carryOver = (*p == 'y');
	}
};
