#define _CRT_RAND_S
#include "pch.h"
#include <sys/stat.h>
#ifdef _WIN32
#include <io.h>
#define chmod _chmod
#endif
#include <sstream>
#include "breakOptimiser.h"

#ifdef _DEBUG
#pragma warning(disable:4244)
#endif

#define OUTPUT_INTERVAL_DURING_OPTIMISATION 10

breakOptimiser::breakOptimiser(const char * rootDir,int Nchrs,int scoreChr,int pssmOffset) :
	breakDataSet(),COptimise(rootDir),m_scoreChr(scoreChr),m_pssmOffset(pssmOffset)
{

}
#define MULTIPLIER

ErrorPair breakOptimiser::ErrorFunc()
{
	m_pssmBlocks.ConvertXyzToPssm();
	ErrorPair Score = calculatePssmWeights(m_scoreChr);

	for (size_t block = 0; block < m_pssmBlocks.size(); block++)
	{

		for (size_t pssm = 0; pssm < m_pssmBlocks[block].size(); pssm++)
		{
			//	This bits stops the optimiser coming up with pssms with negative cooefficients
			if((pssm == 0) || (m_pssmBlocks[block][pssm].m_scalers.size() == 0))
			{
				pssmVector pv;
				for (size_t i = 0; i <  m_pssmBlocks[block][pssm].size(); i++)
				{
					pv.ConvFrom(m_pssmBlocks[block][pssm].m_xyzWeights[i]);

#ifdef _DEBUG
					pssmDataType & pd = m_pssmBlocks[block][pssm];
#endif
					for (int j = 0; j < 4; j++)
					{
#ifdef LINEAR_PSSM
						if (pv[j].value < 0)
							Score.WithWeightings -= (pv[j].value * 100);
						if (pv[j].value > 1)
							Score.WithWeightings += ((pv[j].value -1) * 100);
#else
						if (pv[j].value < -100)
							Score.WithWeightings -= 100 * (pv[j].value + 100);
#endif
					}
				}
			}
			//	And this bit stops it producing scalers that are outside the range 0-1.
			for (size_t i = 0; i < m_pssmBlocks[block][pssm].m_scalers.size(); i++)
			{
				if(m_pssmBlocks[block][pssm].m_scalers[i].value > 1)
					Score.WithWeightings += 10*(m_pssmBlocks[block][pssm].m_scalers[i].value-1);
				else if(m_pssmBlocks[block][pssm].m_scalers[i].value < 0)
					Score.WithWeightings -= 10*(m_pssmBlocks[block][pssm].m_scalers[i].value);
			}
		}
	}
	return Score;
}

void breakOptimiser::SaveResults(bool toFile)
{
	cout << "\n" << m_totalIterations << "," << m_ErrLast.WithWeightings << "," << m_ErrLowest.WithWeightings << "\n";
	if (toFile)
	{
		outputWeights(m_ErrLast);
		SaveProgress();
	}

}

#define WEIGHTS_FILENAME "weights.csv"
#define WEIGHTS_OLD_FILENAME "weightsOld.csv"
#define SCORES_FILENAME "results.csv"


bool breakOptimiser::readWeights()
{
	if(!m_pssmBlocks.open(m_rootDir + WEIGHTS_FILENAME,this))
	{
		if(!m_pssmBlocks.open(m_rootDir + WEIGHTS_OLD_FILENAME,this))
			return false;
	}
	return true;
}

void breakOptimiser::outputWeights(const ErrorPair & Score)
{
	std::stringstream rl;
	rl << DEL_COM << " \"" << m_rootDir << WEIGHTS_OLD_FILENAME << "\"";
	system(rl.str().c_str());
	rl.str("");

	rl << MOVE_COM << " \"" << m_rootDir << WEIGHTS_FILENAME << "\"  \"" << m_rootDir << WEIGHTS_OLD_FILENAME << "\"";
	system(rl.str().c_str());

	FILE * f = fopen((m_rootDir + WEIGHTS_FILENAME).c_str(),"w");
	if (f == 0) 
	{
		if (chmod((m_rootDir + WEIGHTS_FILENAME).c_str(),S_IREAD | S_IWRITE) == 0)
		{
			f = fopen((m_rootDir + WEIGHTS_FILENAME).c_str(),"w");
		}
	}
	if (f == 0) 
	{
		cout << "Failed to open " << WEIGHTS_FILENAME << "\n";
		return;
	}

	fprintf(f,"[score]\n");
	fprintf(f,"Score,%e\n",Score.NoWeightings);
	fprintf(f,"With weights,%e\n",Score.WithWeightings);
	fprintf(f,"Total nucs,%lld\n",m_totalNucs);
	fprintf(f,"Total seqs,%lld\n",m_totalSeqs);
	fprintf(f,"Total breaks,%lld\n",m_totalBreaks);
	fprintf(f,"Threshold,%i\n",m_threshold);
	fprintf(f,"Parameters,%i\n",m_pssmBlocks.Nparameters());
	if(!m_combinedChrList[0].empty())
	{
		fprintf(f,"Chr list 1,%s\n",m_combinedChrList[0].c_str());
		fprintf(f,"Chr list 2,%s\n",m_combinedChrList[1].c_str());
	}
	fprintf(f,"Optimise/calculate using,%i\n",m_scoreChr);
	fprintf(f,"log Offset,%f\n",LOG_OFFSET);

	outputWeightParameters(f);

	m_pssmBlocks.output(f,4,"offset","Multiplier","Coeff2","leakage");
	fclose(f);
}

void breakOptimiser::duplicatePssm(int block,int dupPssm)
{
	m_pssmBlocks[block].push_back(m_pssmBlocks[block][dupPssm]);
	m_pssmBlocks[block].m_gains.push_back(m_pssmBlocks[block].m_gains[dupPssm]);
}

#define FILE_SIZE 32768

void breakOptimiser::initialisePssm()
{
	int pssmLen = m_bdi.length();
	m_pssmBlocks.resize(1);
	pssmVector pvt(0.25,true);
	xyzVector pxy;

	m_pssmBlocks[0].resize(1);
	m_pssmBlocks[0][0].clear();
	m_pssmBlocks[0][0].m_optimising = true;

	pxy.ConvFrom(pvt);
	for (int i = 0;i < pssmLen;i++)
	{
		m_pssmBlocks[0][0].push_back(pvt,pxy);
	}

	m_pssmBlocks[0].m_gains.resize(1);
	m_pssmBlocks[0].m_gains[0].optimise = true;
	m_pssmBlocks[0].m_gains[0].value = log10(1.0);

	m_pssmBlocks.m_globalWeights.resize(3);
	m_pssmBlocks.m_globalWeights[0].value = 0;
	m_pssmBlocks.m_globalWeights[0].optimise = false;
	m_pssmBlocks.m_globalWeights[1].value = 1;
	m_pssmBlocks.m_globalWeights[1].optimise = false;
	m_pssmBlocks.m_globalWeights[2].value = 0;
	m_pssmBlocks.m_globalWeights[2].optimise = false;
}

#define N_NUCS 8

int power(int a, int b)
{
     int c=1;
     for (int n=0; n<b; n++) c*=a;
     return c;
}

//	Create 8 files, one for each of the nucleotides, which shows in detail how the numbers are derived
//#define OUTPUT_DETAILS

void breakOptimiser::calculateSignificance(int offset,const char * name)
{
	char filename[200];
	sprintf(filename,"%ssignificance.csv",m_rootDir.c_str());
	FILE * fAll = fopen(filename,"w");
#ifdef  OUTPUT_DETAILS
	sprintf(filename,"%sinfo_pernuc.csv",m_rootDir.c_str());
	FILE * fnuc = fopen(filename,"w");
#endif
	fprintf(fAll,"%s,Variation with model,,Variation within genome\n",name);
	fprintf(fAll,"Nucleotide,Ave std dev,Std dev of std dev,Ave std dev,Std dev of std dev\n");

/*	
		The nucleotide sequence can be represented as XXNYYYYY

		nuc is the index of the nucleotide, from the right hand side, which is why we
		start with nuc = 7, which the the first in conventional sequence numbering, ie left hand nucleotide
	
*/
	for (int nuc = N_NUCS-1; nuc >= 0 ;nuc--)
	{
		dataVec Dkl(16384),							//Array of KullbackLeibler divergence values for all 16384 7-mers
			res2h(16384),resd(16384),resdh(16384);
		vector<dataVec> resAll(4,16384);
		int count = 0;

#ifdef  OUTPUT_DETAILS
		sprintf(filename,"%ssig%i.csv",m_rootDir.c_str(),7-nuc);
		FILE * f = fopen(filename,"w");
		if (f == 0)
		{
			cout << "Failed to open " << filename << "\n";
			exit(PROC_FAILURE);
		}
		fprintf(f,"index,root sequence,instances,breaks,KullbackLeibler,std dev, stddev two halfs");
		for (int i = 1; i <= 4; i++)
			fprintf(f,",seq%i,sequences%i,breaks%i,over/under%i",i,i,i,i);
		fprintf(f,"\n");

#endif

		for (int i = 0;i < power(4,nuc); i++)
		{
			/*	i is used to generate sequences to the right of the nucleotide being examined */
			for (int j = 0;j < power(4,N_NUCS-nuc-1);j++)
			{
			/*	and j to the left.   Then we create the index to the data where the nucleotide
				is an A, ie 0, and generate the sequence */

				int k = j * power(4,nuc+1);
				int base = i + k;
				int index;
				int tot_breaks_in_subset = 0;
				int tot_8mers_in_subset = 0;

				string seqBase = m_bdi.asString(base);

				dataVec seqCounts(4),breaks(4);		//Data for the 4 nucleotide values over the full genome
				dataVec half1(4),half2(4);	//And then for the two halves
				dataVec frag_8mer_marginal_probs(4),frag_8mer_probs(4);

				vector<string> seqs(4);		//And the sequences, so we can output the data, and check it.
#define MUT_INFO
#ifdef MUT_INFO
				dataVec seq_probs(4);
				dataVec prob_seq_half1(4),prob_seq_half2(4);

				dataVec prob_ratios(4);
				for (int x = 0;x < 4;x++)
				{
					index = base + (power(4,nuc)*x);
					seqs[x] = m_bdi.asString(index);
					resAll[x][count] = chr(0)[index].scaledBreaks;
					breaks[x] = chr(0).m_data[index].breaks;			//Pf(s)
					half1[x] = chr(1).m_data[index].breaks;
					half2[x] = chr(2).m_data[index].breaks;

					seqCounts[x]=chr(0).m_data[index].count;			//P(s)

					prob_seq_half1[x]=chr(1).m_data[index].count;
					prob_seq_half1[x]=chr(2).m_data[index].count;

					tot_breaks_in_subset += (int)chr(0)[index].breaks;
					tot_8mers_in_subset += (int)chr(0)[index].count;
				}

				//	We now need to convert both the top and the bottom to a marginal probability distribution
//				DOUBLE_TYPE tot_breaks_in_subset = breaks.total();
//				DOUBLE_TYPE tot_8mers_in_subset = seqCounts.total();

				//	Normalise the bias, and then find the information content
				//	a measure of how different they are
#define MARGINAL_PROBABILITY
				//	Has to be marginal probabilities because Kullback-Leibler divergence assumes that the set of probabilities 
				//	add up to one
#ifdef MARGINAL_PROBABILITY
				seq_probs = seqCounts/tot_8mers_in_subset;	//Vector division!
				frag_8mer_probs = breaks/(double)m_totalBreaks;
				if (tot_breaks_in_subset == 0)
					frag_8mer_marginal_probs = 0.25;
				else
					frag_8mer_marginal_probs = breaks/tot_breaks_in_subset;	//Normalise so that they sum to one, also vector division

#else
				if (tot_breaks_in_subset == 0)
				{
					frag_8mer_probs = 0;
				}
				else
				{
					frag_8mer_probs = breaks/m_totalBreaks;	//Normalise so that they sum to one, also vector division
				}
				seq_probs = seqCounts/m_totalNucs;
#endif

				for (int x = 0;x < 4;x++)
					prob_ratios[x] = frag_8mer_marginal_probs[x]/seq_probs[x];

				DOUBLE_TYPE ave = frag_8mer_marginal_probs.average();
				DOUBLE_TYPE varf = frag_8mer_marginal_probs.std_dev();

				//	Now calculate the KullbackLeibler divergence 
				DOUBLE_TYPE Dkl_temp = 0;
				for (size_t i = 0;i < 4;i++)
				{
					if (frag_8mer_marginal_probs[i] > 0)
						Dkl_temp += frag_8mer_marginal_probs[i] *  log(prob_ratios[i])/Log2;
				}

//	And now various overall scalings
//	Scale based on the overall number of breaks
//				Dkl_temp *= tot/m_totalBreaks;	

//	Scale based on the overall number of sequences
#ifdef MARGINAL_PROBABILITY
				Dkl_temp *= (16384.0 * tot_8mers_in_subset/m_totalNucs);
//				Dkl_temp *= (16384.0 * tot_breaks_in_subset/m_totalBreaks);
#endif
//	Equal weight given to all 
//				Dkl_temp /= (m_totalSeqs/4);		

#else
				/*	For each, of the 4 nucleotides,   */
				for (int x = 0;x < 4;x++)
				{
					index = base + (power(4,nuc)*x);
					seqs[x] = m_bdi.asString(index);
					resAll[x][count] = chr(0)[index].scaledBreaks;
					full[x] = chr(0)[index].scaledBreaks;
					half1[x] = chr(1)[index].scaledBreaks;
					half2[x] = chr(2)[index].scaledBreaks;
					total += (int)chr(0)[index].count;
					totals[x] = (int)chr(0)[index].count;
				}
				DOUBLE_TYPE ave = full.average();
				DOUBLE_TYPE varf = full.std_dev(ave);
				//	Normalise the bias, and then find the information content
				//	a measure of how different they are
				if (ave == 0)
				{
					full = 1;
				}
				else
				{
					full /= (ave);	//Normalise so that they sum to one.
				}
				DOUBLE_TYPE info = full.info_content();


#endif



				DOUBLE_TYPE var2h = 0;
				for (int x = 0;x < 4;x++)
				{
					var2h += sqr(half1[x]-half2[x]);
				}
				var2h = sqrt(var2h/4/3);



#define INFO
#ifdef INFO
				Dkl[count] = Dkl_temp;
#else
				resf[count] = varf;
#endif
				res2h[count] = var2h;

#ifdef  OUTPUT_DETAILS
				fprintf(f,"%i,%s,%i,%i,%e,%f,%f",base,seqBase.c_str(),tot_8mers_in_subset,tot_breaks_in_subset,Dkl_temp,varf,var2h);
				for (size_t x = 0;x < 4; x++)
					fprintf(f,",%s,%i,%i,%f",seqs[x].c_str(),(int)seqCounts[x],(int)breaks[x],prob_ratios[x]);
				fprintf(f,"\n");
#endif
				count++;
			}
		}
#ifdef  OUTPUT_DETAILS
		fclose(f);
#endif

		dataVec perNucAves(4),perNucHalf1Aves(4),perNucHalf2Aves(4);

		for (int x = 0;x < 4; x++)
		{
			perNucAves[x] = resAll[x].average();
		}
#ifdef DISTANCE
		for (int i = 0;i < count;i++)
		{
			double df = 0;
			for (int x = 0;x < 4; x++)
			{
				df += sqr(resAll[x][i]-perNucAves[x]);
			}
			resd[i] = sqrt(df/3);
		}
#else
		dataVec info_per_nuc(4);
		for (int x = 0;x < 4; x++)
		{
			resAll[x] /= perNucAves[x];
			info_per_nuc[x] = resAll[x].info_content()/count;
		}
#ifdef  OUTPUT_DETAILS
		fprintf(fnuc,"%f,%f,%f,%f\n",info_per_nuc[0],info_per_nuc[1],info_per_nuc[2],info_per_nuc[3]);
#endif
#endif

		DOUBLE_TYPE VarSig2 = info_per_nuc.average();

		//	And now find the info content for each of the 16384 sequences
bool logres = false;
#define LOG_AVE
//	Calculate ave and std dev of the log values as this has a log normal distribution
#ifdef LOG_AVE
		Dkl.log();
		logres = true;
		DOUBLE_TYPE geometric_mean = exp(Dkl.average());// * 1E6;
#else
		DOUBLE_TYPE geometric_mean = resf.total();
#endif
		DOUBLE_TYPE mutInfo2h = res2h.total();
		DOUBLE_TYPE mutInfod = resd.total();

		//	And also the standard deviation of the log
#define LOG_SD
#ifdef LOG_SD

		if (!logres)
			Dkl.log();	//Take logs if we have not already done so
		DOUBLE_TYPE av_log = Dkl.average();
		//	The -1.0 removes the typical value for random data
		DOUBLE_TYPE sigf = exp(Dkl.std_dev());// - 0.95;

#else
		DOUBLE_TYPE sigf = resf.std_dev(mutInfoF/count,count);
		sigf = sigf*count/mutInfoF;
//		sigf = sigf/avef/(double)m_totalBreaks;
#endif
		DOUBLE_TYPE sig2h = res2h.std_dev();
		DOUBLE_TYPE sigd = resd.std_dev();


		//	And now find the variance that is due to nucleotide variation
#ifdef INFO
		double varSig = geometric_mean;
#else
		double varSig = sqrt(sqr(avef)-sqr(ave2h));
#endif

		int pos = 8-nuc+offset;
		if (pos <= 0) pos--;

		fprintf(fAll,"%i,%0.9f,%0.9f,%0.9f,%0.9f,%0.9f,%0.9f,%0.9f,%0.9f\n",pos,geometric_mean,sigf,mutInfo2h,sig2h,mutInfod,sigd,varSig,VarSig2);
	}
#ifdef  OUTPUT_DETAILS
	fclose(fnuc);
#endif
	fclose(fAll);
}

void breakOptimiser::outputScores(int Nchrs,bool dataIsLogarithmic)
{
	char strCommand[MED_LINE_LENGTH];
	sprintf(strCommand, "%s \"%sresults*.csv\"\n", DEL_COM,m_rootDir.c_str());
	system(strCommand);

	char filename[200];
	sprintf(filename,"%sresultsAll.csv",m_rootDir.c_str());
	FILE * fAll = fopen(filename,"w");
	fprintf(fAll,"seq,index,PSSM,model,breaks,count,error,abs ratio");

	sprintf(filename,"%smodelBias.csv",m_rootDir.c_str());
	FILE * fSum = fopen(filename,"w");
	fprintf(fSum,"Summary,%i,%i\n",seqLength(),seqOffset());


	for (int j = 0;j <= Nchrs; j++)
	{
		if (Nchrs == 2)
			fprintf(fAll,",%s,Tot %i,C %i",chr(j).chrName.c_str(),j,j);
		else
			fprintf(fAll,",%s",chr(j).chrName.c_str());
	}
	fprintf(fAll,",cumPoiss\n");

	//	There are three output files, two with lots of detail where the distribution is split into
	//	two for the purposes of plottting in Excel (Plots cant hold more than 32000 points)
	//	and a full distribution
	for (size_t count = 0;count < at(0).size()/FILE_SIZE+1;count++)
	{
		sprintf(filename,"%sresults%i.csv",m_rootDir.c_str(),count);

		FILE * f = fopen(filename,"w");
		if (f == 0) 
		{
			if (chmod(filename,S_IREAD | S_IWRITE) == 0)
			{
				f = fopen(filename,"w");
			}
		}
		if (f == 0) 
		{
			cout << "Failed to open " << filename << "\n";
			return;
		}

		fprintf(f,"seq,PSSM,model,breaks,count,ratio,error");
		for (int j = 1;j <= Nchrs; j++)
		{
			fprintf(f,",%s,count%i,ratio%i,error%i",chr(j).chrName.c_str(),j,j,j);
		}
		fprintf(f,"\n");
		for (size_t i = count * FILE_SIZE; i < min<size_t>(at(0).size(),(count+1)*FILE_SIZE);i++)
		{

			//	Are we above the threshold, and are there breaks in both Chromosome 1 and 
			//	the final chromsome

			//Always print the summary data
			fprintf(fSum,"%s,%i,%f,%i\n",m_break_data[i].Seq(),
				i,m_break_data[i][0].m_value,m_break_data[i].m_PSSM);


			if (m_break_data[i].m_include)
			{

				fprintf(fAll,"%s,%i,%i,%f",m_break_data[i].Seq(),
					i,m_break_data[i].m_PSSM,
					m_break_data[i][0].m_value);

				bool printPlotData = ((chr(1)[i].breaks > 0) && (chr(Nchrs)[i].breaks > 0));

				if (printPlotData) 
					fprintf(f,"%s,%i,%f",m_break_data[i].Seq(),
						m_break_data[i].m_PSSM, m_break_data[i][0].m_value);

				countInfo & ci = at(0)[i];

				fprintf(fAll,",%f,%f,%f,%f",ci.breaks,ci.count,
					m_break_data[i][0].m_error,
#ifdef LOG2NORM
					dataIsLogarithmic?exp(ci.scaledBreaks * Log2):ci.scaledBreaks);
#else
					dataIsLogarithmic?exp(ci.scaledBreaks)-1:ci.scaledBreaks);
#endif

				/*				countInfo & ci = at(0)[i];
				fprintf(f,"%s,%i,%f,%i,%i,%f,%f",m_break_data[i].Seq(),
				m_break_data[i][0].m_PSSM,

				(int)ci.breaks,
				ci.count,
				ci.scaledBreaks,
				m_break_data[i][0].m_error);*/


				for (int j = 0;j <= Nchrs; j++)
				{
					countInfo & cn = chr(j)[i];
					//	Only print the data to be plotted if both of the chromosomes 
					//	have none zero counts.
					if (printPlotData) 
						fprintf(f,",%f,%f,%f,%f",cn.breaks,cn.count,
							cn.scaledBreaks, m_break_data[i][j].m_error);
	
					if (Nchrs == 2)
						fprintf(fAll,",%f,%f,%f",cn.scaledBreaks,cn.breaks,cn.count);
					else
						fprintf(fAll,",%f",cn.scaledBreaks);
				}
				if (printPlotData) 
					fprintf(f,"\n");

				double lambda = (double)ci.count/m_totalNucs * m_totalBreaks;
				double x = at(0)[i].breaks;
				double distance = (x-lambda)/sqrt(lambda);

				fprintf(fAll,",%f\n",distance);
			}
		}
		fclose(f);
	}
	fprintf(fAll,"0\n");	//So that we can find the last line
	fclose(fAll);
	fclose(fSum);
}

void breakOptimiser::outputPssms(const char * filename,bool RNA,double scale)
{
	m_pssmBlocks.outputPssms(m_rootDir,filename,scale,RNA,-m_pssmOffset);
}


bool breakOptimiser::open(const string & filename,bool logNorm,int & Nchrs,size_t threshold)
{
	m_totalNucs = 0;
	m_totalSeqs = 0;
	m_totalBreaks = 0;
	m_threshold = threshold;

	if (breakDataSet::open(filename,logNorm,Nchrs))
	{
		m_break_data.resize(at(0).size(),break_data(Nchrs));
		for (size_t i = 0;i < m_break_data.size();i++)
		{
			if (at(0)[i].count >= threshold)
			{
				m_break_data[i].m_include = true;
				m_totalSeqs++;
				m_totalNucs += (int)at(0)[i].count;
				m_totalBreaks += (int)at(0)[i].breaks;
			}
			else
				m_break_data[i].m_include = false;

			m_break_data[i].SetSeq(m_bdi.asString(i).c_str());
		}

		return true;
	}
	return false;
}

double breakOptimiser::scaleForSequence(const char * seq,int & index)
{
	index = m_bdi.asIndex(seq);
	if (index == -1)
		return 0;

	return (at(0).m_data[index].scaledBreaks);
}


ErrorPair breakOptimiser::calculatePssmWeights(int Chromosome,bool all)
{
	m_pssmBlocks.setGains();


	ErrorPair error;
/*	Calculates the weightings for the N PSSMs for the sequence at a location bindSiteptr in the genome
	The iteration variable is because we may end up calling this more than once for the same location 
	in the same iteration, and it stops us recalculating the values uncessarily
	*/
	static DOUBLE_TYPE log4 = log(4.0);
	int PSSM_no = -1;

	for (size_t i = 0; i < at(0).size(); i++)
	{
		if(all ||(m_break_data[i].m_include))
		{
			const char * seq = m_break_data[i].Seq();

			int pssmCount = 0;
			double cumWeight = m_pssmBlocks.calcWeight(seq,m_break_data[i][Chromosome],m_break_data[i].m_PSSM);

#ifdef LINEAR_PSSM
			if (cumWeight > 0)
				cumWeight = log(cumWeight)/Log2;
			else
				cumWeight = DEFAULT_IF_NOBREAKS;
#else
			cumWeight = log(cumWeight)/Log2;
#endif
			double err = (chr(Chromosome)[i].scaledBreaks - cumWeight);

			err = err * err;

			m_break_data[i][Chromosome].m_value = cumWeight;
			m_break_data[i][Chromosome].m_error = err;

			error.NoWeightings += err;
		}
	}
	error.WithWeightings = error.NoWeightings;
	return error;
}


DOUBLE_TYPE breakOptimiser::Optimise(int iterations,bool verbose)
{
	m_iteration = 0;

	optiDataType optiData;

	for (size_t block = 0; block < m_pssmBlocks.size(); block++)
	{
		for (size_t pssm = 0;pssm < m_pssmBlocks[block].size(); pssm++)
		{
			//	There could be a number of unweighted pssms, followed by a pssm to which weights
			///	are applied
			if((m_pssmBlocks[block][pssm].m_scalers.size() == 0) || (pssm == 0))
			{
				if (m_pssmBlocks[block][pssm].m_optimising)
				{
					for (size_t i = 0;i < m_pssmBlocks[block][pssm].size();i++)
					{
						xyzVector & xyz = m_pssmBlocks[block][pssm].m_xyzWeights[i];
						if(xyz.optimising())
							optiData.push_back(&xyz);
					}
				}
			}
			if (m_pssmBlocks[block][pssm].m_scalers.size())
			{
				if (m_pssmBlocks[block][pssm].m_optimising)
					optiData.push_back(&m_pssmBlocks[block][pssm].m_scalers);
			}
		}
		optiData.push_back(&m_pssmBlocks[block].m_gains);

	}
	optiData.push_back(&m_pssmBlocks.m_globalWeights);

	DOUBLE_TYPE minError = optimise(optiData,iterations,
		verbose?OUTPUT_INTERVAL_DURING_OPTIMISATION:(OUTPUT_INTERVAL_DURING_OPTIMISATION*100),(DOUBLE_TYPE)0.001);

	return minError;
}

