#pragma once
#ifndef BREAK_OPTIMISER_H
#define BREAK_OPTIMISER_H
#include "breakData.h"
#include "Optimise.h"
#include "pssmData.h"
using namespace std;

#pragma warning (disable: 4996)  //strdup warning

#define TRIPLE_PSSM



class break_data
{
private:
	char * m_seq;
	vector<chromosome_break_data> m_data;

public:
	int m_PSSM;
	int m_include;
	chromosome_break_data & operator[](size_t i){return m_data[i];};


	break_data(int chromosomes=0): m_seq(0){m_data.resize(chromosomes+1);};
	~break_data(){free(m_seq);};

	void SetSeq(const char * seq){free(m_seq);m_seq = strdup(seq);};
	const char * Seq(){return m_seq;};

	const vector<chromosome_break_data> & Data() const {return m_data;};

};

class weightFileUser
{
public:
	virtual void outputWeightParameters(FILE * f){};
	virtual void readWeightParameters(std::ifstream & f,mystring & buffer)
	{
		while (getDataLine(f,buffer))
		{
			const char * p;
			if (buffer.startsWith("parameter name,",p))
				int x = atoi(p);
		}

	};
};

class breakOptimiser : public breakDataSet , protected COptimise , public weightFileUser
{
protected:

	int m_iteration;
	int m_pssmOffset;

	size_t m_threshold;
	vector<break_data> m_break_data;

	long long m_totalSeqs,m_totalBreaks,m_totalNucs;
	int m_scoreChr;

public:
	pssmDataBlocks m_pssmBlocks;

	breakOptimiser(const char * rootDir,int Nchrs,int scoreChr,int pssmOffset);
	bool open(const string & filename,bool logNorm,int & Nchrs,size_t threshold);

	virtual ErrorPair ErrorFunc();
	virtual void SaveResults(bool toFile);

	ErrorPair calculatePssmWeights(int Chromosome,bool all = false);
	double scaleForSequence(const char * seq,int & index);

	bool readWeights();

	void outputWeights(const ErrorPair & Score);
	void calculateSignificance(int offset = 0,const char * name = 0);

	void initialisePssm();
	void duplicatePssm(int block = 0,int pssm = 0);
	void outputScores(int Nchrs,bool dataIsLogarithmic);
	void outputPssms(const char * filename,bool RNA,double scale);

	DOUBLE_TYPE Optimise(int iterations,bool verbose);

	const vector<break_data> & BreakData(){ return m_break_data;};
};

#endif
