#include "pch.h"
#include "codFile.h"

void filename::create(const string & path, const string & name,const char * suffix)
{
	ostrstream sstr;
	sstr << path.c_str() << name.c_str() << "." << suffix << (char)0;
	assign(sstr.str());
}



void orderedInt::print(FILE * f,const char * label)
{
	fprintf(f,"\n%s,%i",label,size());
	int j = 0;
	for (const_iterator i = begin();
		(i != end()) && (j < 100);i++,j++)
	{
		fprintf(f,",%i",(*i).first);
	}
}

bool codDataItem::overlaps(const codDataItem & b,int extend)
{
	return ((start < b.end-extend) && (end > b.start+extend));
}

void codDataItem::output(FILE * f,const char * chr)
{
	fprintf(f,"%i\t%s\t%i\t%i\t%s\t%c\t%f\t%f\t%f\n",id,chr,start,end,site.c_str(),
		strand,value1
#ifdef FULL_COD		
		,value2,value3
#endif
		);
}

chromosomeData::chromosomeData(const char * to, size_t l,codDataVec * cdv,bool m)
	:mapsTo(to),length(l),mirror(m)
{
	if (cdv)
	{
		size_t expand_count = 0;
		size_t offset = 0;
		for(size_t i = 0;i < cdv ->size();i++)
		{
			codDataItem & cdi = (*cdv)[i];
			shrink_offsets.insert(map<size_t,size_t>::value_type(cdi.start,offset));
			expand_offsets.insert(map<size_t,size_t>::value_type(cdi.start - offset,offset));
			offset += cdi.end - cdi.start;
		}
		shrink_offsets.insert(map<size_t,size_t>::value_type((size_t)length,offset));
		length -= offset;
		expand_offsets.insert(map<size_t,size_t>::value_type((size_t)length,offset));
	}
};

bool chrDataMap::remap(codDataItem & item,char * chr)
{
	iterator i = find(chr);
	if (i == end())
		return false;
	chromosomeData & chr1Data = (*i).second;
	chromosomeData & chr2Data = (*find(chr1Data.mapsTo)).second;
	
	size_t length = item.end - item.start;

	offsetMap::iterator j = chr1Data.shrink_offsets.lower_bound(item.start);
	if (j != chr1Data.shrink_offsets.end())
		item.start -= (*j).second;

	if(chr1Data.mirror)
	{
		item.start = (size_t)(chr1Data.length - item.start);
		j = chr1Data.expand_offsets.lower_bound(item.start);
		if (j != chr1Data.expand_offsets.end())
			item.start += (*j).second;
	}
	else
	{
		strcpy(chr,chr1Data.mapsTo.c_str());
		item.start = (size_t)(item.start * (chr2Data.length/chr1Data.length));

		j = chr2Data.expand_offsets.lower_bound(item.start);
		if (j != chr2Data.expand_offsets.end())
			item.start += (*j).second;
	}
	item.end = item.start + length;
	return true;
}

void chrDataMap::insert(const char * from, const char * to, size_t from_length,bool mirror)
{
	codFile::iterator i = gapFile.find(from);
	codDataVec * cdv = 0;
	if (i != gapFile.end())
		cdv = &(*i).second;
	map<string,chromosomeData>::insert(chrDataMap::value_type(from,chromosomeData(to,from_length,cdv,mirror)));
}



void codDataVec::add(codDataItem & item)
{
	push_back(item);
}

void codDataVec::makeHash()
{
	if (!hash.empty())
	{
		cout << "Asked to make a hash that already existed\n";
		m_loaded = true;
		return;
	}
	for (iterator i = begin(); i != end();i++)
	{
		hash.insert(dataItemMap::value_type((*i).start,&(*i)));
	}
	m_loaded = true;
}

void codDataVec::ClearData()
{
	clear();
	hash.clear();
}

bool codDataVec::mergeItems()
{
	bool itemsRemoved = false;
	for (iterator i = begin(); i != end();i++)
	{
		bool elementRemoved = true;
		while (elementRemoved)
		{
			elementRemoved = false;
			iterator j = i;
			j++;
			for (; j != end();j++)
			{
				if ((*j).overlaps(*i))
				{
					(*i).start = min((*i).start,(*j).start);
					(*i).end = max((*i).end,(*j).end);
					erase(j);
					itemsRemoved = true;
					elementRemoved = true;
					break;
				}
			}
		}
	}
	return itemsRemoved;
}

bool codDataVec::overlaps(const codDataItem & data,int extend)
{
	for (iterator i = begin(); i != end();i++)
	{
		if ((*i).overlaps(data),extend) 
			return true;
	}
	return false;
}


bindVecCollection codFiles::chrData(const char * chr)
{
	bindVecCollection results;
	for (size_t i = 0; i < data.size(); i++)
	{
		map<string,codDataVec>::iterator j = data[i].find(chr);
		if (j != data[i].end())
			results.push_back(&(*j).second);
	}
	return results;

}
codDataVec codDataVec::overlappingItems(const codDataItem & data,int extend)
{
	return overlappingItems(data.start - extend,data.end + extend);
}

codDataVec codDataVec::overlappingItems(size_t start,size_t finish)
{
	codDataVec results;
	pair<dataItemMap::const_iterator, dataItemMap::const_iterator> data2Range = 
		hash.equal_range(start);

	dataItemMap::const_iterator i = data2Range.first;

	if (i == hash.end())
	{
		return results;
	}

	if ((i == hash.end() || (*i).second ->start > finish))
	{
		if (i == hash.begin())
		{
			return results;
		}

		i--;

		if (i == hash.begin())
		{
			return results;
		}

		if((*i).second ->end < start)
		{
			return results;
		}
	}

	while ((i != hash.end() && (*i).second -> start < finish))
	{
		results.push_back(*(*i).second);
		i++;
	}

	return results;
}

codDataVec bindVecCollection::overlappingItems(int start,int finish)
{
	//	Returns the overlapping binding sites from all of the different 
	//	pssms being considered
	codDataVec results;
	for (size_t i = 0; i < size(); i++)
	{
		codDataVec newdata = at(i) -> overlappingItems(start,finish);
		results.insert(results.end(),newdata.begin(),newdata.end());
	}
	return results;

}

codFile::codFile() : m_file(0),m_posOffset(0)
{
}

codFile::~codFile()
{
}


bool codFile::open (const char * filename,bool openAll,int posOffset,chrDataMap * mapinfo)
{	

	idPos = -1;
	chrPos = -1;
	startPos = -1;
	endPos = -1;
	valuePos = -1;
	negPos = -1;
	posPos = -1;
	sitePos = -1;
	strandPos = -1;
	maxParam = 0;
	bool header = false;

	m_posOffset = posOffset;

	matFile = false;

	if (m_file)
		fclose(m_file);

	clear();

	m_file = fopen(filename,"r");
	if (m_file == 0)
	{
		std::cout << "Failed to open input file" << filename << "\n";
		return false;
	}

	char line[10000];

	fgets(line,10000,m_file);

	char * a = line;
	char * b = strchr(a,'\t');
	int i = 0;
	while (*a)
	{
		*b = 0;
		if (strcmp(a,"#rank") == 0)
		{
			header = true;
			idPos = i++;
		}
		else if (strcmp(a,"#seq_id") == 0)
		{
			header = true;
			matFile = true;
			idPos = i++;
		}
		else if ((strcmp(a,"chr") == 0) || (strcmp(a,"chromosome") == 0)|| (strcmp(a,"chrom") == 0))
		{
			header = true;
			maxParam = chrPos = i++;
		}
		else if ((strcmp(a,"start") == 0) || (strcmp(a,"chromStart") == 0))
		{
			header = true;
			maxParam = startPos = i++;
		}
		else if ((strcmp(a,"end") == 0)|| (strcmp(a,"chromEnd") == 0))
		{
			header = true;
			maxParam = endPos = i++;
		}
		else if (strcmp(a,"strand") == 0)
		{
			header = true;
			strandPos = i;
			maxParam = i++;
		}
		else if (matFile &&(strncmp(a,"log10",5) == 0))
		{
			header = true;
			valuePos = i;
			maxParam = i++;
		}
		else if (!matFile &&(strncmp(a,"maxN",5) == 0))
		{
			header = true;
			valuePos = i;
			maxParam = i++;
		}
		else if (matFile &&(strcmp(a,"site") == 0))
		{
			header = true;
			sitePos = i;
			maxParam = i++;
			break;
		}
//		else if (!matFile &&(strcmp(a,"max|FC|") == 0))
		else if (!matFile &&(strcmp(a,"maxPosN") == 0))
		{
			header = true;
			valuePos = i;
			maxParam = i++;
			break;
		}
		else if (!matFile &&(strncmp(a,"pos_read",8) == 0))
		{
			header = true;
			posPos = i++;
		}
		else if (!matFile &&(strncmp(a,"neg_read",8) == 0))
		{
			header = true;
			negPos = i++;
		}
		else if (strncmp(a,"chr",3) == 0)
		{
			if (startPos != -1)
			{
				matFile = true;
				idPos = startPos;
				startPos = -1;
			}
			chrPos = i++;
		}
		else if (strspn(a,"0123456789") == strlen(a))
		{
			if (startPos == -1)
				maxParam = startPos = i++;
			else if (endPos == -1)
				maxParam = endPos = i++;
			else if (valuePos == -1)
				maxParam = valuePos = i++;
			else
				i++;
		}
		else if (strspn(a,"FR+-") == strlen(a))
		{
			maxParam = strandPos = i++;
		}
		else if (strspn(a,"+-0123456789.eE") == strlen(a))
		{
			if (valuePos == -1)
			{
				maxParam = valuePos = i++;
				matFile = true;
		}
		}
		else if (strspn(a,"CGATN") == strlen(a))
		{
			maxParam = sitePos = i++;
			break;
		}
		else
			i++;
		a = ++b;
		b += strcspn(a,"\t\n");
	};

	strcpy(m_currChromosomeBeingLoaded,"");
	if (!header)
		rewind(m_file);

	if (openAll)
	{

		if (!loadData("LoadAll",mapinfo))
			return false;

		for (iterator i = begin(); i != end(); i++)
		{
			(*i).second.makeHash();
		}
	}
	return true;

}

bool codFile::loadData (const string & searchChr,chrDataMap * mapinfo)
{
	char line[10000];

	while (!feof(m_file))
	{
		fgets(line,10000,m_file);
		char * a = line;
		char * b = strchr(a,'\t');
		char chr[20];
		codDataItem item;

		if (b == 0) break;

		for (int i = 0;i <= maxParam; i++)
		{
			*b = 0;
			if (i == idPos)
				item.id = atoi(a);
			else if (i == startPos)
				item.start = atoi(a)+ m_posOffset;
			else if (i == endPos)
				item.end = atoi(a) +m_posOffset;
			else if (i == strandPos)
				item.strand = a[0];
			else if (i == chrPos)
				strcpy(chr,a);
			else if (i == valuePos)
			{
				item.value1 = (COD_FLOAT_TYPE)(matFile?atof(a):log(atof(a)));
			}
#ifdef FULL_COD
			else if (i == posPos)
				item.value2 = (COD_FLOAT_TYPE)atoi(a);
			else if (i == negPos)
				item.value2 -= (COD_FLOAT_TYPE)atoi(a);
#endif
			else if (i == sitePos)
				item.site = a;
			a = ++b;
			b += strcspn(a,"\t\n");
		}
		if (mapinfo)
		{
			if(!mapinfo -> remap(item,chr))
			{
				cout << "Error remapping data\n";
				return false;
			}
		}
		if (strcmp(chr,m_currChromosomeBeingLoaded) != 0)
		{
			if (strcmp(m_currChromosomeBeingLoaded,"") != 0)
			{
				iterator i = map<string,codDataVec>::find(m_currChromosomeBeingLoaded);
				(*i).second.SetLoaded();
				//	Dont make the hash at this point, but wait until its needed
			}
			if(searchChr == m_currChromosomeBeingLoaded)
			{
				strcpy(m_currChromosomeBeingLoaded,chr);
				add(item,chr);
				return true;
			}

			strcpy(m_currChromosomeBeingLoaded,chr);
		}

		add(item,chr);
	}

	if (searchChr == "LoadAll")
		return true;

	//	We did not find the chromosome
	cout << "Did not find data for the chromosome " << searchChr << "\n";
	return false;
}

codFile::iterator codFile::getData(const string & chr)
{
	if (!m_currChromosome.empty())
	{
		//	Once we have finished with a Chromosome, delete it.
		iterator i = find(m_currChromosome);
		if (i != end())
		{
			(*i).second.ClearData();

		}
	}
	m_currChromosome = chr;

	iterator i = find(chr);

	if((i == end())  || (!(*i).second.loaded()))
	{
		loadData (chr);
		i = find(chr);	//Structly only needed if not originally found
	}

	if(i != end())
		(*i).second.makeHash();	//Chromosome (e.g. M) may not be in data

	return i;
}



void codFile::add(codDataItem & data,const char * chr)
{
	iterator i = find(chr);
	if (i == end())
	{
		codDataVec vec;
		i = insert(value_type(chr,vec)).first;
	}
	(*i).second.add(data);

}

bool codFile::output (const string & filename,const char * header,outputVals ordered)
{
	FILE * f = fopen(filename.c_str(),"w");
	if (f == 0) return false;

	fprintf(f,"#seq_id\tchromosome\tstart\tend\tsite\tstrand\t%s\n",header);

	vector<int> chrIterators(size());

	for (size_t i = 0; i < size(); i++)
		chrIterators[i] = 0;
	

	if (ordered)
	{
		int j = 0;
		size_t chromsomesWithValidData = size();
		while (chromsomesWithValidData > 0)
		{
			int k = 0;
			for (iterator i = begin(); i != end(); i++,k++)
			{
				while ((chrIterators[k] >= 0) && ((*i).second[chrIterators[k]].id == j))
				{
					(*i).second[chrIterators[k]].output(f,(*i).first.c_str());
					chrIterators[k]++;
					if (chrIterators[k] == (*i).second.size())
					{
						chrIterators[k] = -1;
						chromsomesWithValidData--;
					}
				}
			}
			j++;
		}
	}
	else
	{
		for (iterator i = begin(); i != end(); i++)
		{
			for (size_t j = 0; j < (*i).second.size(); j++)
			{
				(*i).second[j].output(f,(*i).first.c_str());
			}
		}
	}

	return true;
}

size_t codFile::numberOfEntries()
{
	size_t count = 0;
	for (iterator i = begin();i != end(); i++)
	{
		count += (*i).second.size();
	}

	return count;
}

size_t codFile::totalLength()
{
	size_t length = 0;
	for (iterator i = begin();i != end(); i++)
	{
		for (size_t j = 0;j < (*i).second.size(); j++)
		{
			length += ((*i).second[j].end - (*i).second[j].start);
		}
	}
	return length;
}


int codFiles::openFiles(const vector<string> filenames)
{
	for (size_t i = 0; i < filenames.size();i++)
	{
		if (!(*this)[i].open(filenames[i].c_str()))
			return PROC_FAILURE;
	}
	return PROC_SUCCESS;
}

