#ifndef codfile_h
#define codfile_h


#include <vector>
#include <map>
#include <string>
#include "barFile.h"
using namespace std;

#define FULL_COD
#define COD_FLOAT_TYPE float

class filename : public string
{
public:
	void create(const string & path, const string & name,const char * suffix);
};


class codDataItem
{
public:
	int id;
	size_t start;
	size_t end;
	string site;
	COD_FLOAT_TYPE value1;
#ifdef FULL_COD 
	COD_FLOAT_TYPE value2;
	COD_FLOAT_TYPE value3;
#endif
	char strand;

	codDataItem() :id(-1),value1(0),
#ifdef FULL_COD 
value2(0),value3(0),
#endif
start(0),end(0),site(""),strand(' '){};
	bool overlaps(const codDataItem & b,int extend = 0);
	void output(FILE * f,const char * chr);
	size_t midpoint() const {return (start + end)/2;};
	size_t length() const {return end - start;};
};


class dataItemMap : public map<size_t,codDataItem *> {};

class codDataVec: public vector<codDataItem>
{
	bool m_loaded;
public:
	dataItemMap hash;

	codDataVec():m_loaded(false){};
	void add(codDataItem & item);
	void makeHash();
	codDataVec overlappingItems(size_t start,size_t finish);
	codDataVec overlappingItems(const codDataItem & data,int extend = 0);
	bool overlaps(const codDataItem & data,int extend = 0);
	bool mergeItems();
	bool loaded(){return m_loaded;};
	void ClearData();
	void SetLoaded(){m_loaded = true;};

};

class orderedInt : public map<int,int>
{
public:
	void add(const codDataItem & d){insert(value_type(d.id,d.id));};
	void add(const codDataVec::iterator & d){add(*d);};
	void print(FILE * f,const char * label);
};

class chrDataMap;

class codFile : public map<string,codDataVec>
{
	FILE * m_file;
public:
	enum outputVals 
	{
		ORDERED = 1,
		UNORDERED = 0
	};
	codFile();
	~codFile();
	bool open (const char * filename,bool openAll = true,int posOffset = 0,chrDataMap * mapinfo = 0);
	bool open (const string & filename,bool openAll = true){ return open(filename.c_str(),openAll);};
	bool loadData (const string & chr,chrDataMap * mapinfo = 0);
	void add (codDataItem & data,const char * chr);
	bool output (const string & filename,const char * header, outputVals ordered = ORDERED);
	bool opened()
	{
		return (m_file != 0);
	};

	const string & CurrChromosome(){ return m_currChromosome;};


	iterator getData(const string & _Keyval);
	size_t numberOfEntries();
	size_t totalLength();

private:
	int idPos;
	int chrPos;
	int startPos;
	int endPos;
	int valuePos;
	int negPos;
	int posPos;
	int sitePos;
	int strandPos;
	int maxParam;

	int m_posOffset;


	bool matFile;

	string m_currChromosome;

	char m_currChromosomeBeingLoaded[100];

};

class bindVecCollection : public vector<codDataVec *>
{
public:
	codDataVec overlappingItems(int start,int finish);
};

class codFiles 
{
	vector<codFile> data;
public:
	codFiles(size_t size = 0): data(size){};
	size_t size() {return data.size();};
	codFile & operator [](size_t i)
	{	
		if (data.size() <= i)
			data.resize(i);
		return data[i];
	}
	int openFiles(const vector<string> filenames);

	bindVecCollection chrData(const char * chr);
};

/*	This is used to move data to different chromosomes in order to create a control set */
class offsetMap: public map<size_t,size_t> {};

struct chromosomeData
{
	chromosomeData(const char * to, size_t l,codDataVec * cdv,bool m = false);
	string mapsTo;
	double length;
	bool mirror;
	offsetMap shrink_offsets;
	offsetMap expand_offsets;
};

class chrDataMap : public map<string,chromosomeData>
{
public:
	codFile gapFile;
	void insert(const char * from, const char * to, size_t from_length,bool mirror = false);
	bool remap(codDataItem & item,char * chr);
};


#endif
