#include <vector>
#include <map>
#include <string>
#include "barFile.h"
using namespace std;

//	For just processing data in one direction
enum strandMode {
	forward_strand,
	reverse_strand,
	both_strands
};


/*	Classes for holding and processing histogram data
	First a vector of doubles, the core of the histogram  */

class histData : public vector<double>
{
public:
	int m_nHists;
	double m_dataSum;

	int m_dataSets;
	histData(int size);
	void inc(int bin,int amount = 1);
	double addData(const barFile & barData,int reference);
	void average();
	void normalise();
};

/*	And a vector of four of them, so that we can have forward and reverse reads
	for the forward and reverse versions of the PSSM.
	These are then ablt to be combined together, in that forward reads on the PSSM should
	be equivalent to reverse read data on its inverse
*/

class histDataVec : public vector<histData>
{
public:
	histDataVec(int M) : vector<histData>(4,histData(M)){};
	void average();
	void normalise();
//	size_t findCorrelation(size_t maxOffset,int vec1,int vec2);
	void combine(int vec1,int vec2,bool forwardStrand,bool reverseStrand);
};

/*	And then a vector of them making an array so that we can have a separate histogram 
	for each value of some parameter
	*/

class histDataArray : public vector<histDataVec>
{
public:
	//	Creates a data array that holds data for N different distributions, each of which
	//	has M bins.
	histDataArray(int N, int M) : vector<histDataVec>(N,histDataVec(M)){};
	double addData(const barFile & barData,int vecOffset,int pssm_score_group,int reference);
	void normalise();
	size_t combine(int vec1,int vec2,bool forwardStrand,bool reverseStrand,int combineThreshold);
	void output(const string & filename,int vec,int width,const string & cons,
		const char ** titles);
};

class seqData : public vector<string>
{
public:
	int m_Nseqs;

	seqData():m_Nseqs(0){};

	void addSequence(const char * seq,int & count);
	void CombineWithReverseStrands(const seqData & revStrands,size_t offset);
	string consensus();
};

