#include "pch.h"
#include "mRNAinfo.h"
#include "alignFile.h"

#ifdef _WIN32
#define rand_r(a) rand_s(a)
#define R_MAX UINT_MAX
#else
#define R_MAX RAND_MAX
#endif

//#define CALC_BACKGROUND

mRNAinfo::mRNAinfo(const string & name,int size,int chromosome):
		m_name(name),m_chromosome(chromosome),m_size(size)
{
		m_readData[0].reads.resize(size+1);
		m_readData[1].reads.resize(size+1);
};



double mRNAinfo::calculatePSSMWeights(pssmDataBlocks & pssmData,bool justWeights,DOUBLE_TYPE threshold,bool carryOver,int pssmOffset)
{
	static DOUBLE_TYPE log4 = log((DOUBLE_TYPE)4.0);

	double totalErr = 0.0;
	const char * seq = m_sequence.c_str();
	double cumWeight = 0.0;

	/*	The offset is how far to the left of the break the PSSM starts.  This 
		allows the nucleotide to the left of the break to have an influence  */

	double outstanding = 0;


	size_t length = m_model.size();

	for (int j = 0; j < pssmOffset; j++)
	{
		m_modelInfo[j].m_error = 0;
		m_model[j][0] = DEFAULT_IF_NONE;
//		outstanding += m_modelInfo[j].m_breaks;
	}

	//	Step through each of the nucleotides in the gene (minus a few at the end)
	for (size_t j = pssmOffset; j < length-pssmData.m_maxLength; j++)
	{
		int pssmCount = 0;
		double weights[100];

		double cumWeight = 0;

		for (size_t block = 0;block < pssmData.size();block++)
		{


#define MAXWEIGHT
#ifdef MAXWEIGHT
			double maxWeight = -100;
#else
			double maxWeight = 0;
#endif
			for (size_t pssm = 0; pssm < pssmData[block].size(); pssm++)
			{
				pssmDataType & ps(pssmData[block][pssm]);

				size_t pssmSize = ps.size();
				//	Use an ordered match instance associated with the PSSM as it will
				//	have been set to the right length, and does not need to be resized every time
				ps.m_orderedMatchesF.reset();

				weights[pssmCount] = 0;
				for (size_t k = 0;k < pssmSize; k++)
				{
					double wf;
					//	Only need to consider forward direction
					switch (seq[j+k-pssmOffset])
					{
					case 'a':case 'A':
						wf = ps.m_weights[k][0].value; break;
					case 'c':case 'C':
						wf = ps.m_weights[k][1].value; break;
					case 'g':case 'G':
						wf = ps.m_weights[k][2].value; break;
					case 't':case 'T':
						wf = ps.m_weights[k][3].value; break;
					}
					weights[pssmCount] += wf;
				}

#ifdef LINEAR_PSSM
				weights[pssmCount] -= (ps.size() * 0.25);
#else
				weights[pssmCount] += (log4 * ps.size());
				weights[pssmCount] = exp(weights[pssmCount]);
#endif

				weights[pssmCount] *= pssmData.m_absGains[pssmCount];
#ifdef MAXWEIGHT
				if (weights[pssmCount] > maxWeight)
				{
					maxWeight = weights[pssmCount];
					//	Register which of the PSSMs gave the biggest weighting
					//	Do this for the 2nd (typically the 6-13th nt
					//	and the third
					if(block == 1)
						m_modelInfo[j].m_PSSM[0] = pssm;
					if(block == 2)
						m_modelInfo[j].m_PSSM[1] = pssm;
				}
#else
				maxWeight += weights[pssmCount];
#endif

				pssmCount++;
			}
			if (maxWeight < threshold) 
				maxWeight = threshold;
		
//			m_model[j][block] = (float)maxWeight;

			if(block == 2)
				m_model[j][1] = (float)maxWeight;
//			else

			cumWeight += maxWeight;

			if (block == 1)
			{
				cumWeight = pssmData.m_globalWeights[0].value + (pssmData.m_globalWeights[1].value * cumWeight) + 
					(pssmData.m_globalWeights[2].value * cumWeight * cumWeight);

				if (cumWeight < threshold) 
					cumWeight = threshold;

				m_model[j][2] = (float)min<double>(cumWeight,1.0);
				cumWeight = 0;
			}
			else if (block == 3)
				m_model[j][1] = (float)min<double>(cumWeight,1.0);

		}


	}

	//	m_model[j][1]:
	//		3 Blocks, contains the biggest of the third set of PSSMs
	//		4 Blocks, contains the biggest of the third set plus the biggest of the fourth
	//		up to a maximum of 1
	//	m_model[j][2] contains the cumulative weighting for PSSMs 1 and 2, up to a maximum of 1.	

	if (justWeights)
		return  totalErr;


	for (int j = 0; j < pssmOffset; j++)
	{
		size_t pos = j;
		outstanding += m_modelInfo[pos].m_breaks;
	}

#ifdef CALC_BACKGROUND
	float multiplier = 0;
	for (int i = 0;i < length;i++)
	{
		multiplier += m_modelInfo[i].m_breaks;
	}

	multiplier /= length;

	float runningCount = 0;
	int window = 8;
	for (int i = 0;i < window; i++)
	{
		runningCount += m_model[i][2];
	}
	for (int i = 0;i < length-8;i++)
	{
		if (i < (length-window))
			runningCount += m_model[i+window][2];

		int c = window;
		float v = m_model[i][2] * (window);
		for (int j = 1;j < window; j++)
		{
			if((i-j) >= 0)
			{
				c += (window-j);
				v += m_model[i-j][2] * (window-j);
			}
			if((i+j) < length)
			{
				c += (window-j);
				v += m_model[i+j][2] * (window-j);
			}
		}
		float value = v/c * multiplier;
		m_model[i][0] = log(LOG_OFFSET + value)/Log2;

		double err = (m_modelInfo[i+8].m_breaks - value);
		err = err * err;

		m_modelInfo[i].m_error = err;
		totalErr += err;
	}



#else

/*	Start
	m_model[j][1] = first multiplier
	m_model[j][2] = second multiplier

	End
	m_model[j][0] = output
	m_model[j][1] = Number of new breaks added at this point  to outstanding breaks
	m_model[j][2] = second multiplier

*/
	for (size_t j = pssmOffset; j < length; j++)
	{
		size_t pos = j;
//#define SINGLE_MODEL
//#define STRAIGHT_MULT
#ifdef SINGLE_MODEL	
		outstanding += m_modelInfo[pos].m_breaks;
		m_modelInfo[pos].m_outstanding = (float)outstanding;
		double modelReads = (outstanding * m_model[pos][2]);
#else
#ifdef STRAIGHT_MULT
		double modelReads = m_modelInfo[pos].m_breaks * m_model[pos][2];
#else
//***	float breakMult = m_model[pos][2];
		m_model[pos][1] = m_modelInfo[pos].m_breaks * m_model[pos][1];
		outstanding += m_model[pos][1];	
		m_modelInfo[pos].m_outstanding = (float)outstanding;
		double modelReads = (outstanding * m_model[pos][2]);
//***	m_model[pos][2] = breakMult;
#endif
#endif
		if (modelReads>0)
			m_model[pos][0] = max<float>((float)log(LOG_OFFSET + modelReads)/Log2,DEFAULT_IF_NONE);
		else
			m_model[pos][0] = DEFAULT_IF_NONE;

		double err = (m_readData[0].reads[pos] - m_model[pos][0]);
		err = err * err;
//#define EMPHASISE_LARGE
#ifdef EMPHASISE_LARGE
		//	More emphasis on errors associated with bigger reads
		if (m_readData[0].reads[pos] > 0)
			err *= m_readData[0].reads[pos];
		else
			err = 0;
#endif

		if (carryOver)
		{
			outstanding -= modelReads;
	//		outstanding += (1 - pssmData.m_globalWeights[3].value) * m_modelInfo[0].m_breaks;
		}
		else
			outstanding = 0;
		
#ifdef OLD_METHOD
		//	The closer cumWeight gets to m_globalWeights[3], the smaller maxWeight is
		maxWeight *= (1.0 - cumWeight);
		cumWeight += maxWeight;

		//	leakage
		cumWeight *= (pssmData.m_globalWeights[3].value);

#ifdef COMPARE_ABS
		double err = (RNAdata[i].m_readData[0].reads[pos] - maxWeight);
#endif
		if(maxWeight > 0)
			maxWeight = log(LOG_OFFSET + maxWeight * m_readData[0].total)/Log2;
		else
			maxWeight = -1.0;

#ifndef COMPARE_ABS
		double err = (m_readData[0].reads[pos] - maxWeight);
#endif
		err = err * err;
#ifndef COMPARE_ABS
		if (m_readData[0].reads[pos] > 0)
			err *= m_readData[0].reads[pos];
		else
			err = 0;
#endif
		m_model[j] = (float)maxWeight;
#endif

		m_modelInfo[pos].m_error = err;
		totalErr += err;
	}
#endif  //not CALC_BACKGROUND 

	return totalErr;
}

DOUBLE_TYPE mRNAinfo::deriveReadsFromBreaks(vector<DOUBLE_TYPE> & outstanding,
		vector<DOUBLE_TYPE> & modelReads,
		size_t max)
{
	modelReads[0] = m_model[0][0]*m_modelInfo[0].m_breaks;
	outstanding[0] = m_modelInfo[0].m_breaks - modelReads[0];
	for (size_t i = 1;i < max;i++)
	{
		modelReads[i] = (outstanding[i-1] + m_modelInfo[i].m_breaks)* m_model[i][0];
		outstanding[i] = outstanding[i-1] + m_modelInfo[i].m_breaks - modelReads[i];
	}
	return (outstanding[max-1]);
}

mRNAinfo mRNAinfo::complement()
{	
	/* Convert everything to its reverse complement.   In teh forward direction the
	info relating to a break is in the location after the break.  When we map to the
	reverse direction it needs to be shifted by 1.  This means that data at position 0 has
	to be mapped to one off the end of the sequence 
	
	A   C   C   T   T
	1   2   3   4   5
	
	A   A   G   G   T
    X   5   4   3   2
	*/
	mRNAinfo compGene;
	compGene.m_chromosome = m_chromosome;

	size_t len = m_sequence.size();

	if (m_model.size())
	{
		compGene.m_model.resize(len+1);

		for (size_t i = 0;i < len;i++)
		{
			compGene.m_model[i][0] = m_model[len-i][0];
		}
	}

	compGene.m_modelInfo  =m_modelInfo;
	compGene.m_name = m_name;
	compGene.m_offset = m_offset;
	compGene.m_pos = m_pos;
	compGene.m_sequence = m_sequence.complement();
	compGene.m_size = m_size;

	compGene.m_readData[0].reads.resize(len + 1);
	for (size_t i = 0;i < len;i++)
	{
		compGene.m_readData[0].reads[i] = m_readData[1].reads[len-i];
	}
	if (m_modelInfo.size())
	{
		compGene.m_modelInfo.resize(len + 1);
		for (size_t i = 0;i < len;i++)
		{	
			compGene.m_modelInfo[i] = m_modelInfo[len-i];
		}
	}
	compGene.m_readData[0].total = m_readData[1].total;
	compGene.m_readData[0].totalNorm = m_readData[1].totalNorm;
	return compGene;
}

void mRNAinfo::invert()
{	
	size_t len = min<size_t>(m_model.size(),m_modelInfo.size());
	len = min<size_t>(len,m_readData[0].reads.size());


	vector<modelData> convModel(m_model.size());

	for (size_t i = 0;i < len;i++)
	{
		convModel[i] = m_model[len-i-1];
	}
	m_model.swap(convModel);


	vector<modelInfo> convInfo(m_modelInfo.size());
	for (size_t i = 0;i < len;i++)
	{
		convInfo[i] = m_modelInfo[len-i-1];
	}
	m_modelInfo.swap(convInfo);


	for (int i = 0;i < 2;i++)
	{
		vector<float> convReads(m_readData[i].reads.size());
		for (size_t j = 0;j < len;j++)
		{
			convReads[j] = m_readData[i].reads[len-j-1];
		}
		m_readData[i].reads.swap(convReads);
	}
}

//#define SPREAD_OUT
#define TRIANGLE
void mRNAinfo::generateSmoothedBreakInfo(int window,int multiplier,int offset)
{
	/*
	Setup the break data so that each point is the average of the values in the 2* window 
	points on either side of the current point.

	If i = 10 and window = 4, then the following code adds the value for i = 14, and removes
	the value for i = 6, so the window contains the values for the 14 points 7,8,9,10,11,12,13,14
	We then devide the value by (2 * window)
	*/


	initialiseModelArrays();
	float runningCount = 0;
	int len = (int)m_readData[0].reads.size();
	for (int i = 0;i < window; i++)
	{
		runningCount += m_readData[0].reads[i];
	}
	for (int i = len - offset; i < len;i++)
	{
		m_modelInfo[i].m_breaks = 0;
	}

	for (int i = 0;i < len - max<int>(offset,0);i++)
	{
		int shift_pos = i + offset;
		if (shift_pos < (len-window))
			runningCount += m_readData[0].reads[shift_pos+window];

#ifdef SPREAD_OUT
		if (i >= window)
			runningCount -= m_readData[0].reads[i-window];
		m_modelInfo[i].m_breaks = (float)(runningCount/(2*window)) * multiplier;
#elif defined(SHIFT)
		m_modelInfo[i].m_breaks = runningCount * multiplier;
		runningCount = 0;
#elif defined(TRIANGLE)
		int c = window;
		float v = m_readData[0].reads[max<int>(shift_pos,0)] * (window);
		for (int j = 1;j < window; j++)
		{
			if((shift_pos-j) >= 0)
			{
				c += (window-j);
				v += m_readData[0].reads[shift_pos-j] * (window-j);
			}
			if(((shift_pos+j) < len) && (shift_pos+j >= 0))
			{
				c += (window-j);
				v += m_readData[0].reads[shift_pos+j] * (window-j);
			}
		}
		m_modelInfo[i].m_breaks = v/c * multiplier;
#else
		if (i > 0)
			runningCount -= m_readData[0].reads[i-1];
		m_modelInfo[i].m_breaks = (float)(runningCount/(window+1))*multiplier;
#endif

	}
	/*	Also, the read data needs to be mapped to the log2 values to be compatible
		with the rest of the code which processes it as log2 values */
	for (int i = 0;i < len; i++)
	{
		if (m_readData[0].reads[i])
			m_readData[0].reads[i] = log(LOG_OFFSET + m_readData[0].reads[i])/Log2;
		else
			m_readData[0].reads[i] = DEFAULT_IF_NONE;
	}
}


void readInfo::clearCounts()
{
	total = 0;
	totalNorm = 0;
	for (vector<float>::iterator i = reads.begin();i < reads.end();i++){(*i) = 0.0;};
}

void mRNAinfo::clearCounts()
{
	m_readData[0].clearCounts();
	m_readData[1].clearCounts();
	//	The modelData constructor allows all of the entries to be set to zero very neatly
	for (vector<modelData>::iterator i = m_model.begin();i < m_model.end();i++){(*i) = 0.0;};
}

void mRNAinfo::initialiseModelArrays()
{
	m_modelInfo.resize(m_readData[0].reads.size());
	m_model.resize(m_readData[0].reads.size());
}
void mRNAinfo::freeModelArrays()
{
	vector<modelInfo> tmp1;
	m_modelInfo.swap(tmp1);

//	vector<float> tmp2;
//	m_model.swap(tmp2);
}

int geneToChromosome(const string & gene)
{
	if (strncmp(gene.c_str(),"AT",2) == 0)
	{
		size_t i = strcspn(gene.c_str(),"0123456789MC");

		if (gene[i] == 'M')
			return 6;
		else if (gene[i] == 'C')
			return 7;
		else
			return atoi(gene.c_str() + i);
	}
	else
	{
		//	We can't tell teh chromosome from the gene
		const char * i = strchr(gene.c_str(),'.');
		if (i)
			return atoi(i-1);
		else
		{
			i = gene.c_str() + strlen(gene.c_str());
			return(atoi(i-1));
		}
	}
}



bool mRNAinfoSet::loadAll(bool minLoad)
{
	bool fileEnd = false;
	while (!fileEnd)
	{
		push_back(mRNAinfo());
		fileEnd = loadGene(back(),minLoad);
	}

	for (size_t i = 0; i < size();i++)
	{
		string s = at(i).m_name;
		m_map.insert(fastaMap::value_type(s,i));
	}
	//Clears error info such as eof
	f.clear();
	return true;
}

bool mRNAinfoSet::convertAll(const char * filename,convMode mode)
{
	FILE * fout = fopen(filename,"w");
	char geneName[200];
	string seq;

	while (!f.eof())
	{
		seq.clear();
		if (!readEntry(seq))
			return false;
		strcpy(geneName,m_name);
		if (mode == removeSpacesInNames)
		{
			char * c = geneName;
			while(c = strchr(c,' ')) *c = '.';
		}
		fprintf(fout,">%s\n",geneName);
		int j = 0;
		while (j < (int)seq.size())
		{
			fprintf(fout,"%s\n",seq.substr(j,min<int>(50,(int)seq.size()-j)).c_str());
			j+= 50;
		}
	}
	fclose(fout);
	return true;
}


void mRNAinfoSet::clearBreakInfo()
{
	for (iterator i = begin();i != end();i++)
	{
		(*i).m_readData[0].clearCounts();
		(*i).m_readData[1].clearCounts();
	}
}


bool mRNAinfoSet::loadGene(mRNAinfo & fd,bool minLoad)
{
	fd.m_pos = f.tellg();

	if (!readEntry(fd.m_sequence,minLoad))
		return false;
	char * space = strchr(m_name,' ');
	if (space)
		*space = 0;
	fd.m_name = m_name;
	fd.m_chromosome = geneToChromosome(fd.m_name);
	if (fd.m_chromosome > max_chromosome)
		max_chromosome = fd.m_chromosome;
	if (!minLoad)
	{
		//	The sequences need to be one longer in order to deal with the
		//	convention that reverse reads are noted at the position after the break
		fd.m_readData[0].reads.resize(fd.m_sequence.size()+1,0);
		fd.m_readData[1].reads.resize(fd.m_sequence.size()+1,0);
		fd.m_model.resize(fd.m_sequence.size()+1,0.0);
	}
	return (f.eof());
}

mRNAinfo * mRNAinfoSet::findGeneInfo(const string & gene)
{
	fastaMap::iterator i = m_map.find(gene);
	if (i == m_map.end())
		return 0;

	mRNAinfo * gi = &at((*i).second);

	if (gi->m_sequence.size() == 0)
	{
		f.seekg(gi->m_pos);
		readEntry(gi->m_sequence,false);
//		f.clear();	//In case we have just read the last entry
		gi->m_readData[0].reads.resize(gi->m_sequence.size()+1,0);
		gi->m_readData[1].reads.resize(gi->m_sequence.size()+1,0);
		//Dont do this now, as it uses too much memory
//		gi->m_model.resize(gi->m_sequence.size()+1,0.0);

		f.clear();	//We may have reached the end of the file
	}

	return gi;
}

void mRNAinfoSet::printGenesSummary(const char * filename)
{
	FILE * f = fopen(filename,"w");
	if (f == 0)
	{
		cout << "Failed to open " << filename << "\n";
		return;
	}
	fprintf(f,"gene,length,forwardBreaks,reverseBreaks,normalisedFBreaks,normalisedRBreaks\n");
	for (size_t i  = 0; i < size();i++)
	{
		if ((at(i).m_readData[0].total) || (at(i).m_readData[1].total))
		{
			fprintf(f,"%s,%i,%i,%i,%f,%f\n",at(i).m_name.c_str(),
				at(i).m_sequence.size(),
				at(i).m_readData[0].total,
				at(i).m_readData[1].total,
				at(i).m_readData[0].totalNorm,
				at(i).m_readData[1].totalNorm);
		}
	};
}

void mRNAinfoSet::printGenes(const char * filename,int loop,bool normalised,breakItemList * breakList)
{
	FILE * f = fopen(filename,"w");
	if (f == 0)
	{
		cout << "Failed to open " << filename << "\n";
		return;
	}

	size_t maxLen = 0;


	vector<mRNAinfo *> genes;

	if (breakList)
	{
		for (size_t i = 0;i < min<size_t>(120,breakList->size());i++)
		{
			mRNAinfo * j = &at(m_map[(*breakList)[i].pgene ->m_name]);
			genes.push_back(j);
		};
	}
	else
	{
		mRNAinfo * g[] = {&at(m_map["AT1G02930.1"]),&at(m_map["AT5G05730.2"]),
			&at(m_map["AT4G05320.2"]),&at(m_map["AT2G30860.1"])};

		//	Transfer the C++ vector into a STL vector, use the iterator type
		//	constructor, where the first parameter is a iterator that points to
		//	the first element, and the second to the last element
//		genes = vector<mRNAinfo *>(g,g + sizeof(g)/sizeof(mRNAinfo *));
	}

	size_t listsize = genes.size();
	size_t mx = min<size_t>(120 - listsize,topN.size()-listsize);

	for (size_t i = 0;i < listsize;i++)
	{
		fprintf(f,"%s,",genes[i]->m_name.c_str());
		maxLen = max<size_t>(maxLen,genes[i]->m_readData[loop].reads.size());
	}
	for (size_t i  = 0; i < mx;i++)
	{
		fprintf(f,"%s,",at(topN[i]).m_name.c_str());
		maxLen = max<size_t>(maxLen,at(topN[i]).m_readData[0].reads.size());
	};
	fprintf(f,"\n");

	if(normalised)
	{
		for (size_t i = 0;i < listsize;i++)
			fprintf(f,"%f,",genes[i]->m_readData[loop].totalNorm);
		for (size_t j  = 0; j < mx;j++)
		{
			fprintf(f,"%f,",at(topN[j]).m_readData[loop].totalNorm);
		};
	}
	else
	{
		for (size_t i = 0;i < listsize;i++)
			fprintf(f,"%i,",genes[i]->m_readData[loop].total);
		for (size_t j  = 0; j < mx;j++)
		{
			fprintf(f,"%i,",at(topN[j]).m_readData[loop].total);
		};
	}


	fprintf(f,"\n");

	for (size_t i = 0;i < listsize;i++)
		fprintf(f,"%i,",genes[i]->m_sequence.size());
	for (size_t j  = 0; j < mx;j++)
	{
		fprintf(f,"%i,",at(topN[j]).m_sequence.size());
	};
	fprintf(f,"\n");


	for (size_t i = 0; i < maxLen;i++)
	{
		for (size_t j = 0;j < listsize;j++)
		{
			if (i < genes[j]->m_readData[loop].reads.size())
				if (normalised)
				{
					if (i < genes[j]->m_model.size())
						fprintf(f,"%f,",genes[j]->m_model[i][0]);
					else
						fprintf(f,",");
				}
				else
					fprintf(f,"%f,",genes[j]->m_readData[loop].reads[i]);
			else
				fprintf(f,",");
		}
		for (size_t j = 0; j < mx;j++)
		{
			mRNAinfo & gene = at(topN[j]);

			if (i < gene.m_readData[loop].reads.size())
				if (normalised)
				{
					if (i < gene.m_model.size())
						fprintf(f,"%f,",at(topN[j]).m_model[i][0]);
					else
						fprintf(f,"0.0,");
				}

				else
					fprintf(f,"%f,",gene.m_readData[loop].reads[i]);
			else
				fprintf(f,",");
		}
		fprintf(f,"\n");
	};
	fclose(f);
}

void mRNAinfoSet::orderGenes(int loop,int threshold)
{
	breakCountMap breakMap;

	for (size_t i = 0;i < size();i++)
	{
		if (at(i).m_readData[loop].total > threshold)
			breakMap.insert(breakCountMap::value_type(at(i).m_readData[loop].total,i));
	}

	topN.clear();

	for (breakCountMap::reverse_iterator  rIter = breakMap.rbegin();
		rIter != breakMap.rend(); rIter++)
	{
		topN.push_back((*rIter).second);
	}
}

void mRNAinfoSet::saveGenes(const char * filename,bool forward)
{
	FILE * f = fopen(filename,"w");

	for (size_t i = 0;i < topN.size(); i++)
	{
		mRNAinfo & gene = at(topN[i]);
		fprintf(f,">%s %i\n",gene.m_name.c_str(),gene.m_readData[0].total);
		if (forward)
			fprintf(f,"%s\n",gene.m_sequence.c_str());
		else
		{
			string revSeq(gene.m_sequence.size(),' ');
			size_t k = 0;
			for (size_t j = gene.m_sequence.size()-1;j >= 0;j--,k++)
			{
				switch (gene.m_sequence[j])
				{
				case 'A': revSeq[k] = 'T';break;
				case 'C': revSeq[k] = 'G';break;
				case 'G': revSeq[k] = 'C';break;
				case 'T': revSeq[k] = 'A';break;
				default:revSeq[k] = 'X';
				}
			}
			fprintf(f,"%s\n",revSeq.c_str());
		}
	}
	fclose(f);
}


bool mRNAinfoSet::loadBreakdata(const string & filename)
{
	std::ifstream ifs(filename.c_str());

	if (!ifs.is_open())
		return false;

	char buffer[LINE_SIZE];
	while ((ifs.peek() == '>') && (!ifs.eof()))
	{
		ifs.getline(buffer,LINE_SIZE);
		trimR(buffer);
		string gene = buffer+1;
		mRNAinfo * fd = findGeneInfo(gene);

		size_t breakCounter = 0;
		while ((ifs.peek() != '>') && (!ifs.eof()))
		{
			ifs.getline(buffer,LINE_SIZE);
			char * a = buffer;
			char * b = strchr(a,',');
			while(b)
			{
				if (b == a)
				{
					breakCounter++;
				}
				else
				{
					*b  = 0;
					char * c = strchr(a,':');
					if (c) 	*c = 0;
					bool OK_to_write = ((fd) && (breakCounter < fd ->m_readData[0].reads.size()));
					if (OK_to_write)
					{
						int count = atoi(a);
						fd ->m_readData[0].reads[breakCounter] += count;
						fd ->m_readData[0].total += count;

						if (c)
						{
							count = atoi(c+1);
							fd ->m_readData[1].reads[breakCounter] += count;
							fd ->m_readData[1].total += count;
						}
						breakCounter++;
					}
				}
				a = b+1;
				b = strchr(a,',');
			}
		}
	}
	ifs.close();
	return true;
}

bool mRNAinfoSet::loadBreakdata(const vector<string> &  alignmentFile,bool random)
{
	for (size_t fileNo = 0;fileNo < alignmentFile.size(); fileNo++)
	{
		string processedAlignmentFile[2];
		alignFile alignments;
		if(!alignmentFile[fileNo].empty())
		{

			size_t x = alignmentFile[fileNo].find_last_of(".");

			if (alignmentFile[fileNo].substr(x,7) == ".alnmap")
			{
				processedAlignmentFile[fileNo] = alignmentFile[fileNo];
			}
			else
			{
				if (random)
					processedAlignmentFile[fileNo] = alignmentFile[fileNo].substr(0,x) + "_random.alnmap";
				else
					processedAlignmentFile[fileNo] = alignmentFile[fileNo] + ".alnmap";
			}

			if (!loadBreakdata(processedAlignmentFile[fileNo])) 
			{
				if (fileNo == 1)
				{
					clearBreakInfo();
				}
				if(!alignments.open(alignmentFile[fileNo].c_str()))
				{
					cout << "Failed to open alignment file\n";
					return false;
				}

				alignment data;

				mRNAinfo * fd = 0; //findGeneInfo(data.gene);
				int errCount = 0;
				while(alignments.read(data))
				{
					fd = findGeneInfo(data.gene);
					if (fd)
					{
						size_t geneLen = fd->m_sequence.size();
						if (data.forward)
						{
#ifdef _DEBUG
							const char * seq = fd ->m_sequence.c_str() + data.location;
#endif
							if (random)
							{
								unsigned int r;
								rand_r(&r);
								if (geneLen > 300)
									data.location = (size_t)((double)r/R_MAX) * (geneLen-100);
								else
									data.location = (size_t)((double)r/R_MAX) * geneLen;
							}

							fd ->m_readData[0].reads[data.location]++;
							fd ->m_readData[0].total++;
						}
						else
						{
#ifdef _DEBUG
							const char * seq = fd ->m_sequence.c_str() + data.location;
#endif
							//	In Emmas data the genomic location is at the start of the sequence that was 
							//	sequenced, whereas the fragment length is at the end, so need to add the length 
							//	of the sequence.
							size_t pos;
							if (random)
							{
								unsigned int r;
								rand_r(&r);
								if (geneLen > 300)
									pos = (size_t)(((double)r/R_MAX) * (geneLen-100)) + 100;
								else
									pos = (size_t)((double)r/R_MAX) * geneLen;
							}
							else
								pos = data.location + data.sequence.size();
							if (pos >= fd ->m_readData[1].reads.size())
								cout << "Error " << data.gene << " " << pos << "\n";
							else
								fd ->m_readData[1].reads[pos]++;
							fd ->m_readData[1].total++;
						}
					}
					else
						cout << "Unable to find gene data for " << data.gene << "\n";
				}
				//	Now write the processed alignment file
				saveBreakdata(processedAlignmentFile[fileNo]);
				if (fileNo == 1)
				{
					//	Reload the first set of data
					loadBreakdata(processedAlignmentFile[0]);
				}
			}
		}
	}


	return true;
}



bool mRNAinfoSet::saveBreakdata(const string & filename)
{
	FILE * f = fopen(filename.c_str(),"w");
	if(f == 0)
	{
		cout << "Failed to open " << filename << "\n";
		return false;
	}
	for (size_t i = 0;i < size();i++)
	{
		mRNAinfo & gene = at(i);
		fprintf(f,">%s",gene.m_name.c_str());
		if(gene.m_readData[0].total || gene.m_readData[1].total)
		{
			//	Don't output data for non-expressed genes
			for (size_t j = 0;j < gene.m_readData[0].reads.size();j++)
			{
				if ((j % 30) == 0)
					fprintf(f,"\n");
				if ((gene.m_readData[0].reads[j] == 0) && (gene.m_readData[1].reads[j] == 0))
					fprintf(f,",");
				else
					fprintf(f,"%i:%i,",(int)gene.m_readData[0].reads[j],(int)gene.m_readData[1].reads[j]);
			}
		}
		fprintf(f,"\n");
	}
	fclose(f);
	return true;
}
