#ifndef MRNA_INFO_H
#define MRNA_INFO_H

#include "fastaFile.h"
#include "genomeFile.h"
#include "pssmData.h"


typedef multimap<int,size_t> breakCountMap;

struct readInfo
{
	readInfo(int size = 0):total(0),totalNorm(0),reads(size){};
	void clearCounts();
	vector<float> reads;
	int total;
	double totalNorm;
};

struct modelInfo
{
	modelInfo(): m_error(0),m_breaks(0),m_outstanding(0)
		{m_PSSM[0] = 0;m_PSSM[1] = 0;};
	float m_breaks;
	float m_outstanding;
	double m_error;
	size_t m_PSSM[2];
};

struct modelData
{
	modelData(float v  = 0.0){m_data[0] = v,m_data[1] = v,m_data[2] = v;};
	float m_data[3];
	float & operator [](int i) {return m_data[i];};

};

class mRNAinfo
{
public:
	mRNAinfo(const string & name,int size,int chromosome);
	mRNAinfo():m_size(0),m_chromosome(-1){};

	double calculatePSSMWeights(pssmDataBlocks & pssmData,bool justWeights = false,DOUBLE_TYPE threshold = 0,bool carryOver = true,int pssmOffset = 1);
	DOUBLE_TYPE deriveReadsFromBreaks(vector<DOUBLE_TYPE> & outstanding,
		vector<DOUBLE_TYPE> & modelReads,
		size_t max);

	void clearCounts();
	void initialiseModelArrays();
	void freeModelArrays();

	mRNAinfo complement();
	void invert();
	void generateSmoothedBreakInfo(int window,int multiplier,int offset);

	string m_name;
	sequence m_sequence;
	int m_chromosome;
	int m_size;
	int m_offset;
	ifstream::pos_type m_pos;
	readInfo m_readData[2];
	vector<modelData> m_model;
	vector<modelInfo> m_modelInfo;
};

int geneToChromosome(const string & gene);


struct breakItem
{
	breakItem(mRNAinfo * g,short int p,float c):pgene(g),position(p),count(c){};

	mRNAinfo * pgene;
	short int position;
	float count;
};
typedef vector<breakItem> breakItemList;


class mRNAinfoSet : public vector<mRNAinfo> , public fastaFile
{
private:
	int max_chromosome;
	fastaMap m_map;

public:
	enum convMode
	{
		removeSpacesInNames
	};
	vector<size_t> topN;

	mRNAinfoSet():max_chromosome(0){};
	bool loadAll(bool minLoad = true);
	bool loadGene(mRNAinfo & fd,bool minLoad);
	bool convertAll(const char * filename,convMode mode);

	void clearBreakInfo();

	int maxChromosome(){ return max_chromosome;};

	void printGenes(const char * filename,int loop, bool normalised,breakItemList * breakList = 0);
	void printGenesSummary(const char * filename);

	mRNAinfo * findGeneInfo(const string & gene);

	void orderGenes(int loop,int threshold = 0);

	bool loadBreakdata(const string &  filename);
	bool loadBreakdata(const vector<string> &  filenames,bool random=false);

	bool saveBreakdata(const string &  filename);

	void saveGenes(const char * filename,bool forward = true);

};

#endif
