#include "pch.h"
#include "mirData.h"

//struct nameList : private map<string,int>

void nameList::add(const char * name,bool combineFamilies)
{
	char * n = _strdup(name);

	if (combineFamilies)
	{
		if ((n[strlen(n)-1] >= 'a') && (n[strlen(n)-1] <= 'z'))
			n[strlen(n)-1] = 0;
	}
	//	Not a multimap, so identical items overwrite
	insert(nameList::value_type(n,1));
	free(n);
}

nameList::operator string()
{
	string list = "";
	bool first = true;
	nameList::iterator k = begin();
	while (k != end())
	{
		if (first)
			first = false;
		else
			list += ",";
		list += (*k).first;
		k++;
	}
	return list;
}



miRNA::miRNA(const char * species,const char * name)
{
	m_names.insert(map<string,string>::value_type(species,name));
}

matchInfo::matchInfo(const char * name,const char * before,const char * after
					 ,const char * sequence)
{
		m_name = _strdup(name);
		m_before = _strdup(before);
		m_after = _strdup(after);
		m_sequence = _strdup(sequence);
};

matchInfo::matchInfo(const matchInfo & m)
{
	m_name = _strdup(m.m_name);
	m_before = _strdup(m.m_before);
	m_after = _strdup(m.m_after);
	m_sequence = _strdup(m.m_sequence);
}

miRNADataMap::iterator miRNADataMap::insert(const char * species,const char * name,char * seq)
{

	miRNADataMap::iterator i = find(seq);
	if (i == end())
	{
		i = map<string,miRNA>::insert(miRNADataMap::value_type(seq,miRNA(species,name))).first;

		unsigned long index = 0;
		char beforeBuffer[MAX_LINE];
		char * beforeBufferIndex = beforeBuffer+MAX_LINE-1;
		*beforeBufferIndex = 0;

		for (size_t j = 0;j < strlen(seq);j++)
		{
			index *= 4;
			index &= 0x3ff;
			switch(seq[j])
			{
			case 'A': break;
			case 'C': index += 1; break;
			case 'G': index += 2; break;
			case 'U': case 'T': index += 3; break;
			}
			if (j >= 4)
			{
				m_matches.insert(index,name,beforeBufferIndex,seq+j+1,seq);
				beforeBufferIndex--;
				*beforeBufferIndex = seq[j-4];
			}
		}
	}
	else
		(*i).second.m_names.insert(multimap<string,string>::value_type(species,name));

	return i;
}

void matchInfoMap::insert(unsigned long index,const char * name,const char * before,const char * after
		,const char * sequence)
{
	multimap<unsigned long,matchInfo>::insert(matchInfoMap::value_type(index,matchInfo(name,before,after,sequence)));
}

mirData::mirData(void)
{
}

mirData::~mirData(void)
{
}

int mirData::read(const char * species)
{
	int found = 0;
	while (readEntry())
	{
		char * mirName = strchr(m_name,'-');
		*mirName++ = 0;

		char * mirNameEnd = strchr(mirName,' ');
		*(mirNameEnd) = 0;

		char * u = strchr(m_sequence,'U');
		while (u)
		{
			*u = 'T';
			u = strchr(u,'U');
		}

		if ((species[0] == 0) || (strcmp(m_name,species) == 0))
		{
			m_allData.insert(m_name,mirName,m_sequence);
			found++;
		}
	}
	cout << species << " database read. " << found << " entries created\n";

	return found;
}


void miRNADataMap::findMatches(Seq_items & seqItems,const options & opts)
{
	Seq_items::iterator i = seqItems.begin();
	int count = 0;
	int matches = 0;
	while (i != seqItems.end())
	{
		if ((count++ % 10000) == 0)
			cout << count << " : " << matches << "\n";
		const char * seq = (*i).first.c_str();
		unsigned long index = 0;
		bool found = false;

		const matchInfo * bestMatch = 0;
		int bestMatchMismatches = 9999;

		size_t searchLen;
		switch (opts.m_mismatches)
		{
		case 0: searchLen = 6; break;
		case 1: searchLen = 12; break;
		default: searchLen = strlen(seq); break;
		}

		for (size_t j = 0;(j < searchLen) && !found;j++)
		{
			index *= 4;
			index &= 0x3ff;
			switch(seq[j])
			{
			case 'A': break;
			case 'C': index += 1; break;
			case 'G': index += 2; break;
			case 'U': case 'T': index += 3; break;
			}
			if (j >= 4)
			{
				matchInfoMap::iterator mii = m_matches.find(index);
				while ((mii != m_matches.end()) && ((*mii).first == index) && !found)
				{
					int m = opts.m_mismatches;

					const matchInfo * mi = &(*mii).second;

					const char * before = mi -> m_before;
					size_t searchLen = j-4;
					while (searchLen > strlen(before))
						{searchLen--;m--;};

					size_t k;
					for (k = 0;(k < searchLen) && (m >= 0) ;k++)
					{
						if (before[k] != seq[j-5-k])
							m--;
					}
					if (m >= 0)
					{
						const char * after = mi -> m_after;
						size_t searchLen = strlen(seq) - j - 1;

						//	Allow one spurious trailing character
//						if (searchLen > strlen(after))
//							searchLen--;

						while ((searchLen > strlen(after)) && (m >= 0))
							{searchLen--;m--;};

						for (size_t k = 0;(k < searchLen) && (m >= 0) ;k++)
						{
							if (after[k] != seq[j+k+1])
								m--;
						}
					}
					if (m >= 0)
					{
						if ((opts.m_mismatches - m) < bestMatchMismatches)
						{
							bestMatchMismatches = opts.m_mismatches - m;
							bestMatch = mi;
							if (bestMatchMismatches == 0)
								found = true;
						}
/*
						//	We have a match
						(*i).second.m_mirMatches.push_back(mi.m_name);

						//	Find the master data, indexed by the sequence
						miRNADataMap::iterator sdi = find(mi.m_sequence);

						//	There may be a number of different sequences mapped to this
						//	because of mismatches
						foundSeqs & fs = (*sdi).second.m_foundSeqs;
						foundSeqs::iterator fsi = fs.find(seq);

						if (fsi == fs.end())
							fs.insert(foundSeqs::value_type(seq,foundSeq((*i).second.n_pos,opts.m_mismatches-m,seq)));
						else
							(*fsi).second.m_number += (*i).second.n_pos;

						found = true;*/
					}
					mii++;
				}

			}
		}
		if (bestMatch)
		{
			matches++;

			(*i).second.m_mirMatches.push_back(bestMatch -> m_name);

			//	Find the master data, indexed by the sequence
			miRNADataMap::iterator sdi = find(bestMatch -> m_sequence);

			//	There may be a number of different sequences mapped to this
			//	because of mismatches
			foundSeqs & fs = (*sdi).second.m_foundSeqs;
			foundSeqs::iterator fsi = fs.find(seq);

			if (fsi == fs.end())
				fs.insert(foundSeqs::value_type(seq,foundSeq((*i).second.n_pos,bestMatchMismatches,seq)));
			else
				(*fsi).second.m_number += (*i).second.n_pos;
		}
		i++;
	}
}


typedef multimap<string,foundSeq*> orderedByNameType;


void miRNADataMap::print (FILE * f,const options & opts)
{
	orderedByNameType orderedByName;

	for (iterator i = begin(); i != end(); i++)
	{
		if ((*i).second.m_foundSeqs.size())
		{
			//	The new list of sequences, ordered and grouped as appropriate
			foundSeqs orderedList;

			nameMap & names = (*i).second.m_names;

			nameList combinedList;
			for (nameMap::iterator j = names.begin();j != names.end();j++)
				combinedList.add((*j).second.c_str(),opts.m_combineFamilies);

			for (foundSeqs::iterator k = (*i).second.m_foundSeqs.begin();
				k != (*i).second.m_foundSeqs.end();k++)
			{
				orderedByName.insert(orderedByNameType::value_type(combinedList,&(*k).second));
//				orderedList.insert(multimap<int,foundSeq*>::value_type((*j).second.m_number,&(*j).second));
			}
		}
	}

	multimap<int,foundSeq*> orderedByQuanity;
	string currentName;


	orderedByNameType::iterator j = orderedByName.begin();
	if (j != orderedByName.end())
	{
		currentName = (*j).first;

		do 
		{
			orderedByQuanity.insert(multimap<int,foundSeq*>::value_type((*j).second->m_number,(*j).second));
			j++;
			if ((j == orderedByName.end()) || 
				((currentName != (*j).first) && (orderedByQuanity.size() > 0)))
			{
				int count = 1;
				multimap<int,foundSeq*>::iterator l = orderedByQuanity.end();
				do
				{
					l--;
					fprintf(f,">%s-%s_%i_%i_%ix\n",opts.m_title.c_str(),currentName.c_str(),count++,(*l).second -> m_mismatches,(*l).first);
					fprintf(f,"%s\n",(*l).second ->m_seq.c_str());
				}
				while (l != orderedByQuanity.begin());

				if (j != orderedByName.end())
					currentName = (*j).first;
				orderedByQuanity.clear();
			}
		}
		while (j != orderedByName.end());
	}
}


void mirData::findMatches(Seq_items & seqItems,const options & opts)
{
	m_allData.findMatches(seqItems,opts);
}


void mirData::print (FILE * f,const options & opts)
{
	m_allData.print(f,opts);
	fclose(f);
}

