
#ifndef PSSMDATA_DEFINES
#define PSSMDATA_DEFINES

//#define SIMPLE_MAPPING
//#define LINEAR_PSSM
#include "Optimise.h"
#include "stdarg.h"

#define LOG_OFFSET (float)8.0
//#define LOG_OFFSET (float)0.0
#define DEFAULT_IF_NONE log(LOG_OFFSET)/Log2

#define PSSM_FONT_SIZE 13

struct chromosome_break_data
{
	double m_value;
	double m_error;
//	int m_PSSM;
};

inline void printItem(FILE * f,const optiItem & v,const char * text = "") {
	fprintf(f,"%e,%c,%s\n",v.value,v.optimise?'Y':'N',text);};

inline bool getDataLine(std::ifstream & f, mystring & buffer) {
	//	Gets a line of data and return true only if appears to be a line of 
	//	data within a section, as opposed to a section start that begins with a [
	return (getline(f,buffer)) && (buffer[0] != '[') && (!f.eof()); };


class orderedList 
{
	DOUBLE_TYPE * m_data;
#ifdef _DEBUG
#define ORDERED_LIST_DEBUG
	//	This is a copy of the data which is visible during debug, unlike the elements
	//	in the m_data array;
	vector<DOUBLE_TYPE> m_debugData;
#endif
	size_t m_length;
	size_t m_allocated;
public:
	orderedList(size_t size = 0); 
	orderedList(const orderedList & ol); 
	~orderedList(){free (m_data);};
	void resize(size_t size);
	void reset(){m_length = 0;};
	void insert(DOUBLE_TYPE v);
	DOUBLE_TYPE operator[](size_t i){return m_data[i];};

};


struct xyzVector;

struct pssmVector: public optiVector
{
	pssmVector(DOUBLE_TYPE val = 0,bool opti = true):optiVector(4,optiItem(val,opti))
	{
#ifndef LINEAR_PSSM
		for (size_t i = 0;i < size();i++) setValue(i,val);
#endif
	};
	void ConvFrom(const xyzVector & xyz,DOUBLE_TYPE scaler = 1);
	DOUBLE_TYPE maxValue() const;
	DOUBLE_TYPE absValue(size_t i) const;
	DOUBLE_TYPE posValue(size_t i) const;
	void setValue(size_t i,DOUBLE_TYPE value);
	DOUBLE_TYPE infoContent() const;
	void makeComplement();
	void print(FILE * f);

	pssmVector & operator *= (DOUBLE_TYPE a);
	pssmVector & operator += (const pssmVector &);

};

struct xyzVector : public optiVector
{
	xyzVector(DOUBLE_TYPE value = 0,bool opti = true):optiVector(3,optiItem(value,opti)){};
	void ConvFrom(const pssmVector & pv,DOUBLE_TYPE scale = 1.0);
	void Scale();
	double MaxPossAmplify();
	xyzVector & operator *= (DOUBLE_TYPE a);
	xyzVector & operator += (const xyzVector & a);
};

struct pssmScaler: public optiItem
{
	void setValue(DOUBLE_TYPE value);
	void setOptimising(bool mode = true);
};
struct pssmScalers: public optiVector
{
	DOUBLE_TYPE value(size_t i);
};

class xyzVectorVector;
class pssmVectorVector: public vector<pssmVector>
{
public:
	pssmVectorVector(size_t size=0,DOUBLE_TYPE value = 0):vector<pssmVector>(size,value){};
	void ConvFrom(const xyzVectorVector & pv,DOUBLE_TYPE scale = 1.0);
	pssmVectorVector & operator += (const pssmVectorVector &);
	pssmVectorVector & operator *= (DOUBLE_TYPE a);
	void print(FILE * f,int offset);

};

class xyzVectorVector: public vector<xyzVector>
{
public:	
	xyzVectorVector(size_t size=0,DOUBLE_TYPE value = 0):vector<xyzVector>(size,value){};
	void ConvFrom(const pssmVectorVector & pv,DOUBLE_TYPE scale = 1.0);
	xyzVectorVector & operator *= (DOUBLE_TYPE a);
	xyzVectorVector & operator += (const xyzVectorVector & a);
	double MaxPossAmplify();
};


class pssmDataType
{
	size_t m_length;
public:
	pssmDataType():m_optimising(false){};
	pssmVectorVector m_weights;
	xyzVectorVector m_xyzWeights;
	pssmScalers m_scalers;
	orderedList m_orderedMatchesF;	//Used during calculation of matches;
	orderedList m_orderedMatchesR;
	bool m_optimising;

	void push_back(const pssmVector & pv,const xyzVector & xyzv);
	void push_back(const pssmScaler & ps);

	size_t size() { return m_length;};
	void resize(size_t size,pssmVector pv = pssmVector(0.25,true));
	void clear();
	void makeComplement();
	void setPositive();
	void operator = (const pssmDataType & pdt);
};

struct pssmDataBlock : vector<pssmDataType>
{
	void resize(size_t i);
	pssmDataBlock(){};
	optiVector m_gains;
};

class weightFileUser;
struct pssmDataBlocks : vector<pssmDataBlock>
{
	int m_firstScalar;
	int m_maxLength;
	vector<DOUBLE_TYPE> m_absGains;		//For holding non-log gains

public:
	optiVector m_globalWeights;

	bool open(const string filename,weightFileUser * wfu);

	void resetWeights();
	void setGains();
	double calcWeight(const char * seq,chromosome_break_data & cbd,int & PSSM);

	void output(FILE * f,int N,...);
	void outputPssms(string & rootDir,const char * filename,double scale = 1,bool RNA = false,int offset = 0);
	size_t Nparameters();

	void initialise();
	void addNoiseToPssms(double scale);
	void duplicate(int block = -1,int pssm = -1);
	void makeComplement();

	void read(std::ifstream &  f,mystring & buffer);

	void ConvertXyzToPssm();
	void reset();
};



#endif