#include <string.h>

//	STL libraries
#include <string>
#include <iostream>
#include <map>

#include "seqInfo.h"

//#define LONG_LINE_LENGTH 65535
//#define MAX_SEQ 5000000
//#define MAX_DEBUG_LINES 1000

using namespace std;

void mirStats::output(FILE * f)
{
	fprintf(f,"Statistics for %s\n",m_title.c_str());
	fprintf (f,"Number of sequences,%u\n",m_sequences);
	fprintf (f,"Contains Ns,%u\n",m_containsNs);
	fprintf (f,"Adaptor tuncated by 1,%u\n",m_adaptorTruncatedByOne);
	fprintf (f,"Adaptor tuncated by 2,%u\n",m_adaptorTruncatedByTwo);
	fprintf (f,"Adaptors removed from real sequence,%u\n",m_adaptorsRemoved);
	fprintf (f,"One mismatch in first 12 of adaptor,%u\n",m_oneMismatch);

	fprintf(f,"\n");
}


void RemoveAdaptor(const options & opts, char * sequence,mirStats & stats)
{
	//	Removes the adaptor sequence from the end of the sequence.
	//	If smartfilter is enabled, a match of 12 is enough to count
	//	as a match, and a single mismatch is ignored.
	//	Smartfilter not currently enabled.
	StrTrimRight(sequence);
	if (strchr(sequence,'N'))
	{
		fprintf(stats.f,"%s\n",sequence);
		stats.m_containsNs++;
		sequence[0] = 0;
		return;
	}
	size_t len = strlen(sequence);
	for (size_t i = 0;i < len;i++)
	{
		bool match = true;
		int mismatch = opts.m_adaptorErrors;

		size_t compareLength = min<size_t>(len-i,opts.m_adaptor.size());
		if ((mismatch > 0) && (compareLength > 12)) compareLength = 12;

		if (compareLength < 3) mismatch = 0;

		for (size_t j = 0;j<compareLength;j++)
		{
			if (sequence[i+j] != opts.m_adaptor[j])
			{
				if (--mismatch < 0)
				{
					match = false;
					break;
				}
			}
		}
		if (match)
		{
			sequence[i]=0;
			stats.m_adaptorsRemoved++;
			if ((opts.m_adaptorErrors - mismatch) == 1)
				stats.m_oneMismatch++;
			return;
		}
		else
		{
			//	Is this a sequence that is just the adaptor starting at nucleotide 2?
			for (size_t offset = 1;offset < 3; offset++)
			{
				int mismatch = opts.m_adaptorErrors;
				size_t compareLength = 12-offset;

				bool match = true;
				for (size_t j = 0;j<compareLength;j++)
				{
					if (sequence[j] != opts.m_adaptor[j+offset])
					{
						if (--mismatch < 0)
						{
							match = false;
							break;
						}
					}
				}
				if (match)
				{
					sequence[0]=0;
					switch (offset)
					{
					case 1:	stats.m_adaptorTruncatedByOne++; break;
					case 2:	stats.m_adaptorTruncatedByTwo++; break;
					}
					return;
				}
			}

		}
	}
}


/* ----------------------------------------------------------------------- */ 
/*                         char* StrTrimRight()                            */
/* This function is used for deleting '\n' ' ' '\t' '\r' on the right of a */
/* string.                                                                 */
/* ----------------------------------------------------------------------- */ 
char* StrTrimRight(char strLongLine[])
{
	char *endp;
	if(strLongLine != NULL)
	{
		endp=strchr(strLongLine, '\0');
		while(endp != strLongLine)
		{
			endp--;
			if((*endp == ' ') || (*endp == '\t') || (*endp == '\n') || (*endp == '\r'))
			{
				*endp = '\0';
				endp=strchr(strLongLine, '\0');
			}
			else
			{
				break;
			}
		}
	}

	return strLongLine;
}

/* ----------------------------------------------------------------------- */ 
/*                         char* StrTrimLeft()                             */
/* This function is used for deleting '\n' ' ' '\t' '\r' on the left of a  */
/* string.                                                                 */
/* ----------------------------------------------------------------------- */ 
char* StrTrimLeft(char strLongLine[])
{
	char *startp;
	long ni;
	size_t nLength;

	if(strLongLine != NULL)
	{
		nLength = strlen(strLongLine);
		if(nLength>0)
		{
			startp=strLongLine;
			
			/* search for start point */
			ni = 0;
			while(*startp != '\0')
			{
				if((*startp == ' ') || (*startp == '\t') || (*startp == '\n') || (*startp == '\r'))
				{
					ni++;
					startp++;
				}
				else
				{
					break;
				}
			}

			/* memmove */
			if(ni == nLength)
			{
				strLongLine[0] = '\0';
			}
			else if(ni != 0)
			{
				memmove(strLongLine, strLongLine+ni, nLength-ni+1);
			}
		}
	}

	return strLongLine;
}

