#pragma once
#include <string>
#include <map>
#include <vector>
using namespace std;

struct options
{
	bool m_ignoreEnds;
	bool m_shortFasta;
	bool m_combineFamilies;
	int m_mismatches;
	string m_species;
	string m_title;
	string m_adaptor;
	string m_outputPathRoot;
	int m_adaptorErrors;
	options():m_ignoreEnds(false),m_shortFasta(false),
		m_combineFamilies(true),m_mismatches(0),m_adaptorErrors(0){};
	void output(FILE * f);
};


char* StrTrimLeft(char strLongLine[]);
char* StrTrimRight(char strLongLine[]);

struct seqInfo
{
	seqInfo():n_pos(0),n_control(0){};
	seqInfo(int p,int c):n_pos(p),n_control(c){};
	seqInfo(string s,int p,int c):m_seq(s),n_pos(p),n_control(c){};
	int n_pos;
	int n_control;
	string m_seq;
	vector<string> m_mirMatches;
};

class Seq_items : public map<string,seqInfo>
{
};
class Ordered_seq_items : public multimap<int,seqInfo>
{
};

struct mirStats
{
	unsigned long m_sequences;
	unsigned long m_adaptorsRemoved;
	unsigned long m_oneMismatch;
	unsigned long m_adaptorTruncatedByOne;
	unsigned long m_adaptorTruncatedByTwo;
	unsigned long m_containsNs;
	string m_pathRoot;
	string m_title;
	FILE * f;

	mirStats(const string & path,const string & title):m_sequences(0),m_adaptorsRemoved(0),
		m_oneMismatch(0),
		m_adaptorTruncatedByOne(0),m_adaptorTruncatedByTwo(0),
		m_containsNs(0),
		m_pathRoot(path),m_title(title),f(0)
	{
		f = fopen((path + "." + title + ".statsInfo.csv").c_str(),"w");};
	void output(FILE * f);
	~mirStats(){fclose(f);};

};

void RemoveAdaptor(const options & opts, char * sequence,mirStats & stats);
