#include "pch.h"
#include <locale>
#include <iostream>
using namespace std;


#include "barFile.h"
#include "codFile.h"
#include "breakData.h"
#include "genomeFile.h"

#define CORRELATION_SEARCH_RANGE 10
#define CHAR_CHUNK_SIZE 50000
#define RAND_INC 25
#define OUTPUT_LIST

/*
A derivative of normalise breaks, with the normalise stuff removed and extended 
analyse stuff included.

the -d file allows the regions which are duplicated to be excluded from the statistics on the 
number of occurances of each sequence.  This is because what we would expect is for the
number of sequences seen to be proportional to the number of occurances which are unique.

If they were all non-unique then we would never see any of the fragments

"Analyse fragments.xls" looks at the variation in binding probaility as a function of each of 
combinations of any two nucleotides.

"Analyse fragments.xls" looks at the variation in binding probaility as a function of each of 
combinations of any two nucleotides.

Analyse fragments2.xls attempts to do a fitting excercise

*/


// #define OUTPUT_LIST	Outputs the list of breaks, positions and surrounding sequence

int menu_region_stats(int argv, char **argc);

int main(int argv, char **argc)
{
	/* ---- */
	/* menu */
	/* ---- */
	menu_region_stats(argv, argc);

	/* exit */
	exit(EXIT_SUCCESS);
}

enum run_mode
{
	forward_only,
	reverse_only,
	all
};

class seqInfo
{
public:
	string chromosome;
	size_t pos;
	float value;
	seqInfo(const char * c,size_t p,float v): chromosome(c),pos(p),value(v){};
	seqInfo(): pos(0),value(0){};
};


int menu_region_stats(int argv, char **argc)
{

	string strBarFile;
	string strControlBarFile;

	string strRegionFile;
	string duplicatesFileName;

	string strGenomeDirectory;
	int length = 8;

#ifdef OUTPUT_LIST
#define INST_STR_LEN 20
	vector<int> instances;
	char instanceBuffL[INST_STR_LEN + 5];
	char instanceBuffR[INST_STR_LEN + 5];
#endif
	int offset = 0;
	
	int nResult;
	int numberOfChromosomes = 0;	//0 = doAll
	bool iOK = false;
	bool cOK = false;
	bool dualbarfile = false;
	bool logNorm = true;
	bool random = false;
	bool wigFile = false;

	run_mode mode = all;
	  
	/* ---------------------------- */
	/*     region_stats             */
	/* -t top n regions             */
	/* ---------------------------- */
	if(argv < 1)
	{
		printf("Error: parameter wrong!\n");
		exit(EXIT_FAILURE);
	}
	else if(argv == 1)
	{
		printf("/* ----------------------------- */\n");
		printf("     normalise breaks              \n");
		printf(" -c control bar file               \n");
		printf(" -i input bar file			       \n");
		printf(" -s cod file for selecting regions \n");
		printf(" -d duplicates file                \n");
		printf(" -g genome data directory          \n");
		printf(" -o offset                         \n");
		printf(" -s suffix                         \n");
		printf(" -l search length                   \n");
		printf(" -f forward only				   \n");
		printf(" -r reverse only				   \n");
		printf(" -R Random locations			   \n");
		printf(" -w make wig file				   \n");
#ifdef OUTPUT_LIST
		printf(" -I <index> output instance info\n");
#endif
		printf(" -n Number of chromosomes to be processed (0 for all)\n");
		printf(" example: \n");
		printf("    normaliseBreaks \n");
		printf("/* ----------------------------- */\n");
		exit(EXIT_SUCCESS);
	}

	int ni = 1;
	while(ni < argv)
	{
		if(strcmp(argc[ni], "-g") == 0)
		{
			ni++;
			strGenomeDirectory = argc[ni];
		}
		else if(strcmp(argc[ni], "-d") == 0)
		{
			ni++;
			duplicatesFileName = argc[ni];
		}
		else if (strcmp(argc[ni], "-R") == 0)
		{
			random = true;
		}
		else if (strcmp(argc[ni], "-w") == 0)
		{
			wigFile = true;
		}
		else if (strcmp(argc[ni], "-a") == 0)
		{
			logNorm = true;
		}
		else if(strcmp(argc[ni], "-i") == 0)
		{
			ni++;
			strBarFile =  argc[ni];
			if (!cOK)
				strControlBarFile = strBarFile;
			iOK = true;
			if (cOK && (strBarFile != strControlBarFile))
				dualbarfile = true;
		}
		else if(strcmp(argc[ni], "-c") == 0)
		{
			ni++;
			strControlBarFile =  argc[ni];
			cOK = true;
			if (!iOK)
				strBarFile= strControlBarFile;
			if (iOK && (strBarFile != strControlBarFile))
				dualbarfile = true;
		}
		else if(strcmp(argc[ni], "-n") == 0)
		{
			ni++;
			numberOfChromosomes =  atoi(argc[ni]);
		}
		else if(strcmp(argc[ni], "-l") == 0)
		{
			length =  atoi(argc[++ni]);
		}
		else if(strcmp(argc[ni], "-o") == 0)
		{
			offset =  atoi(argc[++ni]);
		}
		else if(strcmp(argc[ni], "-s") == 0)
		{
			strRegionFile =  argc[++ni];
		}
		else if(strcmp(argc[ni], "-f") == 0)
		{
			mode = forward_only;
		}
		else if(strcmp(argc[ni], "-r") == 0)
		{
			mode = reverse_only;
		}
#ifdef OUTPUT_LIST
		else if(strcmp(argc[ni], "-I") == 0)
		{
			instances.push_back(atoi(argc[++ni]));
		}
#endif
		else
		{
			cout << "Error: Input Parameter not correct!\n";
			exit(EXIT_FAILURE);
		}
		ni++;
	}

/*	if((bOK == 0) || (nOK == 0) || (dOK == 0) || (oOK == 0))
	{
		printf("Error: Input Parameter not correct!\n");
		exit(EXIT_FAILURE);
	}
	else*/
	{

		codFile genomeRegions;
		if (strRegionFile != "")
			genomeRegions.open(strRegionFile.c_str());

		codFile duplicatesFile;

#ifdef OUTPUT_LIST
		FILE * flist1 = 0, * flist2 = 0, * flist3 = 0;
		string flistName;
		
		typedef vector<seqInfo> vecSeqInfo;

		typedef map<string,vecSeqInfo> mapSeqInfo;
		
		mapSeqInfo si;

#endif

		genomeFile genFile;
		genFile.setRoot(strGenomeDirectory);

		/*	A two stage process.  First we go through one (or more) chormosomes getting the 
			data that we need for the normalisation process
			*/
		int char_offset = 0;	//This is the offset for the sequence to be matched
		int frag_offset = 0;	//This is the offset to the start of the fragment.  Use for mapability

//		int loop = 1;
		//	We perform the normalisation for the combined data, the forward data and the 
		//	reverse data independantly.
		//	Note that in all cases the assumption is that the genomic coordinate is to the right of
		//	the break, ie the start of the forward fragment, or one to the right of the start
		//	of the reverse fragment

		int start = 1;
		int finish = 2;
		switch (mode)
		{
		case forward_only: finish = 1; break;
		case reverse_only: start = 2; break;
		}

		breakDataSet breakStats(offset,length);

		size_t Nchrs;

		int matchesOffset = 0;
		for (int loop = start; loop <= finish;loop++)
		{
			//	0th loop do combined, 1st loop do forward, 2nd loop do revers
			string thisLoopBarFile = strBarFile;
			string thisDuplicatesFileName = duplicatesFileName;


#ifdef OUTPUT_LIST
			string thisFlistFile = strBarFile + "list";
#endif
			switch (loop)
			{
			case 1:
				thisLoopBarFile += "_F.bar";
#ifdef OUTPUT_LIST
				thisFlistFile += "_F";
#endif
				char_offset = offset + length - 1;
				frag_offset = length - 1;
				breakData::setForward();
				break;
			case 2:
				{
					thisLoopBarFile += "_R.bar";

					int len = thisDuplicatesFileName.length();

					if (len > 7)
					matchesOffset = atoi(thisDuplicatesFileName.substr(len-7,3).c_str());

#ifdef OUTPUT_LIST
					thisFlistFile += "_R";
#endif
					char_offset = -offset - 1;
					frag_offset = - 1;

					breakData::setForward(false);
					break;
				}
			default:;
			}

			if (thisDuplicatesFileName != "")
				duplicatesFile.open(thisDuplicatesFileName.c_str(),false,matchesOffset);

			barFile barData;

#ifdef OUTPUT_LIST
			if (instances.size())
			{
				char buffer[20];
				sprintf(buffer,"_%i",numberOfChromosomes);
				thisFlistFile += buffer;

				for (size_t i = 0;i < instances.size();i++)
				{
					sprintf(buffer,"_%i",instances[i]);
					thisFlistFile += buffer;
				}
				flist1  = fopen((thisFlistFile + ".csv").c_str(),"w");
				flist2  = fopen((thisFlistFile + "_full.csv").c_str(),"w");
				flist3  = fopen((thisFlistFile + "_list.csv").c_str(),"w");
				fprintf(flist2,"Chromosome\tPosition\tStrand\tValue\tIndex\tPreSeq\tPostSeq\n");
				fprintf(flist3,"Chromosome\tPosition\tValue\tPostSeq\n");
			}
#endif
			
			if(!barData.open(thisLoopBarFile,true))
			{
				cout << "failed to open barfile\n";
				return PROC_FAILURE;
			}

			vector<string> chrs = barData.GetChrs();

			//	Choosing a width of 4 means that we go between 
			//	-3 and +4, ie locations -3,-2,-1 and 0 are the last 4 
			//	positions of the sequenced fragment and 1,2,3,4 are the 
			//	first 4 of teh following nucleotides

			//	The resize already caters for the additional ref entry
			if (!breakStats.isSetup())
				breakStats.resize(chrs.size());

			chrs.insert(chrs.begin(),"ref");

			//	Use the data in breakStats[0] (normally the first chromosome from the control
			//	data for normalisation
			breakData & bsNorm = breakStats[0];

			chrs = barData.GetChrs();
			Nchrs = (numberOfChromosomes == 0)?chrs.size():numberOfChromosomes;

			//	And now we normalise the data for all of the chromosomes  */

			for (size_t chr = 0;chr < Nchrs; chr++)
			{
				string & currentChromosome = chrs[chr];
				PTR_TYPE * nucleotide_types = 0;

				const codDataVec * dups = 0;
				codDataVec::const_iterator dupDataItem;
				if (duplicatesFile.opened())
				{
					codFile::const_iterator di = duplicatesFile.getData(currentChromosome);
					if (di != duplicatesFile.end())
					{
						dups = &(*di).second;
						dupDataItem = dups -> begin();
					}
				}

				breakData & bs = breakStats[chr + 1];
				bs.chrName = currentChromosome;

				size_t maxBar = barData.getNEntries(currentChromosome)-1;
				chrData data;
				barData.getNext(currentChromosome,data,maxBar);
				data.location += char_offset;
				size_t chrLength = data.location;

				//	We now need to track down the data in the main file and normalise
				size_t gOffset = 1;
				size_t gEnd = gOffset + CHAR_CHUNK_SIZE + offset + length;

				try
				{
					const char * gseq = genFile.getseq(currentChromosome.c_str(),gOffset,gEnd);

					nucleotide_types = new PTR_TYPE[maxBar];

					size_t p = 0;

					barData.getNext(currentChromosome,data,p);

					if (random)
					{
						data.value = 1;
						data.location = RAND_INC;
					}

					//	Initialise the string
					bs.setCurrPtr(gseq); 

					//	Step right the way through the chromosome nucleotide by nucleotide

					for (size_t i = length;
						i <= chrLength /* + breakData::posOffset()*/;	i++)
						//						for (int i = breakData::posOffset() - breakData::negOffset(); i < 20000000;i++)
					{
						size_t actualPos = i - char_offset;
						size_t frag_pos = i - frag_offset;
						bool ignoreAsDup = false;
						size_t index;
						float value = 0;
						if (dups)
						{
							while ((dups) && (frag_pos > (*dupDataItem).end))
								if (++dupDataItem == dups ->end())
									dups = 0;
							if (dups)
								ignoreAsDup = (frag_pos >= (*dupDataItem).start);
						}

							
						if (actualPos >= data.location)
						{
							//	p is pointing to the next barData that will be fetched!
							if (random)
								data.value = (float)(ignoreAsDup?0.0:1.0);

							index = bs.addData(gseq[i - gOffset],data.value,ignoreAsDup);
							nucleotide_types[p-1] = index;
							value = data.value;

							if (random)
								data.location += RAND_INC;
							else
								barData.getNext(currentChromosome,data,p);
						}
						else
							index = bs.addData(gseq[i - gOffset],0,ignoreAsDup);

#ifdef OUTPUT_LIST
						for (size_t j = 0;j < instances.size();j++)
						{
							if (index == instances[j])
							{
								//									if (data.value > 0)	//Output all locations so we can calculate over representations
								{

									strncpy(instanceBuffR,&gseq[i - gOffset - frag_offset],INST_STR_LEN);
									instanceBuffR[INST_STR_LEN] = 0;
									strncpy(instanceBuffL,&gseq[i - gOffset - frag_offset-INST_STR_LEN],INST_STR_LEN);
									instanceBuffL[INST_STR_LEN] = 0;

									locale loc1;// ( "German_Germany" );

									use_facet<ctype<char> > ( loc1 ).toupper(instanceBuffR,instanceBuffR + INST_STR_LEN);
									use_facet<ctype<char> > ( loc1 ).toupper(instanceBuffL,instanceBuffL + INST_STR_LEN);


									for (int k = 0;k < value; k++)
										fprintf(flist1,"%s\t%i\t+\t%i\n",currentChromosome.c_str(),
										actualPos,index);

									fprintf(flist2,"%s\t%i\t+\t%f\t%i\t%s\t%s\n",currentChromosome.c_str(),
										actualPos,value,index,instanceBuffL,instanceBuffR);

									mapSeqInfo::iterator it = si.find(instanceBuffR);
									if (it == si.end())
									{
										it = (si.insert(mapSeqInfo::value_type(instanceBuffR,vecSeqInfo()))).first;
									}
									(*it).second.push_back(seqInfo(currentChromosome.c_str(),actualPos,value));
								}
								break;
							}
						}

#endif


						if (i > (gOffset + CHAR_CHUNK_SIZE))
						{
							gOffset += CHAR_CHUNK_SIZE;
							gEnd = gOffset + CHAR_CHUNK_SIZE +offset + length;
							if (gEnd > chrLength)
								gEnd = chrLength;
							gseq = genFile.getseq(currentChromosome.c_str(),gOffset,gEnd);
						}
					}

					cout << "Finished "<< currentChromosome << "\n";
					delete [] nucleotide_types;
				}
				catch (...)
				{
					delete [] nucleotide_types;
					//	Usually means that there is no genefile for this chromosome
				}
			}
		}

		for (size_t chr = 0;chr < Nchrs; chr++)
		{
			breakData & bs = breakStats[chr + 1];
			if (!(strchr(bs.chrName.c_str(),'Y') || strchr(bs.chrName.c_str(),'X')))	
				breakStats[0] += bs;
			bs.normalise(logNorm);
		}

		breakStats[0].normalise(logNorm);
		breakStats[0].chrName = "ref";

		const char * mode_type;
		switch (mode)
		{
		case forward_only: mode_type = "for"; break;
		case reverse_only: mode_type = "rev"; break;
		case all: mode_type = "all"; break;
		}

		size_t x = strBarFile.find(".bar");
		string outputFile = strBarFile.substr(0,x) + "_ana" + strBarFile.substr(x);

#ifdef OUTPUT_LIST
		//	If we have instances then we dont re-output the amalgamated list
		if (instances.size() == 0)
#endif
		{
			char filename[1000];
			char randSuffix[20];
			sprintf(randSuffix,"_rand_%i",RAND_INC);
			sprintf(filename,"%s_o%i_l%i_c%i_%s%s%s.csv",outputFile.c_str(),
				offset,length,numberOfChromosomes,mode_type,
				random?randSuffix:"",logNorm?"":"_abs");
			breakStats.output(filename);
		}

#ifdef OUTPUT_LIST
		for (mapSeqInfo::iterator it = si.begin();it != si.end();it++)
		{
			float value = 0;
			for (size_t i = 0;i < (*it).second.size(); i++)
				value += (*it).second[i].value;

			if (value > 0)
			{
				for (size_t i = 0;i < (*it).second.size(); i++)
				{
					fprintf(flist3,"%s\t%i\t%f\t%s\n",(*it).second[i].chromosome.c_str(),
						(*it).second[i].pos,(*it).second[i].value,(*it).first.c_str());
				}
			}
		}


	if (flist1)
	{
		fclose(flist1);
		fclose(flist2);
		fclose(flist3);
	}
#endif
		nResult = PROC_SUCCESS;
	}

	return nResult;
}