#include "pch.h"
#include "genomeFile.h"
#include "breakOptimiser.h"
#include "barFile.h"

#define CHAR_CHUNK_SIZE 50000
#define OUTPUT_BAR

int menu_makeDistribution(int argv, char **argc);

int main(int argv, char **argc)
{
	/* ---- */
	/* menu */
	/* ---- */
	menu_makeDistribution(argv, argc);

	/* exit */
	exit(EXIT_SUCCESS);
}
#define WEIGHTS_FILENAME "weights.csv"

int menu_makeDistribution(int argv, char **argc)
{

	string strGenomeDirectory;
	string breakFile;
	bool useBreaksInfo = false;
	int nChromosomes = 0;
	char outputDirectory[LONG_LINE_LENGTH] = "";
	bool cg = false;
	bool cod = false;
	int window = 1;
	int pssmLength = 0;
	FILE * f = 0;
	double threshold = 2.0;

/*	This creates a bar file containing the distribution of some value
	throughout the genome, or some region thereof.

	-cg option does the cg content
	-b shows the values as predicted by the experimentally observed variation with nmer 
	if none of the obove then -o uses the weightings from the weight file in the output directory

*/

	if(argv < 1)
	{
		printf("Error: parameter wrong!\n");
		exit(EXIT_FAILURE);
	}
	else if(argv == 1)
	{
		printf("/* ----------------------------- */\n");
		printf("     outputDistribution		       \n");
		printf(" -g genome data directory          \n");
		printf(" -cg	get cg distribution        \n");
		printf(" -c 	make code file             \n");
		printf(" -b	break file			           \n");
		printf(" -o	output directory               \n");
		printf(" -n	number of chromosomes          \n");
		printf(" -w data averaged over window size (default = 1)  \n");

		exit(EXIT_SUCCESS);
	}

	int ni = 1;
	while(ni < argv)
	{
		if(strcmp(argc[ni], "-o") == 0)
		{
			ni++;
			strcpy(outputDirectory,argc[ni]);
			size_t lastChar = strlen(outputDirectory)-1;
			if (outputDirectory[lastChar] != FILE_SEP)
			{
				outputDirectory[++lastChar] =FILE_SEP;
				outputDirectory[++lastChar] = 0;
			}
			break;
		}
		ni++;
	}

	ni = 1;
	while(ni < argv)
	{
		if(strcmp(argc[ni], "-g") == 0)
		{
			strGenomeDirectory = argc[++ni];
		}
		else if(strcmp(argc[ni], "-b") == 0)
		{
			useBreaksInfo = true;
			breakFile = argc[++ni];
		}
		else if(strcmp(argc[ni], "-w") == 0)
		{
			window = atoi(argc[++ni]);
		}
		else if(strcmp(argc[ni], "-n") == 0)
		{
			nChromosomes = atoi(argc[++ni]);
		}
		else if(strcmp(argc[ni], "-cg") == 0)
		{
			cg = true;
		}
		else if(strcmp(argc[ni], "-c") == 0)
		{
			cod = true;
			threshold = atof(argc[++ni]);
		}

		ni++;
	}


	genomeFile genFile;
	genFile.setRoot(strGenomeDirectory);

	barFile barData;

	vector<string> chrs;

	size_t NChrs;
	if (nChromosomes == 0)
		NChrs = genFile.NChrs();
	else
		NChrs = nChromosomes;

	for (size_t chr = 0;chr < NChrs; chr++)
	{
		chrs.push_back(genFile.getChrName(chr));
	}

	breakOptimiser optimiser(outputDirectory,-1,0,-1);

	if (cg)
	{
	}
	else if (useBreaksInfo)
	{
		int Nchrs = -1;
		optimiser.open(breakFile,false,Nchrs,0);
	}
	else
	{
		if(!optimiser.readWeights())
		{
			cout << "Failed to read weight file";
			return PROC_FAILURE;
		}
		optimiser.m_pssmBlocks.setGains();
		pssmLength = optimiser.m_pssmBlocks[0][0].m_weights.size();
	}

	barData.Initialise(chrs);

//	int window = 20;
#ifdef OUTPUT_SEQS
	int refIndex[] = {40086,26405};
	FILE * f = fopen((string(outputDirectory) + "seqs.csv").c_str(),"w");
#endif
	for (size_t chr = 0;chr < NChrs; chr++)
	{
		
		int p = 0;
		string currentChromosome = genFile.getChrName(chr);
		size_t chrLength = genFile.getChrLength(chr);

		cout << currentChromosome << "\n";
#ifdef _DEBUG
//	chrLength = 100000;
#endif


		if (cod)
		{
			f = fopen((string(outputDirectory) + "matches.cod").c_str(),"w");
			fprintf(f,"#seq_id\tchromosome\tstart\tend\tstrand\tlog10(Likelihood_Ratio)\tsite\n");
		}
		else
		{
			chrLength = min<size_t>(10000000*window,chrLength);
			barData.setNEntries(currentChromosome,chrLength/window);
		}

		seq_data seq = barData.getSeqData(currentChromosome);

		//	We now need to track down the data in the main file and normalise
		size_t gOffset = 1;

		int offset = (cg)?0:1;
		int length = 20;
		size_t gEnd = gOffset + CHAR_CHUNK_SIZE + offset + length;


#ifndef _DEBUG
		try
#endif
		{
			const char * gseq = genFile.getseq(currentChromosome.c_str(),gOffset,gEnd);

			double weight = 0.0;
			chromosome_break_data cbd;

			for (size_t i = 1;	i <= chrLength-20;	i++)
			{

				if (cg)
				{
					switch(gseq[i-gOffset])
					{
					case 'c': case 'g': case 'C': case 'G':
						weight++;
					}
				}
				else if (useBreaksInfo)
				{
					int index;
					weight += optimiser.scaleForSequence(&gseq[i-gOffset],index);
#ifdef OUTPUT_SEQS
					if ((index == refIndex[0]) || (index == refIndex[1]))
					{
						char buffer[30];
						strncpy(buffer,&gseq[i-gOffset],20);
						buffer[20] = 0;
						fprintf(f,"%i,%s,%i,%s\n",index,currentChromosome.c_str(),i,buffer);
#endif					}

				}
				else
				{	int PSSM;
					weight += optimiser.m_pssmBlocks.calcWeight(&gseq[i-gOffset],cbd,PSSM);
				}


				if (i > (gOffset + CHAR_CHUNK_SIZE))
				{
					gOffset += CHAR_CHUNK_SIZE;
					gEnd = gOffset + CHAR_CHUNK_SIZE + offset + length;
					if (gEnd > chrLength)
						gEnd = chrLength;
					gseq = genFile.getseq(currentChromosome.c_str(),gOffset,gEnd);
					}

				if (cod)
				{
					if (weight > threshold)
					{
						int pos = i - offset;
						char seq[20];
						strncpy(seq,&gseq[i-gOffset],pssmLength);
						seq[pssmLength] = 0;
						fprintf(f,"1\t%s\t%i\t%i\t%c\t%f\t%s\n",currentChromosome.c_str(),
							pos+1,pos + pssmLength,'+',weight - threshold + 2,seq);
					}
					weight = 0.0;
				}
				else
				{
					if ((i % window) == 0)
					{
						double pos = i-(window/2)+offset;
						seq.tagBarData->vData[0]->pMatElement[p] = pos;
						//					cgCount /= window;
						seq.tagBarData->vData[1]->pMatElement[p++] = weight;
						weight = 0.0;
					}
				}
			}
		}
#ifndef _DEBUG
		catch(...)
		{
			return PROC_FAILURE;
		}
#endif
	}

#ifdef OUTPUT_SEQS
	fclose(f);
#endif

	if (cod)
	{
		fclose(f);
	}
	else
	{
		barData.save((string(outputDirectory) + "dist.bar"));
	}
	
	return PROC_SUCCESS;
}