#include "pch.h"
#include "breakOptimiser.h"

#define N_POINTS_FOR_NEXT_PSSM 500

int menu_optimise_breaks(int argv, char **argc);

int main(int argv, char **argc)
{
	/* ---- */
	/* menu */
	/* ---- */
	menu_optimise_breaks(argv, argc);

	/* exit */
	exit(EXIT_SUCCESS);
}
#define WEIGHTS_FILENAME "weights.csv"

int menu_optimise_breaks(int argv, char **argc)
{
	char outputDirectory[MED_LINE_LENGTH] = "";
	char breakFileName[MED_LINE_LENGTH] = "";
	char FileName[MED_LINE_LENGTH] = "";

	int nResult;

	int nThreshold;
	bool optiMode = false;

	enum pssm_modes
	{
		none,
		initialise,
		duplicate,
		randomise,
		complement
	} pssm_mode = none;


	bool verbose = false;
	bool logNorm = true;
	int optimiseCounts = 0;
	int nChromosomes = 0;
	int nScoreChrome = 0;
	double logoScale = 1;

	int offset = 0;		//Derived from break filename


	/* ---------------------------- */
	/*     optimise breaks          */
	/* -g genomeFile				*/
	/* ---------------------------- */
	if(argv < 1)
	{
		printf("Error: parameter wrong!\n");
		exit(EXIT_FAILURE);
	}
	else if(argv == 1)
	{
		printf("/* ----------------------------- */\n");
		printf("     optimsie Breaks          \n");
		printf(" -b break file				  \n");
		printf(" -a absolute break data		  \n");
		printf(" -o output directory		  \n");
		printf(" -p <n> optimise 			  \n");
		printf(" -t <n> threshold			  \n");
		printf(" -n <n> Number of chromosomes \n");
		printf(" -s <n> Score using chromosome\n");
		printf(" -ls logo scale               \n");
		printf("							  \n");
		printf(" -i initialise PSSM           \n");
		printf(" -r randomise PSSMs           \n");
		printf(" -d duplicate PSSMs           \n");
		printf(" -c complement PSSMs           \n");

		exit(EXIT_SUCCESS);
	}

	GetCurrentDir(outputDirectory,MED_LINE_LENGTH);
	char * delim = strrchr(outputDirectory,'\\');

	int ni = 1;
	while(ni < argv)
	{
		if(strcmp(argc[ni], "-o") == 0)
		{
			ni++;
			strcpy(outputDirectory,argc[ni]);
			size_t lastChar = strlen(outputDirectory)-1;
			if (outputDirectory[lastChar] != FILE_SEP)
			{
				outputDirectory[++lastChar] =FILE_SEP;
				outputDirectory[++lastChar] = 0;
			}
			break;
		}
		ni++;
	}

	ni = 1;
	while(ni < argv)
	{
		if(strcmp(argc[ni], "-b") == 0)
		{
			strcpy(breakFileName,argc[++ni]);
			const char * p = strstr(breakFileName,"_o");
			if (p) 
				offset = atoi(p+2);
		}
		else if(strcmp(argc[ni], "-t") == 0)
		{
			nThreshold = atoi(argc[++ni]);
		}
		else if(strcmp(argc[ni], "-ls") == 0)
		{
			logoScale = atof(argc[++ni]);
		}
		else if(strcmp(argc[ni], "-n") == 0)
		{
			nChromosomes = atoi(argc[++ni]);
		}
		else if(strcmp(argc[ni], "-p") == 0)
		{
			optiMode = true;
			optimiseCounts = atoi(argc[++ni]);
		}
		else if(strcmp(argc[ni], "-s") == 0)
		{
			nScoreChrome = atoi(argc[++ni]);
		}
		else if(strcmp(argc[ni], "-i") == 0)
		{
			pssm_mode = initialise;
		}
		else if(strcmp(argc[ni], "-r") == 0)
		{
			pssm_mode = randomise;
		}
		else if(strcmp(argc[ni], "-d") == 0)
		{
			pssm_mode = duplicate;
		}
		else if(strcmp(argc[ni], "-v") == 0)
		{
			verbose = true;
		}
		else if(strcmp(argc[ni], "-c") == 0)
		{
			pssm_mode = complement;
		}
		else if(strcmp(argc[ni], "-a") == 0)
		{
			logNorm = false;
		}

		ni++;
	}

	MakeDir(outputDirectory);

	breakOptimiser optimiser(outputDirectory,nChromosomes,nScoreChrome,-offset);

	optimiser.open(breakFileName,logNorm,nChromosomes,nThreshold);

	ErrorPair score;
	if(!optimiser.readWeights() && (pssm_mode != initialise))
		exit (1);

	switch (pssm_mode)
	{
	case initialise:
		optimiser.initialisePssm();
		break;
	case complement:
		optimiser.m_pssmBlocks.makeComplement();
		break;
	case randomise:
		{
			optimiser.m_pssmBlocks.addNoiseToPssms(0.05);
			break;
		}
	case duplicate:
		if (optimiser.m_pssmBlocks[0].size() < 2)
		{
			optimiser.duplicatePssm();
			break;
		}
		// If > 1 PSSM then do a calculation so that we can work out
		//	which PSSM to duplicaye
	default:
		{
			if (optiMode != 0)
			{
				optimiser.Optimise(optimiseCounts,verbose);
			}

			score = optimiser.calculatePssmWeights(nScoreChrome);

			if (pssm_mode == duplicate)
			{
				int block = 0;

				/* Now that we have all the scores and errors for the selected genes,
					we find the PSSMs associated with the worst scoring points, and make a 
					copy of the PSSM that appears most frequently in the list of the top
					50 worst scoring points.
					This is on teh basis that this PSSM has been left trying to match a 
					wider range of sequences than it is cabale of, and so by giving it two PSSMs
					it can bifurcate and create two PSSMs that cover them between them, and so
					get a better match
					*/
				//  create a map of all the errors and their associated PSSMs
				map<double,int> pssmErrs;
				for (size_t i = 0;i < optimiser.BreakData().size();i++)
				{
					const break_data & bd = optimiser.BreakData()[i];
					const chromosome_break_data & cbd = bd.Data()[nScoreChrome];
					pssmErrs.insert(map<double,int>::value_type(cbd.m_error,bd.m_PSSM));
				}
				//	Now count up how many there are of each PSSM in the top 300.
				vector<int> pssmCount(optimiser.m_pssmBlocks[block].size());
				map<double,int>::reverse_iterator ri = pssmErrs.rbegin();
				for (int j = N_POINTS_FOR_NEXT_PSSM;(j > 0) && (ri != pssmErrs.rend());j--,ri++)
				{
					pssmCount[(*ri).second]++;
				}
				//	And then find the PSSM with the most
				int max = 0;
				int maxPssm = -1;
				for (size_t i = 0;i < pssmCount.size(); i++)
				{
					cout << "Pssm = " << i << " Count = " << pssmCount[i] << "\n";
					if (pssmCount[i] > max)
					{
						maxPssm = (int)i;
						max = pssmCount[i];
					}
				}
				//	And make a copy of it
				cout << "Pssm = " << maxPssm << " selected\n";
				optimiser.duplicatePssm(block, maxPssm);
			}
			else if (optiMode ==0)
			{
				//	Only output the scores if we are not optimising
				for (int j = 0;j <= nChromosomes;j++)
//					if (j != nScoreChrome)
						optimiser.calculatePssmWeights(j,(j==0));

				optimiser.outputScores(nChromosomes,logNorm);
			}
		}
	}

	optimiser.outputWeights(score);
	optimiser.outputPssms("breakPSSM",false,logoScale);

	nResult = PROC_SUCCESS;

	return nResult;
}