#include "pch.h"
#include "RNAStartOptimiser.h"


/*
Used to optimise a PSSM to give the best match between model and experimetal data

There are three modes it can be used.   The first is using a set of sequences and reads from teh start
of teh RNA where it is assumed that there is a fixed set of fragment starts

The second is the set of reads from across one or more genes where we have had a first cut at optimising to predict where
the breaks are.

The third is that it takes the original reads and turns them into a rolling local average.

Note that the code detects from the file name whether or not the reads are reverse reads and 
adjusts accordingly


*/

#define SEQ_LEN 40

int menu_optimise_start_seqs(int argv, char **argc);

int main(int argv, char **argc)
{
	menu_optimise_start_seqs(argv, argc);

	/* exit */
	exit(EXIT_SUCCESS);
}
#define WEIGHTS_FILENAME "weights.csv"



int menu_optimise_start_seqs(int argv, char **argc)
{
	char outputDirectory[MED_LINE_LENGTH] = "";
	char FileName[MED_LINE_LENGTH] = "";

	//	For mode 1, a file containg breaks from teh start.
	vector<string> readFiles(2);


	//	For mode 2
	string geneFileName;

	int nResult;

	enum pssm_modes
	{
		none,
		initialise,
		duplicate,
		randomise,
		complement
	} pssm_mode = none;


	int nThreshold;
	bool optiMode = false;
	bool verbose = false;
	bool logNorm = true;
	int optimiseCounts = 0;
	double logoScale = 1;
	int Ngenes = 1;
	int firstGene = 0;
	string firstGeneName;
	int window_size = 8;
	int window_scale = 5;
	int window_offset = 0;
	bool reverse = false;
	int pssm_offset = -1;

	int useFirstNnucs = SEQ_LEN;


	/* ---------------------------- */
	/*     optimise breaks          */
	/* -g genomeFile				*/
	/* ---------------------------- */
	if(argv < 1)
	{
		printf("Error: parameter wrong!\n");
		exit(EXIT_FAILURE);
	}
	else if(argv == 1)
	{
		printf("/* ----------------------------- */\n");
		printf("     optimsie Breaks          \n");
		printf(" For mode 1					  \n");
		printf(" -b break file				  \n");
		printf(" For mode 2	and mode 3		  \n");
		printf(" -g gene file name			  \n");
		printf(" -b break file				  \n");
		printf(" -n1 number of genes used	  \n");
		printf(" -n2 first gene used	(can be name)	  \n");
		printf(" -po <n> pssm offset		  \n");
		printf(" -w <n> mode 4, window size	  \n");
		printf(" -ws <n> window shift   	  \n");
		printf(" -wo <n> window offset   	  \n");
		printf(" -R reverse					  \n");
		printf("							  \n");
		printf(" -a absolute break data		  \n");
		printf(" -o output directory		  \n");
		printf(" -p <n> optimise 			  \n");
		printf(" -t <n> threshold			  \n");
		printf(" -ls logo scale               \n");
		printf("							  \n");
		printf(" -i initialise PSSM           \n");
		printf(" -r randomise PSSMs           \n");
		printf(" -d duplicate PSSMs           \n");
		printf(" -c complement PSSMs           \n");

		exit(EXIT_SUCCESS);
	}

	GetCurrentDir(outputDirectory,MED_LINE_LENGTH);
	char * delim = strrchr(outputDirectory,'\\');

	int ni = 1;
	while(ni < argv)
	{
		if(strcmp(argc[ni], "-o") == 0)
		{
			ni++;
			strcpy(outputDirectory,argc[ni]);
			size_t lastChar = strlen(outputDirectory)-1;
			if (outputDirectory[lastChar] != FILE_SEP)
			{
				outputDirectory[++lastChar] =FILE_SEP;
				outputDirectory[++lastChar] = 0;
			}
			break;
		}
		ni++;
	}

	ni = 1;
	while(ni < argv)
	{
		if(strcmp(argc[ni], "-b") == 0)
		{
			readFiles[0] = argc[++ni];
		}
		else if(strcmp(argc[ni], "-b2") == 0)
		{
			readFiles[1] = argc[++ni];
		}
		else if(strcmp(argc[ni], "-g") == 0)
		{
			geneFileName = argc[++ni];
		}
		else if(strcmp(argc[ni], "-po") == 0)
		{
			pssm_offset = atoi(argc[++ni]);
		}
		else if(strcmp(argc[ni], "-n1") == 0)
		{
			Ngenes = atoi(argc[++ni]);
		}
		else if(strcmp(argc[ni], "-n2") == 0)
		{
			++ni;
			if(strspn(argc[ni],"0123456789") == strlen(argc[ni]))
				firstGene = atoi(argc[ni]);
			else
				firstGeneName = argc[ni];
		}
		else if(strcmp(argc[ni], "-t") == 0)
		{
			nThreshold = atoi(argc[++ni]);
		}
		else if(strcmp(argc[ni], "-w") == 0)
		{
			window_size = atoi(argc[++ni]);
		}
		else if(strcmp(argc[ni], "-ws") == 0)
		{
			window_scale = atoi(argc[++ni]);
		}
		else if(strcmp(argc[ni], "-R") == 0)
		{
			reverse = true;
		}
		else if(strcmp(argc[ni], "-ls") == 0)
		{
			logoScale = atof(argc[++ni]);
		}
		else if(strcmp(argc[ni], "-p") == 0)
		{
			optiMode = true;
			optimiseCounts = atoi(argc[++ni]);
		}
		else if(strcmp(argc[ni], "-d") == 0)
		{
			pssm_mode = duplicate;
		}
		else if(strcmp(argc[ni], "-v") == 0)
		{
			verbose = true;
		}
		else if(strcmp(argc[ni], "-i") == 0)
		{
			pssm_mode = initialise;
		}
		else if(strcmp(argc[ni], "-r") == 0)
		{
			pssm_mode = randomise;
		}
		else if(strcmp(argc[ni], "-c") == 0)
		{
			pssm_mode = complement;
		}
		else if(strcmp(argc[ni], "-a") == 0)
		{
			logNorm = false;
		}

		ni++;
	}

	MakeDir(outputDirectory);

	RNAStartOptimiser optimiser(outputDirectory,window_size,window_scale,window_offset,pssm_offset);
	ErrorPair score;
	if (pssm_mode != initialise)
		optimiser.readWeights();

	switch (pssm_mode)
	{
	case initialise:
		optimiser.m_pssmBlocks.initialise();
		break;
	case complement:
		optimiser.m_pssmBlocks.makeComplement();
		break;
	case randomise:
		optimiser.m_pssmBlocks.addNoiseToPssms(0.005);
		break;
	case duplicate:
		#ifdef TRIPLE_PSSM
		if (//(optimiser.m_pssmBlocks[0].size() < 2) || 
			(optimiser.m_pssmBlocks[0][0].m_optimising && (optimiser.m_pssmBlocks[0].size() < 2)) ||
//			(optimiser.m_pssmBlocks[1][0].m_optimising && (optimiser.m_pssmBlocks[1].size() < 2)) ||
			(optimiser.m_pssmBlocks[2][0].m_optimising && (optimiser.m_pssmBlocks[2].size() < 2)))
		#else
		if (optimiser.m_pssmBlocks[0].size() < 3)
		#endif
		{
			optimiser.m_pssmBlocks.duplicate();
			break;
		}
		//	If 3 or greater, run straight on into calculating PSSM weights
	default:
		{

			mRNAinfoSet genes;
			if (geneFileName.empty())
			{
				//	Mode 1, use a pre created file
				optimiser.open(readFiles[0].c_str(),useFirstNnucs,logNorm);
			}
			else
			{
				//	Mode 2 or mode 3
				if(!genes.open(geneFileName.c_str()))
				{
					cout << "Failed to open gene file\n";
					return PROC_FAILURE;
				}
				genes.loadAll();

				if (optimiser.m_window_size)
				{
					//	Mode 3
					//  load the original breakdata
					genes.loadBreakdata(readFiles);
					genes.orderGenes(reverse?1:0);
					optimiser.getTopGenes(genes,Ngenes,firstGene,firstGeneName);
					if (reverse)
						optimiser.complementGenes();
					optimiser.generateSmoothedBreakInfo();

				}
				else
				{
					//	Mode 2
					//	Use an output file from an optimisation with a previous PSSM set
					optimiser.open(readFiles[0].c_str(),genes,Ngenes,logNorm);
					useFirstNnucs = 0;	//Use the full length
				}
			}

			if (optiMode != 0)
			{
				optimiser.Optimise(optimiseCounts,verbose);
			}

			score = optimiser.calculatePssmWeights();

			if (pssm_mode == duplicate)
			{
				int block,pssmInfo;
				if (optimiser.m_pssmBlocks[0][0].m_optimising)
				{

					block = 0;
					pssmInfo = 0;
				}
				else if (optimiser.m_pssmBlocks[2][0].m_optimising)
				{

					block = 2;
					pssmInfo = 1;
				}
				else 
				{
					block = 1;
					pssmInfo = 1;
				}

				/* Now that we have all the scores and errors for the selected genes,
					we find the PSSMs associated with the worst scoring points, and make a 
					copy of the PSSM that appears most frequently in the list of the top
					50 worst scoring points.
					This is on teh basis that this PSSM has been left trying to match a 
					wider range of sequences than it is cabale of, and so by giving it two PSSMs
					it can bifurcate and create two PSSMs that cover them between them, and so
					get a better match
					*/
				//  create a map of all the errors and their associated PSSMs
				map<double,int> pssmErrs;
				for (size_t i = 0;i < optimiser.RNAdata.size();i++)
				{
					for (size_t j = 10; j < optimiser.RNAdata[i].m_modelInfo.size()-10; j++)
					{
						modelInfo & mi = optimiser.RNAdata[i].m_modelInfo[j];
						pssmErrs.insert(map<double,int>::value_type(mi.m_error,mi.m_PSSM[pssmInfo]));
					}
				}
				//	Now count up how many there are of each PSSM in the top 50.
				vector<int> pssmCount(optimiser.m_pssmBlocks[block].size());
				map<double,int>::reverse_iterator ri = pssmErrs.rbegin();
				for (int j = 50;j > 0;j--,ri++)
				{
					pssmCount[(*ri).second]++;
				}
				//	And then find the PSSM with the most
				int max = 0;
				int maxPssm = -1;
				for (size_t i = 0;i < pssmCount.size(); i++)
				{
					cout << "Pssm = " << i << " Count = " << pssmCount[i] << "\n";
					if (pssmCount[i] > max)
					{
						maxPssm = (int)i;
						max = pssmCount[i];
					}
				}
				//	And make a copy of it
				cout << "Pssm = " << maxPssm << " selected\n";
				optimiser.m_pssmBlocks.duplicate(block, maxPssm);
			}
			optimiser.outputScores(logNorm);
			optimiser.outputPssms("rnaPSSM",true,logoScale);
		}
	}

	optimiser.outputWeights(score);
	optimiser.outputPeaks();

	nResult = PROC_SUCCESS;

	return nResult;
}