/* ----------------------------------------------------------------------- */
/*  AffyLib.h : interface of the affymetrix library                        */
/*  Author : Ji HongKai ; Time: 2004.08                                    */
/* ----------------------------------------------------------------------- */

/* ----------------------------------------------------------------------- */
/*                              Declaration                                */
/* ----------------------------------------------------------------------- */
#ifndef AFFYLIB_H
#define AFFYLIB_H


#define FLOAT_SIZE sizeof(float)
#define DOUBLE_SIZE sizeof(double)
#define INT_SIZE sizeof(int)
#define DWORD_SIZE sizeof(unsigned int)
#define ULONG_SIZE sizeof(unsigned long)
#define LONG_SIZE sizeof(long)
#define SHORT_SIZE sizeof(short)
#define USHORT_SIZE sizeof(unsigned short)
#define CHAR_SIZE sizeof(char)
#define UCHAR_SIZE sizeof(unsigned char)
#define WCHAR_SIZE sizeof(wchar_t)
#define WCHAR16_SIZE sizeof(unsigned short)

typedef unsigned short wchar_t16;

struct tagBARSeq
{
	/* sequence name */
	struct tagString *pSeqName;

	/* sequence group name */
	struct tagString *pSeqGroupName;
	
	/* sequence version */
	struct tagString *pSeqVersion;

	/* paramenter name/value pairs */
	int nParamNum;
	struct tagString **vParamName;
	struct tagString **vParamValue;

	/* data points */
	int nDataNum;
	int nColNum;
	struct DOUBLEMATRIX **vData;
};

struct tagBARData
{
	/* magic number; version number */
    char strMagicnumber[9];
	float fVersionnumber;
	
	/* number of sequences and columns, */
	int nSeqNum;
	int nColNum;

	/* column type */
	struct INTMATRIX *pFieldType;

	/* paramenter name/value pairs */
	int nParamNum;
	struct tagString **vParamName;
	struct tagString **vParamValue;

	/* sequence information */
	struct tagBARSeq **vSeqData;
};

/* ----------------------------------------------------------------------- */ 
/*  Affy_BARSeq_Create()                                                   */
/*  create a BARSeq object.                                                */
/* ----------------------------------------------------------------------- */ 
struct tagBARSeq *Affy_BARSeq_Create();

/* ----------------------------------------------------------------------- */ 
/*  Affy_BARSeq_Destroy()                                                  */
/*  delete a BARSeq object.                                                */
/* ----------------------------------------------------------------------- */ 
void Affy_BARSeq_Destroy(struct tagBARSeq **pBARSeq);

/* ----------------------------------------------------------------------- */ 
/*  Affy_BARData_Create()                                                  */
/*  create a BARData object.                                               */
/* ----------------------------------------------------------------------- */ 
struct tagBARData *Affy_BARData_Create();

/* ----------------------------------------------------------------------- */ 
/*  Affy_BARData_Destroy()                                                 */
/*  delete a BARData object.                                               */
/* ----------------------------------------------------------------------- */ 
void Affy_BARData_Destroy(struct tagBARData **pBARData);

/* ----------------------------------------------------------------------- */ 
/*  Affy_LoadBAR()                                                         */
/*  Loading raw data from a single affymetrix's *.bar file                 */
/* ----------------------------------------------------------------------- */ 
struct tagBARData *Affy_LoadBAR(char strFileName[]);

/* ----------------------------------------------------------------------- */ 
/*  Affy_LoadBAR_Fast()                                                    */
/*  Loading raw data from a single affymetrix's *.bar file                 */
/* ----------------------------------------------------------------------- */ 
struct tagBARData *Affy_LoadBAR_Fast(char strFileName[]);

/* ----------------------------------------------------------------------- */ 
/*  Affy_SaveBAR_Columns_Fast()                                            */
/*  Saving specified columns of a bar data object to a *.bar file          */
/* ----------------------------------------------------------------------- */ 
int Affy_SaveBAR_Columns_Fast(char strFileName[], struct tagBARData *pBARData,
							  struct INTMATRIX *pCol);

#endif
