/* ----------------------------------------------------------------------- */
/*  SequenceLib.c : implementation of the sequence library                 */
/*  Author : Ji HongKai ; Time: 2004.07                                    */
/*  Abridged version Nigel Dyer August 2011								   */
/* ----------------------------------------------------------------------- */

#include "stdlib.h"
#include "stdio.h"
#include "string.h"
#include "math.h"
#include "limits.h"
#include "time.h"

#include "MatrixLib.h"
#include "StringLib.h"
#include "SequenceLib.h"
/* ----------------------------------------------------------------------------- */ 
/*                       struct tagSequence *SequenceCreate()                    */
/* This function is used for creating a sequence structure.                      */
/* ----------------------------------------------------------------------------- */ 
struct tagSequence *SequenceCreate()
{
	struct tagSequence *pNewSequence;

	/* Init */
	pNewSequence = NULL;

	/* new */
	pNewSequence = (struct tagSequence *)malloc(sizeof(struct tagSequence));
	if( pNewSequence == NULL )
		return NULL;
	pNewSequence->m_nIndex = -1;
	pNewSequence->m_nLength = 0;
	pNewSequence->m_pSequence = NULL;
	strcpy(pNewSequence->m_strAlias, "");
	pNewSequence->m_nStart = 0;
	pNewSequence->m_nEnd = 0;
	pNewSequence->m_pNext = NULL;

	/* return */
	return pNewSequence;
}

/*                           void SequenceDelete()                               */
/* This function is used for deleting a sequence.                                */
/* ----------------------------------------------------------------------------- */ 
void SequenceDelete(struct tagSequence *pDelSequence)
{
	if( pDelSequence != NULL )
	{
		pDelSequence->m_pNext = NULL;
		DeleteString(pDelSequence->m_pSequence);
		free(pDelSequence);
	}	
}

/* ----------------------------------------------------------------------------- */ 
/*                            int  SequenceAddTail()                             */
/* This function is used for adding a segment to the tail of a sequence.         */
/* If the process is successful, it will return 1;                               */
/* else it will return 0.                                                        */            
/* ----------------------------------------------------------------------------- */ 
int SequenceAddTail(struct tagSequence *pSequence, char strLine[])
{
	/* nLen is used for recording the length of a string */
	int nLen;
	/* pString is used for manipulating the string */
	char *pString;

	/* check parameter */
	if(pSequence == NULL)
		return PROC_FAILURE;
	nLen = (int)strlen(strLine);
	if(nLen == 0)
		return PROC_SUCCESS;

	/* add */
	if(pSequence->m_pSequence == NULL)
	{
		pSequence->m_pSequence = CreateString(nLen);
		if(pSequence->m_pSequence == NULL)
		{
			printf("Error in SequenceAddTail: cannot create string file...\n");
			return PROC_FAILURE;
		}
		pString = pSequence->m_pSequence->m_pString;
		strcat(pString, strLine);
		pSequence->m_nLength = nLen;
		pSequence->m_nStart = 1;
		pSequence->m_nEnd = nLen;
	}
	else
	{
		int currLen = pSequence->m_pSequence->m_nLength;
		pSequence->m_pSequence->m_nLength += nLen;
		pSequence->m_pSequence->m_pString = (char *)realloc(pSequence->m_pSequence->m_pString,
			sizeof(char)*(pSequence->m_pSequence->m_nLength+1));
//		strcat(pSequence->m_pSequence->m_pString, strLine);
		strcpy(pSequence->m_pSequence->m_pString + currLen, strLine);
		pSequence->m_nLength += nLen;
		pSequence->m_nEnd += nLen;

/*
		Old inefficient code in its entirity
		pSequence->m_pSequence->m_nLength += nLen;
		pString = pSequence->m_pSequence->m_pString;
		pTempString = (char *)malloc(sizeof(char)*(pSequence->m_pSequence->m_nLength+1));
		if(pTempString == NULL)
			return PROC_FAILURE;
		strcpy(pTempString, pString);
		free(pString);
		strcat(pTempString, strLine);
		pSequence->m_pSequence->m_pString = pTempString;
		pSequence->m_nLength += nLen;
		pSequence->m_nEnd += nLen;
		*/
	}

	/* return */
	return PROC_SUCCESS;
}
